(***************************************************************************

Abstract syntax trees for JavaCard source files

VerifiCard Project - Dmons research team - LRI - Universit Paris XI

$Id: ast_types.ml,v 1.13 2005/03/08 13:36:53 marche Exp $

***************************************************************************)

(*s Qualified identifiers *)

type identifier = Location.t * string

type qualified_ident = identifier list

(*s Modifiers *)
 
type modifier =
    [ `STATIC | `FINAL | `PUBLIC | `PRIVATE | `PROTECTED | `NATIVE 
    | `SYNCHRONIZED | `ABSTRACT | `TRANSIENT (* "threadsafe" ? *)
    | `PURE | `SPEC_PUBLIC | `MODEL ]  

type modifiers = modifier list;;

(* Javadoc comments *)

type javadoc = string;;

(*s type expressions *)

open Java_context_types;;

type type_expr = 
  | Base_type of string
  | Type_name of qualified_ident
  | Array_type_expr of type_expr
;;
 
(*s literals *)

type literal =
  | Int of int
  | Float of Why.float_repr
  | Bool of bool
  | String of string
  | Char of char
  | Null

(*s expressions *)

type quantifier = Forall | Exists;;

type variable_id =
  | Simple_id of identifier
  | Array_id of variable_id
;;

type incrdecrop = [ `PREINCR | `PREDECR | `POSTINCR | `POSTDECR ]

type field_access =
  | Super_access of identifier
  | Primary_access of expr * identifier
  | Typed_primary_access of expr * instance_variable_entry
  | Typed_constant_access of constant_entry
  | Array_length of expr

and expr =
  { expr_loc : Location.t ;
    expr_node : expr_node }

and expr_node =
  | Lit of literal
  | Bin of expr * string * expr         (*r binary operations *)
  | Un of string * expr                 (*r (pure) unary operations *)
  | PrePostIncrDecr of incrdecrop * expr (*r pre-post incr/decr operations *)
  | Var of identifier
  | Typed_var of local_variable_entry
  | Static_class of class_entry
  | Static_interface of interface_entry
  | Assign_name of qualified_ident * string * expr  
  | Assign_local_var of local_variable_entry * string * expr  
  | Assign_field of field_access * string * expr  
  | Assign_array of expr * expr * string * expr  
      (*r [Assign_array(e1,e2,op,e3)] is [e1[e2] op e3] *)
      (*r assignment op is =, +=, etc. *)
  | Typed_assign_array of java_type * expr * expr * string * expr  
  | If_expr of expr * expr * expr       
  | This
  | Field_access of field_access
  | Method_call of expr option * identifier * expr list
  | Typed_method_call of expr * method_entry * expr list
  | Class_instance_creation of qualified_ident * expr list
  | Typed_class_instance_creation of constructor_entry * expr list
  | Array_creation of array_creation
  | Array_access of expr * expr
  | Typed_array_access of java_type * expr * expr
  | Cast of type_expr * expr
  | Typed_cast of java_type * expr
  | Instanceof of expr * type_expr
  | Typed_instanceof of expr * java_type
      (* in JML annotations only *)
  | Result  
  | Old of expr 
  | Fresh of expr 
  | Type of type_expr 
  | Typed_type of java_type
  | Typeof of expr 
  | Typed_typeof of expr * java_type
  | Quantifier of quantifier * type_expr * variable_id list * expr * expr  
  | Typed_quantifier of quantifier * 
      local_variable_entry list * expr * expr  
  | Fpi of expr * (string * string * string) * (string * string * string)

and array_creation =
  | Explicit_array_creation of expr * array_creation
  | Implicit_array_creation of type_expr
  | Typed_implicit_array_creation of java_type


(*s variable declarations *)
      
type variable_initializer =
  | Simple_initializer of expr
  | Array_initializer of variable_initializer list
;;

type variable_declarator = 
    { variable_id : variable_id ;
      variable_initializer : variable_initializer option }
;;
 
      
type variable_declaration =
    { variable_modifiers : modifiers ;
      variable_type : type_expr ;
      variable_decls : variable_declarator list }
;;

(*s formal parameters *)

type parameter =
  | Simple_parameter of type_expr * identifier
  | Array_parameter of parameter
;;

(*s JML annotations *)

type privacy =
  | Privacy_none 
  | Privacy_public 
  | Privacy_protected 
  | Privacy_private
;;

type set_ref = 
  | Set_array_index of expr  
      (*r e[e] *)
  | Set_array_interval of expr * expr  
      (*r e[e..e] *)
  | Set_array
      (*r e[*] *)
  | Set_field of instance_variable_entry
      (* e.f *)
  | Set_fieldraw of identifier
;;

type assignable_object =
  | Assignable_everything
  | Assignable_nothing
  | Assignable_set_ref of expr * set_ref list 
  | Typed_assignable_set_ref of Java_context_types.java_type * expr * set_ref list 
  | Assignable_all_fields of expr
;;

type assignable_clause = (assignable_object * expr option) list      (*r obj if e *)
;;

type loop_annotation = 
    {
      loop_invariant : expr;
      loop_modifies : assignable_clause list option;
      typed_loop_modifies : assignable_clause option;
      loop_decreases : expr;
    }

  
type signals_clause =
    {
      signals_exception : type_expr; 
      signals_variable : identifier option;
      signals_post : expr;
    }

type ensures_clause = expr;;

type requires_clause = expr;;

type when_clause = expr;;

type measured_clause = expr;;

type diverges_clause = expr;;

type local_spec_var_decl =
  | Local_spec_var_decl_model of type_expr * variable_id list
  | Local_spec_var_decl_ghost of type_expr * variable_id list
;;

type spec_var_decls =
    { forall_var_decls : (type_expr * variable_id list) list ;
      let_var_decls : local_spec_var_decl list}
;; 

type spec_header =
    { requires_clauses : requires_clause list ;
      when_clauses : when_clause list ;
      measured_clauses : measured_clause list }
;;

type exceptional_simple_spec_body =
    { exceptional_assignable_clauses : assignable_clause list;
      exceptional_signals_clauses : signals_clause list ;
      exceptional_diverges_clauses : diverges_clause list } 
;;

type normal_simple_spec_body =
    { normal_assignable_clauses : assignable_clause list;
      normal_ensures_clauses : ensures_clause list ;
      normal_diverges_clauses : diverges_clause list } 
;;

type simple_spec_body =
    { assignable_clauses : assignable_clause list;
      ensures_clauses : ensures_clause list ;
      signals_clauses : signals_clause list ;
      diverges_clauses : diverges_clause list }
;;

type generic_spec_body =
  | Simple_spec_body of simple_spec_body
  | Nested_spec_body of generic_spec_case list

and generic_spec_case =
    { generic_spec_var_decls : spec_var_decls ;
      generic_spec_header : spec_header ;
      generic_spec_body : generic_spec_body }
;;

type exceptional_spec_body =
  | Simple_exceptional_spec_body of exceptional_simple_spec_body
  | Nested_exceptional_spec_body of exceptional_spec_case list

and exceptional_spec_case =
    { exceptional_spec_var_decls : spec_var_decls ;
      exceptional_spec_header : spec_header ;
      exceptional_spec_body : exceptional_spec_body }
;;

type normal_spec_body =
  | Simple_normal_spec_body of normal_simple_spec_body
  | Nested_normal_spec_body of normal_spec_case list

and normal_spec_case =
    { normal_spec_var_decls : spec_var_decls ;
      normal_spec_header : spec_header ;
      normal_spec_body : normal_spec_body }
;;

type spec_case = 
  | Generic of generic_spec_case
  | Behavior of privacy * generic_spec_case
  | Exceptional_behavior of privacy * exceptional_spec_case
  | Normal_behavior of privacy * normal_spec_case
;;

type generic_conjoinable_spec =
    { conjoinable_spec_var_decls : spec_var_decls ;
      conjoinable_spec_body : simple_spec_body }
;;

type conjoinable_spec =
  | Generic_conjoinable_spec of generic_conjoinable_spec
  | Behavior_conjoinable_spec of spec_case

type typed_compact_spec =
    {
      typed_compact_decls : spec_var_decls;
      typed_compact_requires : expr;
      typed_compact_assignable : assignable_clause;
      typed_compact_ensures : expr;
      typed_compact_signals : (java_type * 
			       local_variable_entry option *
			       expr) option;
    }

type compact_spec =
    { compact_decls : spec_var_decls; 
      compact_requires : requires_clause ;
      compact_when : when_clause (* list ? *) ;
      compact_measured : measured_clause (* list ? *);
      compact_assignable : assignable_clause (* list *);
      compact_ensures : ensures_clause ;
      compact_signals : signals_clause ;
      compact_diverges : diverges_clause }
;;

type method_specification =
  | Non_extending of spec_case list
  | Extending_also of spec_case list
  | Extending_and of conjoinable_spec list
  | Compact of compact_spec
  | Typed_compact of typed_compact_spec
;;

type statement_specification =
    { (* decls : spec_var_decls; *)
      statement_requires : requires_clause list ;
      (* when : when_clause (* list ? *) *) 
      (* compact_measured : measured_clause (* list ? *); *)
      statement_assignable : assignable_clause list ;
      statement_ensures : ensures_clause list ;
      statement_signals : signals_clause list ;
      (* compact_diverges : diverges_clause *) }
;;

type jml_variable_declaration =
    { jml_variable_modifiers : modifiers;
      jml_variable_type : type_expr;
      jml_variable_decls : variable_id list }
;;

type jml_declaration =
  | Invariant of expr
  | Jml_variable_declaration of jml_variable_declaration
  | Represents of identifier * expr
  | Typed_represents of instance_variable_entry * expr
;;


type jml_specification =
  | Jml_pure
  | Jml_spec_public
  | Jml_declaration of jml_declaration list
  | Jml_method_specification of method_specification
  | Jml_loop_annotation of loop_annotation
  | Jml_assertion of expr


(*s statements *)

type switch_label =
  | Case of expr
  | Default

type statement =
  { statement_loc : Location.t ;
    statement_node : statement_node }

and statement_node =
  | Skip                  (*r empty statement *)
  | Expr of expr
  | Typed_expr of java_type * expr
  | Var_decl of variable_declaration
  | Typed_var_decl of 
      local_variable_entry 
      * variable_initializer option * statement 
  | Throw of expr
  | Return of expr option
  | Typed_return of java_type * expr option
  | Break of string
  | Continue of string
  | Label of string * statement
  | If_statement of expr * statement * statement
  | While of loop_annotation * expr * statement
  | Do of loop_annotation * statement * expr
  | For of 
      loop_annotation * statement list * expr * statement list 
      * statement  
  | For_decl of 
      loop_annotation * variable_declaration * expr * statement list 
      * statement  
  | For_typed_decl of 
      local_variable_entry * variable_initializer option
      * loop_annotation * expr * statement list * statement  
  | Try of block * (parameter * block) list * block option
  | Typed_try of block * (string * Java_context_types.java_type * block) list * block option
  | Switch of expr * (switch_label list * block) list
  | Block of block
  | Synchronized of expr * block
  | Jml_assert of expr
  | Annotated of statement_specification * statement
  | Typed_annotated of typed_compact_spec * statement

and block = statement list
;;

(*s method declarations *)

type method_declarator =
  | Simple_method_declarator of identifier * parameter list
  | Array_method_declarator of method_declarator
;;

type method_declaration =
    { method_specification : method_specification ;
      method_modifiers : modifiers ;
      method_return_type : type_expr option ;
      method_declarator : method_declarator ;
      method_throws : qualified_ident list }
;;

(*s constructor declarations *)


type explicit_constructor_invocation =
  | Invoke_none
  | Invoke_this of expr list
  | Invoke_super of expr list

type constructor_declaration =
    { constr_specification : method_specification ;
      constr_modifiers : modifiers ;
      constr_name : identifier ;
      constr_parameters : parameter list ;
      constr_throws : qualified_ident list;
      constr_body : explicit_constructor_invocation * block }
;;

      
      
      
(*s class declarations *)

type field_declaration =
  | Method of method_declaration * block option
  | Typed_method of 
      method_entry * method_specification * block option
  | Constructor of constructor_declaration 
  | Typed_constructor of 
      constructor_entry * method_specification *
      explicit_constructor_invocation * block 
  | Variable of variable_declaration
  | Typed_variable of 
      instance_variable_entry * variable_initializer option
  | Static_initializer of block
  | Specification of jml_declaration list
;;


type class_declaration =
    { class_modifiers : modifiers;
      class_name : identifier;
      class_extends : qualified_ident option;
      class_implements : qualified_ident list;
      class_fields : (javadoc * field_declaration) list
    }

(*s interface declarations *)

type interface_member_declaration =
  | Constant_declaration of variable_declaration
  | Typed_constant_declaration of 
      constant_entry *
      (*i instance_variable_entry * i*)
      variable_initializer option
  | Method_header of method_declaration
  | Typed_method_header of method_entry * method_specification
;;

type interface_declaration =
    { interface_modifiers : modifiers;
      interface_name : identifier;
      interface_extends : qualified_ident list;
      interface_members : (javadoc * interface_member_declaration) list
    }
;;

(*s compilation units *)

type type_declaration =
  | Class of class_declaration
  | Interface of interface_declaration
  | Typed_class of class_entry * class_declaration
  | Typed_interface of interface_entry * interface_declaration


type import_statement =
  | Import_package of qualified_ident
  | Import_class_or_interface of qualified_ident

type compilation_unit =
    {
      cu_package : qualified_ident;
      cu_imports : import_statement list;
      cu_type_decls : (javadoc * type_declaration) list
    }

