
let c = ref stdout;;

let log =
  c := open_out "krakatoa.log";
  Format.formatter_of_out_channel !c
;;

let close_log () =
  Format.fprintf log "End of log.@.";
  close_out !c;
  c := stdout
;;

let dump_env = ref false;;


(* Java input *)

let source_package = ref ".";;

(*

  parsing only, for testing the parser

*)

let parse_only = ref false;;


let methods = ref [];;

let files_to_parse () = !methods;;

let add_method m =
  methods := !methods @ [m]
;;

let mem_methods c m =
    List.mem (c ^ "." ^ m) !methods
    or
    List.mem (c ^ ".*") !methods
;;

(*
let output_dir = ref ".";;
*)

let rec split ch s =
  try
    let i = String.index s ch in
    let h = String.sub s 0 i
    and t = String.sub s (i+1) (String.length s - i - 1)
    in
    h::(split ch t)
  with
    Not_found -> [s]
;;
      
let get_krakatoa_path () =
  try
    let v = Sys.getenv "KRAKATOAPATH" in
    Format.fprintf log "KRAKATOAPATH is set to %s@." v;
    if v="" then [] else split ':' v
  with
      Not_found -> 
	let p = Filename.concat Version.libdir "java_api" in
	Format.fprintf log 
	  "KRAKATOAPATH is not set, using %s as default@." p;
	[p]
;;

let input_path = ref (get_krakatoa_path ());;

let get_krakatoa_lib () =
  try
    let v = Sys.getenv "KRAKATOALIB" in
    Format.fprintf log "KRAKATOALIB is set to %s@." v;
    v;
  with
      Not_found -> 
	Format.fprintf log "KRAKATOALIB is not set, using %s as default@." 
	  Version.libdir;
	Version.libdir;;

let krakatoa_lib = ref (get_krakatoa_lib ());;

let nojavalang = ref false;;


(*

  model

*)


let use_value_model = ref false;;

let use_global_memory_model = ref false;;

(*

Miscellaneous

*)

let stop_on_error = ref true;;

let verbose = ref 0;;

(*

  Output language

*)

type available_output_language = [`COQ | `SIMPLIFY | `HARVEY | `PVS]

let output_language = ref `COQ;;


let why_validation = ref false;;

let why_black_box = ref false

let coqopts = ref "";;

let coqdirs = ref "";;



(* 

   Parsing of arguments

*)

let options =
  [

   "-dump", Arg.Set(dump_env)
     , "dump typing environments" ;

   "-p", Arg.String((:=) source_package)
     , "main source package" ;
   "-parse-only", Arg.Set(parse_only)
     , "perform only parsing" ;
   "-I", Arg.String(fun s -> input_path := s :: !input_path)
     , "input path" ;
   "-nojavalang", Arg.Set(nojavalang)
     , "do not import java.lang package" ;

   "-coqdir", Arg.String(fun s -> coqdirs := !coqdirs ^ " -I " ^ s)
     , "additional input path to pass to coqc using -I" ;
   "-coqopt", Arg.String(fun s -> coqopts := !coqopts ^ " " ^ s)
     , "additional option to give to coqc" ;
(*
   "-d", Arg.String((:=) output_dir), "output directory" ;
*)
   "-v", Arg.Unit(fun () -> incr verbose), "increments verbosity" ;
   "-k", Arg.Clear(stop_on_error), "do not stop on first error" ;

   "-valid", Arg.Set(why_validation)
     , "produce validation (incompatible with -bb)" ;
   "-novalid", Arg.Clear(why_validation)
     , "do not produce validation" ;
   "-bb", Arg.Set why_black_box
     , "use Why black boxes (incompatible with -valid)" ;

(*
   "-valuemodel",  Arg.Set(use_value_model)
     , "use the value model for translation (unstable)" ;
*)
   "-globalmemorymodel",  Arg.Unit(
     fun () -> 
       if !output_language <> `COQ	   
       then 
	 Format.printf "option -globalmemorymodel only valid with Coq output, ignored@."
	else
	  use_global_memory_model := true)
     , "use the global memory model for translation" ;
   "-localmemorymodel",  Arg.Clear(use_global_memory_model)
     , "use the local memory model for translation (default)" ;

   "-coq", Arg.Unit(fun () -> output_language := `COQ)
     , "produce output for the Coq proof assistant" ;
   "-simplify", Arg.Unit(
     fun () -> 
       if !use_global_memory_model	   
       then 
	 Format.printf "option -simplify is incompatible with -globalmemorymodel, ignored@."
       else
	 output_language := `SIMPLIFY)
     , "produce output for the Simplify prover" ;
   "-harvey", Arg.Unit(
     fun () -> 
       if !use_global_memory_model	   
       then 
	 Format.printf "option -harvey is incompatible with -globalmemorymodel, ignored@."
       else
	 output_language := `HARVEY)
     , "produce output for the haRVey prover" ;
   "-pvs", Arg.Unit(
     fun () -> 
       if !use_global_memory_model	   
       then 
	 Format.printf "option -pvs is incompatible with -globalmemorymodel, ignored@."
       else
	 output_language := `PVS)
     , "produce output for PVS" ;
   
  ]
;;

let usage = "krakatoa [options] class.method ..."

let parse_args () =
  Format.fprintf log "Command line is:";
  for i=0 to Array.length Sys.argv - 1 do
    Format.fprintf log " %s" Sys.argv.(i);
  done;
  Format.fprintf log "@.";
  Arg.parse options add_method usage;
  if !methods = []
  then (Arg.usage options usage; exit 1);
  if !why_validation && !why_black_box then (Arg.usage options usage; exit 1)
;;


let split_extension filename =
  try
    let dotindex = String.rindex filename '.' in
    (String.sub filename 0 dotindex,
     String.sub filename (dotindex+1) (String.length filename - dotindex - 1))
    with Not_found -> filename,""
;;

