
open Java_context_types;;
open Ast_types;;
open Ast;;

let ill_typed msg = Typing.ill_typed Location.dummy msg;;

let effects_union eff1 eff2 =
  {
    reads_heap = eff1.reads_heap or eff2.reads_heap;
    reads_alloc_table = eff1.reads_alloc_table or eff2.reads_alloc_table;
    reads = Inttagset.union eff1.reads eff2.reads;
    writes_heap = eff1.writes_heap or eff2.writes_heap;
    writes_everything = eff1.writes_everything or eff2.writes_everything;
    writes_alloc_table = eff1.writes_alloc_table or eff2.writes_alloc_table;
    writes = Inttagset.union eff1.writes eff2.writes;
    reads_int_array = eff1.reads_int_array or eff2.reads_int_array;
    writes_int_array = eff1.writes_int_array or eff2.writes_int_array;
    reads_bool_array = eff1.reads_bool_array or eff2.reads_bool_array;
    writes_bool_array = eff1.writes_bool_array or eff2.writes_bool_array;
    reads_obj_array = eff1.reads_obj_array or eff2.reads_obj_array;
    writes_obj_array = eff1.writes_obj_array or eff2.writes_obj_array;
    throws = eff1.throws or eff2.throws 
	       (*i Inttagset.union eff1.throws eff2.throws i*);
    breaks = eff1.breaks or eff2.breaks
	       (*i Inttagset.union eff1.breaks eff2.breaks i*);
  }
;;

let rec expr_effects eff e =
  match e.expr_node with
    | Lit _ -> eff
    | Un (op, e) -> expr_effects eff e 
    | PrePostIncrDecr (op, e) -> 
	prepostincrdecr_effects eff e
    | Bin (e1, op, e2) ->
	let eff1 = expr_effects eff e1 in
	expr_effects eff1 e2
    | If_expr (e1, e2, e3) ->
	let eff1 = expr_effects eff e1 in
	let eff2 = expr_effects eff1 e2 in
	expr_effects eff2 e3
    | Typed_var lve -> eff
    | Field_access fa ->
	field_access_effects eff fa
    | Typed_array_access (t, e1, e2) -> 
	let eff = expr_effects eff e1 in
	let eff = expr_effects eff e2 in
	let eff = {eff with reads_alloc_table = true } in
	begin
	  match t with
	    | Integer_type ->
		{ eff with reads_int_array = true }
	    | Boolean_type ->
		{ eff with reads_bool_array = true }
	    | _ -> 
		{ eff with reads_obj_array = true }
	end
    | Old e -> expr_effects eff e
    | Fresh e -> expr_effects eff e
    | Typed_instanceof (e,t) -> expr_effects eff e
    | Typed_type te -> eff
    | Typed_typeof (e,t) -> expr_effects eff e
    | Typed_cast (t, e) -> expr_effects eff e
    | Typed_assign_array (t,e1,e2,op,e3) ->
	let eff = expr_effects eff e1 in
	let eff = expr_effects eff e2 in
	let eff = expr_effects eff e3 in
	let eff = { eff with reads_heap = true; writes_heap = true } in
	begin
	  match t with
	    | Integer_type ->
		{ eff with reads_int_array = true ; writes_int_array = true }
	    | Boolean_type ->
		{ eff with reads_bool_array = true ; writes_bool_array = true }
	    | _ -> 
		{ eff with reads_obj_array = true ; writes_obj_array = true }
	end
    | Assign_field (f,op,e) -> 
	let eff1 = expr_effects eff e in
	begin
	  match f with
	    | Super_access(id) -> assert false
	    | Primary_access(f,(loc,id)) -> assert false
	    | Typed_constant_access _ -> assert false
	    | Typed_primary_access(f,ive) ->	
		{eff1 with 
		   reads = Inttagset.add 
			      ive.instance_variable_entry_tag
			      ive eff1.reads ;
		   writes = Inttagset.add 
			      ive.instance_variable_entry_tag
			      ive eff1.writes ;
		   writes_heap = 
			     if ive.instance_variable_entry_is_static
			     then eff1.writes_heap
			     else true }
	    | Array_length(_) -> assert false
	end
    | Assign_local_var (lve, op, e) -> 
	expr_effects eff e
    | Method_call _ -> failwith "effect(untyped method call)"
    | Typed_method_call (e, me, args) -> 
	let eff1 =
	  List.fold_right
	    (fun e eff -> expr_effects eff e)
	    args
	    (expr_effects eff e)
	in
	effects_union me.method_entry_routine.routine_entry_effects eff1
    | Typed_quantifier(_,lve, e1, e2) ->
	let eff1 = expr_effects eff e1 in
	expr_effects eff1 e2
    | Fpi(e,f1,f2) -> expr_effects eff e
    | Result -> eff
    | Typed_class_instance_creation (ce, args) -> 
	let eff =
	  List.fold_right
	    (fun e eff -> expr_effects eff e)
	    args
	    eff
	in
	let eff = 
	  effects_union ce.constructor_entry_routine.routine_entry_effects eff
	in 
	{ eff with reads_alloc_table = true ; writes_alloc_table = true }
    | Array_creation ac -> 
	let eff = array_creation_effects eff ac in
	{ eff with reads_alloc_table = true ; writes_alloc_table = true }   
    | Static_interface _ -> eff
    | Static_class _ -> eff
    | Quantifier (_, _, _, _, _) -> failwith "effect(untyped quantifier)"
    | Instanceof (_, _) -> failwith "effect(untyped instanceof)"
    | Type (_) -> failwith "effect(untyped \\type)"
    | Typeof (_) -> failwith "effect(untyped \\typeof)"
    | Cast (_, _) -> failwith "effect(untyped cast)"
    | Array_access (_, _) -> failwith "effect(untyped array access)"
    | Assign_array (_, _, _, _) -> failwith "effect(untyped array assign)"
    | Class_instance_creation (_, _) -> failwith "effect(untyped new)"
    | Assign_name (_, _, _) -> failwith "effect(assign name)"
    | Var _ -> Typing.ill_typed e.expr_loc "effect(var)"; eff
    | This -> failwith "effect(this)"

and prepostincrdecr_effects eff e =
  match e.expr_node with
    | Typed_var lve -> eff
    | Field_access fa -> 
	begin
	  match fa with
	    | Typed_primary_access (e,ive) -> 
		if ive.instance_variable_entry_is_final 
		then assert false
		else
		  {eff with 
		     reads = Inttagset.add 
			       ive.instance_variable_entry_tag
			       ive eff.reads ;
		     writes = Inttagset.add 
			       ive.instance_variable_entry_tag
			       ive eff.writes ;
		     reads_heap = if ive.instance_variable_entry_is_static
		     then eff.reads_heap
		     else true;
		     writes_heap = if ive.instance_variable_entry_is_static
		     then eff.writes_heap
		     else true;
		     reads_alloc_table = match ive.instance_variable_entry_type with
		       | Integer_type | Boolean_type | Float_type -> 
			   eff.reads_alloc_table
		       | _ -> true 
	}
	    | Array_length _ -> assert false
	    | Typed_constant_access _ -> assert false
	    | Primary_access _ -> failwith "effect(prepost untyped primary access)"
	    | Super_access _ -> failwith "effect(prepost super access)"
	end
    | Typed_array_access (t, e1, e2) -> assert false (* TODO *)
    | Typed_method_call _ ->  assert false (* TODO *)
    | _ -> assert false


and array_creation_effects eff ac =
  match ac with
    | Explicit_array_creation(e,a) ->
	array_creation_effects (expr_effects eff e) a
    | Implicit_array_creation(t) -> 
	{ eff with writes_heap = true }
    | Typed_implicit_array_creation(t) ->
	{ eff with writes_heap = true }


and field_access_effects eff fa =
  match fa with
    | Typed_primary_access (e,ive) -> 
	if ive.instance_variable_entry_is_final 
	then eff
	else
	  {eff with 
	     reads = Inttagset.add 
		       ive.instance_variable_entry_tag
		       ive eff.reads ;
	     reads_heap = if ive.instance_variable_entry_is_static
	     then eff.reads_heap
	     else true;
          }
    | Array_length _ -> {eff with reads_heap = true; reads_alloc_table = true }
    | Typed_constant_access _ -> eff
    | Primary_access _ -> failwith "effect(untyped primary access)"
    | Super_access _ -> failwith "effect(super access)"
;;

let expr = expr_effects Java_context.default_effects;;

let var_init_effects eff vi =
    match vi with
      | Simple_initializer e -> expr_effects eff e
      | Array_initializer _ -> assert false
;;

let rec statement_effects eff i =
  match i.statement_node with
    | Skip -> eff
    | Block b -> block_effects eff b
    | Typed_expr (_, e) -> expr_effects eff e
    | Typed_var_decl (vd, e, b) -> 
	let eff1 = Option.fold eff (var_init_effects eff) e in
	statement_effects eff1 b
    | If_statement (e, i1, i2) ->
	let eff1 = expr_effects eff e in
	let eff2 = statement_effects eff1 i1 in
	statement_effects eff2 i2	
    | While (annot, e, i') ->
	let eff1 = expr_effects eff annot.loop_invariant in
	let eff2 = expr_effects eff1 e in
	let eff3 = statement_effects eff2 i' in
	{ eff3 with breaks = false } (* FAUX si label !! *)
    | Return e -> 
	Option.fold eff (expr_effects eff) e
    | Typed_return(ty,e) -> 
	Option.fold eff (expr_effects eff) e
    | Switch (e, cases) -> 
	let eff1 = expr_effects eff e in
	let eff2 = List.fold_left
		     (fun eff (_,b) -> block_effects eff b)
		     eff1
		     cases
	in
	{eff2 with breaks = false} (* FAUX si label !! *)
    | For_typed_decl (lve,init,annot,e,incr,body) ->
	let eff = expr_effects eff annot.loop_invariant in
	let eff = expr_effects eff e in
	let eff = block_effects eff incr in
	let eff = statement_effects eff body in
	{ eff with breaks = false } (* FAUX si label !! *)
    | Expr _ -> failwith "effect(expr)"
    | Synchronized (_, _) -> failwith "effect(synchronized)"
    | Try (_, _, _) -> failwith "effect(try)"
    | Typed_try(b,catches,bopt) ->
	let eff1 = block_effects eff b in
	let eff2 = List.fold_left (fun e (_,_,b) -> block_effects e b) eff1 catches in
	let eff3 = Option.fold eff2 (block_effects eff2) bopt in
	eff3
    | For_decl (_, _, _, _, _) -> failwith "effect(for decl)"
    | For(annot,init,e,step,body) -> 
	let eff1 = expr_effects eff annot.loop_invariant in
	let eff2 = block_effects eff1 init in
	let eff3 = expr_effects eff2 e in
	let eff4 = block_effects eff3 step in
	let eff5 = statement_effects eff4 body in
	{ eff5 with breaks = false } (* FAUX si label !! *)
    | Do (_, _, _) -> failwith "effect(do)"
    | Label (_, _) -> failwith "effect(label)"
    | Continue _ -> eff
    | Break _ -> { eff with breaks = true }
    | Throw e -> 
	let eff = expr_effects eff e in
	{ eff with throws = true}
    | Var_decl _ -> assert false
    | Jml_assert e -> expr_effects eff e
    | Annotated _ -> assert false
    | Typed_annotated (_,st) -> 
	statement_effects eff st
	  
and block_effects eff body = 
  match body with
    | [] -> eff
    | i::l ->
	let eff' = statement_effects eff i in
	 block_effects eff' l
;;  

let block = block_effects Java_context.default_effects ;;

let statement = statement_effects Java_context.default_effects ;;

let set_ref_effects t eff refs =
  let eff =
    match refs with
    | Set_array_index(e) -> 
	expr_effects {eff with writes_heap = true } e
    | Set_array_interval(e1,e2) -> 
	expr_effects (expr_effects {eff with writes_heap = true } e1) e2
    | Set_array -> 
	{eff with writes_heap = true } 
    | _ -> eff
  in
  match refs with
    | Set_array_index _  
    | Set_array_interval _ 
    | Set_array -> 
	begin
	  match t with
	    | Array_type Integer_type ->
		{eff with reads_int_array = true; writes_int_array = true }
	    | Array_type Boolean_type ->
		{eff with reads_bool_array = true; writes_bool_array = true }
	    | Array_type _ ->
		{eff with reads_obj_array = true; writes_obj_array = true }
	    | _ -> assert false
	end
    | Set_field(ive) ->
	{eff with 
	   writes = Inttagset.add ive.instance_variable_entry_tag
		      ive eff.writes ;
	   writes_heap =
             if ive.instance_variable_entry_is_static 
	     then eff.writes_heap
	     else true }
  | Set_fieldraw(id) ->
      ill_typed "set_ref_effects(name)"

let mod_effects eff mods =
  match fst mods with
    | Assignable_everything -> 
	{eff with writes_heap = true; writes_everything = true }
    | Assignable_nothing -> eff
    | Assignable_set_ref(e,refs) -> assert false
    | Typed_assignable_set_ref(t,e,refs) -> 
	let eff1 = expr_effects eff e in
	List.fold_left (set_ref_effects t) eff1 refs	  
    | Assignable_all_fields e -> 
	begin
	  let t =
	    match e.expr_node with
	      | Typed_var lve -> lve.local_variable_entry_type
	      | _ -> assert false
	  in
	  let eff = expr_effects eff e in
	  match t with
	    | Class_type ce ->
		List.fold_left
		(fun eff (_,f) ->
		   match f with
		     | Instance_variable_entry ive ->
			 {eff with 
			    reads = Inttagset.add 
				      ive.instance_variable_entry_tag 
				      ive eff.reads;
			    writes = Inttagset.add 
				       ive.instance_variable_entry_tag 
				       ive eff.writes; }
		     | _ -> eff)
		eff
		ce.class_entry_contents
	    | Array_type _ ->
		set_ref_effects t eff Set_array 
	    | _ -> 
		Format.eprintf "all_fields : %a@."
		Java_context.fprintf_java_type t;
		assert false
	end
;;

let spec_effects is_pure eff spec =
  match spec with
    | Typed_compact tcc -> 
	let eff0 = expr_effects eff tcc.typed_compact_requires in
	let eff1 = 
	  List.fold_left mod_effects eff0 tcc.typed_compact_assignable 
	in
	let eff2 = expr_effects eff1 tcc.typed_compact_ensures in
	let eff3 = 
	  match tcc.typed_compact_signals with
	    | None -> eff2
	    | Some(_,_,e) -> 
		expr_effects {eff2 with reads_alloc_table = true ; 
				throws = true} e
	in
	if is_pure && eff3.writes_heap then
	  Typing.ill_typed tcc.typed_compact_ensures.expr_loc 
	    "a pure method cannot modify the heap";
	if is_pure && not (Inttagset.is_empty eff3.writes) then
	  Typing.ill_typed tcc.typed_compact_ensures.expr_loc 
	    "a pure method cannot modify a variable";
	eff3
    | _ -> ill_typed "spec_effects(jml desugaring)"
;;




(*


  effects of methods and constructors

*)

let effects_changed = ref false;;

let not_same_effects e1 e2 =
  e1.reads_heap <> e2.reads_heap
  ||
  e1.reads_alloc_table <> e2.reads_alloc_table
  ||
  e1.writes_heap <> e2.writes_heap
  ||
  (not (Inttagset.equal e1.reads e2.reads))
  ||
  (not (Inttagset.equal e1.writes e2.writes))
  ||
  e1.reads_int_array <> e2.reads_int_array
  ||
  e1.writes_int_array <> e2.writes_int_array
  ||
  e1.reads_bool_array <> e2.reads_bool_array
  ||
  e1.writes_bool_array <> e2.writes_bool_array
  ||
  e1.reads_obj_array <> e2.reads_obj_array
  ||
  e1.writes_obj_array <> e2.writes_obj_array
  ||
  e1.throws <> e2.throws
;;

let report_spec_effects_error name msg se be =
  Format.printf 
    "Warning: effects of %s do not match its specification: %s@." 
	name msg;
  Format.fprintf Config.log 
    "Warning: effects of %s do not match its specification: %s@." 
    name msg;
  Format.fprintf Config.log "Effects of its spec: %a@."
    Java_context.fprintf_effects se;
  Format.fprintf Config.log "Effects of its body: %a@."
    Java_context.fprintf_effects be;
(*
  ill_typed "incorrect spec effects"
*)
;;
 
let verif_spec_effects name routine se (* spec effects *) be (* body effects *) =
  if se.writes_everything
  then 
    Format.fprintf Config.log 
      "Effects of %s not verified because specified as modifying everything.@." 
      name 
  else
    begin
      if se.writes_int_array <> be.writes_int_array
      then 
	begin
	  report_spec_effects_error name "on modifying some int[]" se be;
	  routine.routine_entry_int_array_writes_nothing <- true;
	end;
      if se.writes_bool_array <> be.writes_bool_array
      then 
	begin
	  report_spec_effects_error name "on modifying some boolean[]" se be;
	  routine.routine_entry_bool_array_writes_nothing <- true;
	end;
      if se.writes_obj_array <> be.writes_obj_array
      then 
	begin      
	  report_spec_effects_error name "on modifying some Object[]" se be;
	  routine.routine_entry_obj_array_writes_nothing <- true;
	end;
      if not (Inttagset.equal se.writes be.writes)
      then 
	begin
	  let fields = Inttagset.diff be.writes se.writes in
	  let msg = 
	    Inttagset.fold
	      (fun ive acc -> 	     
		 acc ^ " " ^ ive.instance_variable_entry_name)
	      fields
	      "on modifying field(s)"
	  in	
	  report_spec_effects_error name msg se be;
	  routine.routine_entry_fields_writes_nothing <- fields
	end;
      if se.throws <> be.throws
      then 
	report_spec_effects_error name "on expection throwing" se be
    end
;;  


let param_effects eff routine =
  List.fold_left
    (fun eff (_,x) ->
       match x with
	 | Local_variable_entry lve ->
	     begin
	       match lve.local_variable_entry_type with
		 | Class_type ce ->
		     effects_union
		     { eff with reads_heap = true }
		     ce.class_entry_invariant_effects
		 | Interface_type _ 
		 | Array_type _ -> 
		     { eff with reads_heap = true }
		 | _ -> eff
	     end
	 | _ -> assert false)
    eff
    routine.routine_entry_parameters
;;


let default_effects = Java_context.default_effects
	     
let method_effects verif_only me ms body=
  Format.fprintf Config.log "finding effects for method %s.@."
    me.method_entry_name;
  let eff1 = 
    if me.method_entry_is_static then default_effects else
      effects_union
	{default_effects with reads_heap = true; reads_alloc_table = true }  (* reads this *)
	(match me.method_entry_class_or_interface with
	   | Class_entry c -> c.class_entry_invariant_effects
	   | _ -> assert false)
  in
  let routine = me.method_entry_routine in
  let param_e = param_effects eff1 routine in 
  let old_e = routine.routine_entry_effects in
  if routine.routine_entry_produce_body 
  then
    begin
      match body with
	| None -> 
	    let new_e =
	      spec_effects me.method_entry_is_pure 
		(effects_union old_e param_e) ms
	    in
	    if not_same_effects new_e old_e
	    then
	      begin
		effects_changed := true;
		Format.fprintf Config.log "Effects for %s changed.@."
		  me.method_entry_name;
		routine.routine_entry_effects <- new_e;
	      end
	| Some block -> 
	    if verif_only then
	      begin
		let spec_eff = 
		  spec_effects me.method_entry_is_pure param_e ms
		in
		verif_spec_effects 
		  me.method_entry_name 
		  me.method_entry_routine spec_eff old_e;
		end
	    else
	      begin
		let new_e = block_effects 
			      (effects_union old_e param_e) block in
		if not_same_effects new_e old_e
		then
		  begin
		    effects_changed := true;
		    Format.fprintf Config.log "Effects for %s changed.@."
		      me.method_entry_name;
		    if new_e.throws then 
		      Format.fprintf Config.log 
			"In particular, it may throw an exception.@." ;
		    routine.routine_entry_effects <- new_e;
		  end
	      end
    end
  else
    let spec_eff = 
      spec_effects me.method_entry_is_pure 
	(effects_union old_e param_e) ms 
    in
    if not_same_effects spec_eff old_e
    then
      begin
	effects_changed := true;
	Format.fprintf Config.log "Effects for %s changed.@."
	  me.method_entry_name;
	routine.routine_entry_effects <- spec_eff;
      end
;;
  
let constructor_effects verif_only ce ms eci body = 
  Format.fprintf Config.log "finding effects for constructor %s.@."
    ce.constructor_entry_trans_name;
  let eff1 = 
    effects_union
      {default_effects with reads_heap = true; 
	 writes_heap = true;
	 reads_alloc_table = true;
	 writes_alloc_table = true; } (* writes this *)
      ce.constructor_entry_class.class_entry_invariant_effects
  in
  let routine = ce.constructor_entry_routine in
  let param_e = param_effects eff1 routine in 
  let old_e = routine.routine_entry_effects in
  if routine.routine_entry_produce_body
  then
    begin
      if verif_only then
	begin
	  let spec_eff = 
	    spec_effects false (* ce.constructor_entry_is_pure *) 
	      param_e ms
	  in
	  verif_spec_effects 
	    ce.constructor_entry_trans_name ce.constructor_entry_routine 
	    spec_eff old_e
	end
      else
	let new_e = block_effects (effects_union old_e param_e) body in
	if not_same_effects new_e old_e
	then
	  begin
	    effects_changed := true;
	    routine.routine_entry_effects <- new_e;
	  end 
    end
  else
    begin
      let spec_eff = 
	spec_effects false (* ce.constructor_entry_is_pure *) 
	  (effects_union old_e param_e) ms 
      in
      if not_same_effects spec_eff old_e
      then
	begin
	  effects_changed := true;
	Format.fprintf Config.log "Effects for %s changed.@."
	  ce.constructor_entry_trans_name;
	  routine.routine_entry_effects <- spec_eff;
	end
    end
;;

let field_effects verif_only ce f =
  match f with
    | Method _ -> assert false
    | Typed_method (me,ms,body) -> 
	method_effects verif_only me ms body
    | Constructor _ -> assert false
    | Typed_constructor(ce,ms,eci,body) ->
	constructor_effects verif_only ce ms eci body
    | Variable _  -> assert false
    | Typed_variable (ive,_) ->
	begin	  
	  match ive.instance_variable_entry_type with
	    | Integer_type 
	    | Boolean_type
	    | Float_type -> ()
	    | Class_type _ 
	    | Interface_type _ 
	    | Array_type _ ->
		if ive.instance_variable_entry_is_final
		then ()
		else
		let old_eff = ce.class_entry_invariant_effects in
		let new_eff =
		  { old_eff  with
		      reads = 
		      Inttagset.add 
			ive.instance_variable_entry_tag ive old_eff.reads;
		      reads_alloc_table = true} 
		in
		if not_same_effects old_eff new_eff 
		then
		  begin
		    effects_changed := true;
		    Format.fprintf Config.log "Effects for class invariant of %s changed.@."
		      ce.class_entry_name;
		    ce.class_entry_invariant_effects <- new_eff
		  end
	    | Null_type -> assert false
	    | Type_type -> assert false
	end
    | Static_initializer _ -> ()
    | Specification s -> 
	List.iter
	  (function
	     | Invariant a ->
		 let old_eff = ce.class_entry_invariant_effects in
		 let new_eff = expr_effects old_eff a in
		 if not_same_effects old_eff new_eff 
		 then
		   begin
		     effects_changed := true;
		     Format.fprintf Config.log "Effects for class invariant of %s changed.@."
		       ce.class_entry_name;
		     ce.class_entry_invariant_effects <- new_eff
		   end		 
	     | _ -> ())
	  s
;;

let member_effects verif_only ie m =  match m with
    | Constant_declaration _ 
    | Typed_constant_declaration _ -> ()
    | Method_header _ -> assert false
    | Typed_method_header(me,ms) -> ()
	
;;

let type_decl_effects verif_only env td =
  match td with
    | Class _ -> assert false
    | Interface _ -> assert false
    | Typed_class(ce,cd) ->
	List.iter 
	  (fun (_,f) -> field_effects verif_only ce f) 
	  cd.class_fields
    | Typed_interface(ie,id) ->
	List.iter 
	  (fun (_,f) -> member_effects verif_only ie f) 
	  id.interface_members
;;

let comp_unit_effects verif_only env cu =
  List.iter 
    (fun (_,td) -> type_decl_effects verif_only env td) 
    cu.cu_type_decls
;;


let rec compute env comp_units =
  effects_changed := false;
  List.iter (comp_unit_effects false env) comp_units;
  if !effects_changed then 
    begin
      Format.fprintf Config.log 
	"Fixpoint not reached in computation of effects, running again.@.";
      compute env comp_units
    end
  else
    begin
      Format.fprintf Config.log 
	"Fixpoint reached in computation of effects, verifying effects of specs.@.";
      List.iter (comp_unit_effects true env) comp_units;
    end
;;

    
let compute_translated_type ive =
  if ive.instance_variable_entry_is_static
  then
    let t = match ive.instance_variable_entry_type with
      | Integer_type -> "int"
      | Boolean_type -> "bool"
      | Float_type -> "float"
      | Null_type -> assert false
      | Type_type -> assert false
      | Class_type _ -> "value"
      | Array_type _ -> "value"
      | Interface_type _ -> "value"
    in Why.Base_type([],t)
  else
    let t = match ive.instance_variable_entry_type with
      | Integer_type -> "int"
      | Boolean_type -> "bool"
      | Float_type -> "float"
      | Null_type -> assert false
      | Type_type -> assert false
      | Class_type _ -> "value"
      | Array_type _ -> "value"
      | Interface_type _ -> "value"
    in Why.Base_type([Why.base_type t],"memory")
;;



let params_reads eff params = 
  let params = 
    if eff.reads_obj_array then  
      ("objA",Why.Base_type([Why.base_type "value"],"memarray"))::params
    else params
  in
  let params = 
    if eff.reads_bool_array then  
      ("boolA",Why.Base_type([Why.bool_type],"memarray"))::params
    else params
  in
  let params = 
    if eff.reads_int_array then  
      ("intA",Why.Base_type([Why.int_type],"memarray"))::params
    else params
  in
  let params =
    Inttagset.fold
      (fun ive params ->
	 let t = compute_translated_type ive in
	 (ive.instance_variable_entry_trans_name,t) :: params)
      eff.reads
      params
  in
  let params = 
    if eff.reads_alloc_table then  
      ("heap",Why.Base_type([],"store"))::params
    else params
  in
  params
;;

