
(** Krakatoa project - Class structure *)

Require Import preambule.

(** * Java types parameterized by the class classId *)

Inductive javaT (classId:Set) : Set :=
  | IntT : javaT classId
  | BoolT : javaT classId
  | ClassT : classId -> javaT classId
  | ArrayT : javaT classId -> javaT classId.

(** * The class structure defined as a module type *)

Module Type ClassStructure.

(** Class identifiers *)
Parameter classId : Set.

(** Any class structure needs an Object class *)
Parameter ObjectClass : classId.

(** Decidable equality on class identifiers *)
Axiom eq_classId_dec : forall i1 i2:classId, {i1 = i2} + {i1 <> i2}.

(** [super] associates to a class, its super-class, 
    it is defined as a partial function *)
Parameter super : classId -> option classId.

(** [ObjectClass] should not have a super-class *)
Axiom superObject : super ObjectClass = None.

(** The relation associated to [super] is well-founded : 
    following super-links is always finite *)
Axiom super_wf : well_founded (fun i1 i2 => def (super i2) i1).

(** Names of fields *)
Parameter fieldId : Set.

(** Decidable equality on field identifiers *)
Axiom eq_fieldId_dec : forall f1 f2:fieldId, {f1 = f2} + {f1 <> f2}.

(** Each field should be declared in one class *)
Parameter defined_in : fieldId -> classId.

(** No field defined in class Object *)
Axiom no_field_in_Object : forall f:fieldId, defined_in f <> ObjectClass.

(** Each field has a type *)
Parameter fieldT : fieldId -> javaT classId.

End ClassStructure.

(** * Definition of properties of classes 
      We use a functor parameterized by the class structure *)

Module ClassGeneric (C: ClassStructure).
Import C.

(** ** Definition of [specializes]

  [specializes] is the reflexive transitive closure of the 
  [super] relation : [(specializes i2 i1)] is true when [i2] is
  reachable following [super] links from [i1]

*)

Inductive specializes (i2:classId) : classId -> Prop :=
  | Specialize_refl : specializes i2 i2
  | Specialize_super :
      forall i1 i3:classId,
        def (super i1) i3 -> specializes i2 i3 -> specializes i2 i1.
Hint Constructors specializes.

(** ** Properties of [specializes] *)

Lemma specializes_eq : forall i1 i2:classId, i1 = i2 -> specializes i1 i2.
Proof.
simple destruct 1; auto.
Qed.
Hint Resolve specializes_eq.

Lemma specializes_trans :
 forall i1 i2 i3:classId,
   specializes i1 i2 -> specializes i2 i3 -> specializes i1 i3.
Proof.
simple induction 2; auto.
intros.
apply Specialize_super with i4; trivial.
Qed.

Lemma specializes_Object :
 forall i:classId, specializes i ObjectClass -> i = ObjectClass.
Proof.
intros.
inversion H; trivial.
rewrite superObject in H0.
discriminate.
Qed.

(** ** Definition of [extends] as the symmetric of [specializes] *)

Definition extends (i1 i2:classId) := specializes i2 i1.
Hint Unfold extends.

Lemma extends_eq : forall i1 i2:classId, i1 = i2 -> extends i1 i2.
Proof.
auto.
Qed.
Hint Resolve extends_eq.

Lemma extends_trans :
 forall i1 i2 i3:classId, extends i1 i2 -> extends i2 i3 -> extends i1 i3.
Proof.
unfold extends in |- *; intros; apply specializes_trans with i2; trivial.
Qed.

Lemma extends_Object :
 forall i:classId, extends ObjectClass i -> i = ObjectClass.
Proof.
unfold extends in |- *.
apply specializes_Object.
Qed.

Lemma extends_comparable :
 forall i1 i2 i3:classId,
   extends i1 i2 -> extends i1 i3 -> extends i2 i3 \/ extends i3 i2.
intros i1 i2 i3 H; generalize i3; unfold extends in |- *; elim H;
 clear i1 i3 H; intros; auto.
inversion H2.
right; apply Specialize_super with i3; trivial.
assert (i3 = i5); auto.
apply def_fun with (super i1); auto.
apply (H1 i0).
rewrite H6; trivial.
Qed.

(** ** Property *)
(**  if [i2] can be reached from [i1] and is not [i1] 
       then it can be reached from [(super i1)] *)

Definition specializes_super_inv :
  forall i2 i1 i:classId,
    specializes i2 i1 -> i2 <> i1 -> def (super i1) i -> specializes i2 i.
Proof.
simple destruct 1; intros.
case H0; trivial.
replace i with i3; trivial.
apply def_fun with (super i0); trivial.
Defined.

(** ** Decidability of the [specializes] and [extends] relation *)
(** Defined by well-founded induction *)

Definition specializes_dec :
  forall i2 i1:classId, {specializes i2 i1} + {~ specializes i2 i1}.
intros; pattern i1 in |- *; apply (Fix super_wf); clear i1.
intros i1 srec; case (eq_classId_dec i2 i1); auto.
intro; case (is_def (super i1)); intros.
case (srec y); intros; auto.
left; apply Specialize_super with y; auto.
right; red in |- *; intro; apply n0; apply specializes_super_inv with i1;
 auto.
right; red in |- *; intro; apply n.
inversion H; trivial.
case undef_not_def with (2 := H0); trivial.
Defined.

Definition extends_dec (i1 i2:classId) :
  {extends i1 i2} + {~ extends i1 i2} := specializes_dec i2 i1.


Definition specializes_bool (i2 i1:classId) : bool :=
  dec_bool (specializes_dec i2 i1).

Definition extends_bool (i1 i2:classId) : bool := specializes_bool i2 i1.


(** ** Properties of the [specializes_bool] and [extends_bool] functions *)
Lemma specializes_Prop_to_bool :
 forall i2 i1:classId, specializes i2 i1 -> specializes_bool i2 i1 = true.
Proof fun i2 i1 => dec_bool_true_intro (specializes_dec i2 i1).

Lemma extends_Prop_to_bool :
 forall i1 i2:classId, extends i1 i2 -> extends_bool i1 i2 = true.
intros i1 i2; exact (specializes_Prop_to_bool i2 i1).
Qed.

Lemma specializes_bool_to_prop :
 forall i2 i1:classId, specializes_bool i2 i1 = true -> specializes i2 i1.
Proof fun i2 i1 => dec_bool_true_elim (specializes_dec i2 i1).

Lemma extends_bool_to_Prop :
 forall i1 i2:classId, extends_bool i1 i2 = true -> extends i1 i2.
intros i1 i2; exact (specializes_bool_to_prop i2 i1).
Qed.

Definition exists_in (f:fieldId) (c:classId) : Prop :=
  extends c (defined_in f).
Definition exists_in_bool (f:fieldId) (c:classId) : bool :=
  extends_bool c (defined_in f).

Definition javaType := javaT classId.
Definition IntType : javaType := IntT classId.
Definition BoolType : javaType := BoolT classId.
Definition ClassType : classId -> javaType := ClassT classId.
Definition ArrayType : javaType -> javaType := ArrayT classId.
Definition fieldType : fieldId -> javaType := fieldT.

Lemma eq_javaType_dec : forall t1 t2:javaType, {t1 = t2} + {t1 <> t2}.
decide equality.
apply eq_classId_dec.
Defined.

Fixpoint subType (t1 t2:javaType) {struct t2} : Prop :=
  match t1, t2 with
  | IntT, IntT => True
  | BoolT, BoolT => True
  | ClassT c1, ClassT c2 => extends c1 c2
  | ArrayT t1, ArrayT t2 => subType t1 t2
  | ArrayT t, ClassT c => c = ObjectClass
  | _, _ => False
  end.

Lemma subType_refl : forall t:javaType, subType t t.
simple induction t; simpl in |- *; auto.
Qed.

Lemma subType_trans :
 forall t1 t2 t3:javaType, subType t1 t2 -> subType t2 t3 -> subType t1 t3.
simple induction t1; simple destruct t2; simple destruct t3; simpl in |- *;
 intros; auto; try contradiction.
apply extends_trans with c0; trivial.
subst c; apply extends_Object; assumption.
apply H with j0; trivial.
Qed.

Lemma subType_int : forall t:javaType, subType t IntType -> t = IntType.
simple destruct t; simpl in |- *; intuition.
Qed.

Lemma subType_bool : forall t:javaType, subType t BoolType -> t = BoolType.
simple destruct t; simpl in |- *; intuition.
Qed.

Lemma subType_dec :
 forall t1 t2:javaType, {subType t1 t2} + {~ subType t1 t2}.
intros t1 t2; generalize t1; clear t1; elim t2; simple destruct t1;
 simpl in |- *; intros; auto.
apply extends_dec.
apply eq_classId_dec.
Defined.

End ClassGeneric.

(*
Hint extends_hint := Constructors specializes.
Hints Resolve specializes_eq extends_eq.
*)