
(** Krakatoa project - Model of objects using adresses in the heap *)

Require Import preambule.
Require Import local_classes.
Require Import ZArith.
Require Import map.
Require Import fmap.
Open Scope Z_scope.

(** * Destructive model of objects *)
Set Implicit Arguments.
Unset Strict Implicit.
Module ObjectAdr (C: local_classes.ClassStructure).

Import C.
Module CGen := ClassGeneric C.
Import CGen.

Definition adrObject : Set :=Z.

Definition eq_adrObject_dec : forall a1 a2:adrObject, {a1 = a2} + {a1 <> a2}
  := Z_eq_dec.

Inductive kind : Set :=
  | KInt : kind
  | KBool : kind
  | KObj : javaType -> kind.
    
Definition eq_Arrkind_dec : forall v1 v2:kind, {v1 = v2} + {v1 <> v2}.
decide equality.
apply eq_javaType_dec.
Defined.

Inductive tag : Set :=
  | Obj : classId -> tag
  | Arr : N -> kind -> tag.

Definition ArrInt (n:Z) := Arr (BinInt.Zabs_N n) KInt.
Definition ArrBool (n:Z) := Arr (BinInt.Zabs_N n) KBool.
Definition ArrObj (n:Z) (t:javaType) := Arr (BinInt.Zabs_N n) (KObj t).

Definition eq_tag_dec : forall v1 v2:tag, {v1 = v2} + {v1 <> v2}.
decide equality.
apply eq_classId_dec.
apply eq_Arrkind_dec.
apply entier_eq_dec.
Defined.

Inductive value : Set :=
  | Null : value
  | Ref : adrObject -> value.

Definition neqv (v1 v2:value) := v1 <> v2.
Notation "x != y" := (neqv x y) (at level 70, no associativity).
Hint Unfold neqv.

Lemma neqv_sym : forall v1 v2:value, v1 != v2 -> v2 != v1.
red in |- *; intros.
apply sym_not_eq; trivial.
Qed.
Hint Immediate neqv_sym.

Definition eq_value_dec : forall v1 v2:value, {v1 = v2} + {v1 != v2}.
unfold neqv in |- *; decide equality.
apply eq_adrObject_dec.
Defined.

Definition eq_value :
  forall x y:value, {b : bool | if b then x = y else x != y}.
Proof.
intros.
case (eq_value_dec x y); intros.
exists true; auto.
exists false; auto.
Qed.

Definition ne_value :
  forall x y:value, {b : bool | if b then x != y else x = y}.
Proof.
intros.
case (eq_value_dec x y); intros.
exists false; auto.
exists true; auto.
Qed.

(** Java values of type int, byte will be represented in Z, 
    bool corresponds to bool, objects and arrays 
    are represented by object of type value *)

(** Table for storing allocated adresses with their type *)

Definition store := t tag.

Definition tag_of (h:store) (v:value) : option tag :=
  match v with
  | Null => None
  | Ref a => find h a
  end.

Definition has_tag (h:store) (v:value) (t:tag) : Prop :=
  match v with
  | Null => False
  | Ref a => contains h a t
  end.

Lemma has_tag_eq :
 forall (h:store) (v:value) (t:tag), has_tag h v t -> tag_of h v = Some t.
simple destruct v; simpl in |- *; intuition.
Qed.

(** ** Typing *)

Definition typeof (h:store) (v:value) (t:javaType) : Prop :=
  match tag_of h v with
  | Some (Obj c1) =>
      match t with
      | ClassT c2 => c1=c2
      | _ => False
      end
  | Some (Arr n ak) =>
      match t, ak with
      | ArrBoolT, KBool => True
      | ArrIntT, KInt => True
      | ArrayT t2, KObj t1 => t1=t2
      | _, _ => False
      end
  | None => v = Null
  end.

Definition instanceof (h:store) (v:value) (t:javaType) : Prop :=
    exists t1 , (typeof h v t1) /\ (subType t1 t).

Lemma typeof_subtype_instanceof : forall (h:store) (v:value) (t1 t2:javaType),
  (typeof h v t1) -> (subType t1 t2) -> (instanceof h v t2).
intros; exists t1; auto.
Save.

Lemma typeof_instanceof : forall (h:store) (v:value) (t:javaType),
  (typeof h v t) -> (instanceof h v t).
intros; exists t; auto.
Save.
Hint Resolve typeof_instanceof.

Lemma typeof_null : forall (h:store) (t:javaType), typeof h Null t.
Proof.
red in |- *; simpl in |- *; trivial.
Qed.
Hint Resolve typeof_null.

Lemma instanceof_null : forall (h:store) (t:javaType), instanceof h Null t.
auto.
Save.
Hint Resolve instanceof_null.

Lemma type_obj :
 forall (h:store) (c:classId) (v:value),
   has_tag h v (Obj c) -> typeof h v (ClassType c).
Proof.
red in |- *; intros.
rewrite (has_tag_eq H); simpl; auto.
Qed.

Lemma type_arr :
 forall (h:store) (l:Z) (v:value) (t:javaType),
   has_tag h v (ArrObj l t) ->
   typeof h v (ArrayType t).
Proof.
red in |- *; intros.
rewrite (has_tag_eq H); simpl in |- *; auto.
Qed.

Lemma type_arr_int :
 forall (h:store) (l:Z) (v:value),
   has_tag h v (ArrInt l) -> typeof h v ArrIntType.
Proof.
red in |- *; intros.
rewrite (has_tag_eq H); simpl in |- *; auto.
Qed.

Lemma type_arr_bool :
 forall (h:store) (l:Z) (v:value),
   has_tag h v (ArrBool l) -> typeof h v ArrBoolType.
Proof.
red in |- *; intros.
rewrite (has_tag_eq H); simpl in |- *; auto.
Qed.


Hint Resolve type_obj type_arr type_arr_int type_arr_bool. 


Lemma instanceof_subType_trans :
 forall (h:store) (this:value) (t1:javaType),
   instanceof h this t1 ->
   forall t2:javaType, subType t1 t2 -> instanceof h this t2.
destruct 1.
exists x; intuition.
apply subType_trans with t1; auto.
Save.

(* Arbitrary value for the type of Null *)
Definition NullType := ArrIntType.

Definition typeof_dec :
  forall (h:store) (v:value),
    {t : javaType| typeof h v t} + {forall t, ~ typeof h v t}.
unfold typeof; intros.
case (tag_of h v); intuition.
case t; simpl; intuition.
left; exists (ClassType c); simpl; auto.
case k; intuition.
left; exists ArrIntType; simpl; auto.
left; exists ArrBoolType; simpl; auto.
left; exists (ArrayType j); simpl; auto.
case (eq_value_dec v Null).
left; exists NullType; trivial.
auto.
Defined.

Lemma typeof_unique : 
  forall (h:store) (v:value) (t1 t2: javaType),
    v!=Null -> (typeof h v t1) -> (typeof h v t2) -> t1 = t2.
unfold typeof; intros h v t1 t2 H.
case (tag_of h v); try contradiction; auto.
destruct t; destruct t1; destruct t2; 
try destruct k; try contradiction; auto.
intros; rewrite <- H1; rewrite <- H0; trivial.
intros; rewrite <- H1; rewrite <- H0; trivial.
Save.


Definition instanceof_dec :
  forall (h:store) (v:value) (t:javaType),
    {instanceof h v t} + {~ instanceof h v t}.
unfold instanceof; intros.
case (typeof_dec h v).
intros (t1,H1).
case (subType_dec t1 t); intros.
left; exists t1; auto.
case (eq_value_dec v Null).
left; exists t; simpl; intuition. 
rewrite e; auto.
intros; right; red; intros.
case H; intros t2 (H2,H3).
apply n; rewrite (@typeof_unique h v t1 t2); auto.
right; red; intros.
case H; intros t2 (H2,H3).
apply (n t2); auto.
Defined.

(** Array length *)
Definition arraylength (h:store) (v:value) : Z :=
  match tag_of h v with
  | Some (Arr n _) => BinInt.Z_of_N n
  | _ => 0
  end.

Definition arrayelemtype (h:store) (v:value) : javaType :=
  match tag_of h v with
  | Some (Arr n (KObj t)) => t
  | _ => ClassType ObjectClass
  end.

Lemma typeof_elemtype : 
forall (h:store) (v:value) (t:javaType),
       v !=Null -> (typeof h v (ArrayType t)) -> (arrayelemtype h v)=t.
Proof.
unfold typeof, arrayelemtype; intros h v t.
case (tag_of h v); try contradiction; auto.
destruct t0; simpl; try contradiction; auto.
destruct k; simpl; try contradiction; auto.
Save.

(** * Alive values *)

Definition alive (t:store) (v:value) :=
  match v with
  | Null => True
  | Ref a => find t a <> None
  end.

Definition fresh (t:store) (v:value) := ~ alive t v.

Lemma alive_fresh_diff :
 forall (t:store) (v1 v2:value), alive t v1 -> fresh t v2 -> v1 != v2.
repeat red in |- *; intros.
apply H0; subst; trivial.
Qed.

Lemma fresh_alive_diff :
 forall (t:store) (v1 v2:value), fresh t v1 -> alive t v2 -> v1 != v2.
repeat red in |- *; intros.
apply H; subst; trivial.
Qed.

Definition store_extends (t t':store) :=
  forall v:value, alive t v -> tag_of t v = tag_of t' v.

Lemma alive_extends_stable :
 forall t t':store,
   store_extends t t' -> forall v:value, alive t v -> alive t' v.
intros.
generalize (H v H0) H0; case v; simpl in |- *; intros; auto.
rewrite <- H1; trivial.
Qed.

Lemma fresh_extends_stable :
 forall t t':store,
   store_extends t t' -> forall v:value, fresh t' v -> fresh t v.
unfold fresh in |- *; red in |- *; intros.
apply H0; apply alive_extends_stable with t; trivial.
Qed.

Lemma store_extends_refl : forall t:store, store_extends t t.
Proof.
red in |- *; auto.
Qed.

Lemma store_extends_trans :
 forall t1 t2 t3:store,
   store_extends t1 t2 -> store_extends t2 t3 -> store_extends t1 t3.
Proof.
unfold store_extends in |- *; intros.
transitivity (tag_of t2 v); auto.
apply H0; apply alive_extends_stable with t1; auto.
Qed.

Hint Resolve store_extends_refl.

Lemma tag_of_fresh :
 forall (h:store) (v:value), fresh h v -> tag_of h v = None.
Proof.
intros h v; unfold fresh, alive, tag_of in |- *.
case v; simpl in |- *; auto.
intro a; case (find h a); intuition.
Qed.


Lemma null_alive : forall (h:store) (v:value), v = Null -> alive h v.
Proof.
intros h v vn; rewrite vn; simpl in |- *; auto.
Qed.

Hint Resolve null_alive.
   
Lemma fresh_not_null : forall (h:store) (v:value), fresh h v -> v != Null.
Proof.
repeat red in |- *; intros.
apply H; auto.
Qed.

Definition alive_dec :
  forall (t:store) (v:value), {alive t v} + {~ alive t v}.
Proof.
simple destruct v; simpl in |- *; auto.
intro a; case (find t a); auto.
Defined.


Lemma typeof_alive :
 forall (h:store) (t:javaType) (v:value), typeof h v t -> alive h v.
Proof.
simple destruct v; unfold typeof, alive in |- *; simpl in |- *; auto.
red in |- *; intro a; case (find h a); try (intros; discriminate); case t;
 tauto.
Qed.

Lemma instanceof_alive :
 forall (h:store) (t:javaType) (v:value), instanceof h v t -> alive h v.
Proof.
destruct 1; intuition.
apply typeof_alive with x; trivial.
Qed.




(*
Definition is_valid_update (h:store) (this v:value) 
  (t:javaType) := this != Null /\ instanceof h v t.

Hint Unfold is_valid_update.
*)

Definition is_valid_index (h:store) (this:value) (z:Z) :=
  0 <= z < arraylength h this.

Lemma is_valid_index_alive :
 forall (h:store) (this:value) (z:Z), is_valid_index h this z -> alive h this.
unfold is_valid_index, arraylength in |- *; simple destruct this;
 simpl in |- *; auto.
intros a z; case (find h a); intros; auto with zarith.
Qed.

Definition is_valid_array_update (h:store) (this:value) 
  (z:Z) (v:value) :=
  is_valid_index h this z /\ instanceof h v (arrayelemtype h this).

Lemma is_valid_array_alive :
 forall (h:store) (this:value) (z:Z) (v:value),
   is_valid_array_update h this z v -> alive h this.
simple destruct 1; intros; apply is_valid_index_alive with z; auto.
Qed.

Definition mod_loc := value -> Prop.
Definition array_mod_loc := value -> Z -> Prop.

(** Each field is represented as a function
    from adresses to values or primitive objects *)

Section Memory.

Variable A : Set.

Definition memory := map.t value A.

Definition memory_init (def:A) : memory := map.cte value def.

Definition acc (m:memory) (v:value) : A := map.find m v.

Definition update (m:memory) (v:value) (w:A) : memory :=
  map.add eq_value_dec m v w.

Lemma acc_update :
 forall (m:memory) (v:value) (w:A), acc (update m v w) v = w.
Proof.
intros; unfold acc, update in |- *; apply map.find_add; trivial.
Qed.

Lemma acc_update_eq :
 forall (m:memory) (v1 v2:value),
   v1 = v2 -> forall w:A, acc (update m v1 w) v2 = w.
simple destruct 1; intros; apply acc_update; trivial.
Qed.

Lemma acc_update_other :
 forall (m:memory) (v1 v2:value),
   v1 != v2 -> forall w:A, acc (update m v1 w) v2 = acc m v2.
intros; unfold acc, update in |- *; apply map.find_add_other; auto.
Qed.

Definition modifiable (h:store) (m m':memory) (unchanged:mod_loc) : Prop :=
  forall v:value, alive h v -> unchanged v -> acc m v = acc m' v.

Lemma modifiable_refl :
 forall (h:store) (m:memory) (unchanged:mod_loc), modifiable h m m unchanged.
red in |- *; auto.
Qed.

Lemma modifiable_trans :
 forall (h:store) (m1 m2 m3:memory) (unch:mod_loc),
   modifiable h m1 m2 unch ->
   modifiable h m2 m3 unch -> modifiable h m1 m3 unch.
unfold modifiable in |- *; intuition.
transitivity (acc m2 v); auto.
Qed.

Lemma modifiable_sub :
 forall (h1 h2:store) (m1 m2:memory) (unch1 unch2:mod_loc),
   (forall v:value, alive h1 v -> unch1 v -> unch2 v) ->
   store_extends h1 h2 ->
   modifiable h2 m1 m2 unch2 -> modifiable h1 m1 m2 unch1.
unfold store_extends, modifiable in |- *; intros.
assert (alive h2 v).
apply alive_extends_stable with h1; trivial.
firstorder.
Qed.

Lemma modifiable_trans_sub :
 forall (h1 h2:store) (m1 m2 m3:memory) (unch1 unch2:mod_loc),
   (forall v:value, alive h1 v -> unch1 v -> unch2 v) ->
   store_extends h1 h2 ->
   modifiable h1 m1 m2 unch1 ->
   modifiable h2 m2 m3 unch2 -> modifiable h1 m1 m3 unch1.
intros; apply modifiable_trans with m2; trivial.
apply modifiable_sub with h2 unch2; trivial.
Qed.

Lemma modifiable_update :
 forall (h:store) (m:memory) (unch:mod_loc) (v:value) 
   (a:A), ~ unch v -> modifiable h m (update m v a) unch.
red in |- *; intuition.
rewrite acc_update_other; auto.
red in |- *; intro; subst; intuition.
Qed.

Lemma modifiable_fresh_update :
 forall (h:store) (m1 m2:memory) (unch:mod_loc) (v:value) 
   (a:A),
   fresh h v ->
   modifiable h m1 m2 unch -> modifiable h m1 (update m2 v a) unch.
unfold modifiable in |- *; intuition.
rewrite acc_update_other; auto.
apply fresh_alive_diff with h; auto.
Qed.

End Memory.
Hint Resolve modifiable_refl modifiable_update.

(** Memory for arrays *)

Section Memarray.

Variable A : Set.

Definition memarray := memory (map.t Z A).

Definition array_update (m:memarray) (v:value) (n:Z) 
  (w:A) : memarray := update m v (map.add Z_eq_dec (acc m v) n w).

Definition array_acc (m:memarray) (v:value) (n:Z) : A :=
  map.find (acc m v) n.

Lemma array_acc_same :
 forall (m:memarray) (v:value) (n1 n2:Z),
   n1 = n2 -> array_acc m v n1 = array_acc m v n2.
simple destruct 1; trivial.
Qed.

Lemma array_acc_update :
 forall (m:memarray) (v:value) (n:Z) (w:A),
   array_acc (array_update m v n w) v n = w.
Proof.
unfold array_acc, array_update in |- *; intros.
rewrite acc_update; trivial.
apply map.find_add.
Qed.

Lemma array_acc_update_eq :
 forall (m:memarray) (v1 v2:value) (n1 n2:Z) (w:A),
   v1 = v2 -> n1 = n2 -> array_acc (array_update m v1 n1 w) v2 n2 = w.
Proof.
intros; subst v1 n1; apply array_acc_update; trivial.
Qed.

Lemma array_acc_update_other :
 forall (m:memarray) (v1 v2:value) (n1 n2:Z) (w:A),
   v1 != v2 \/ n1 <> n2 ->
   array_acc (array_update m v1 n1 w) v2 n2 = array_acc m v2 n2.
Proof.
unfold array_acc, array_update in |- *; intros.
case (eq_value_dec v1 v2); intro.
subst v1; rewrite acc_update; trivial.
case H; intro.
case H0; trivial.
apply map.find_add_other; trivial.
rewrite acc_update_other; auto.
Qed.

Definition array_modifiable (h:store) (m m':memarray)
  (unchanged:array_mod_loc) : Prop :=
  forall v:value,
    alive h v ->
    forall z:Z, unchanged v z -> array_acc m v z = array_acc m' v z.

Lemma array_modifiable_refl :
 forall (h:store) (m:memarray) (unchanged:array_mod_loc),
   array_modifiable h m m unchanged.
red in |- *; auto.
Qed.

Lemma array_modifiable_trans :
 forall (h:store) (m1 m2 m3:memarray) (unch:array_mod_loc),
   array_modifiable h m1 m2 unch ->
   array_modifiable h m2 m3 unch -> array_modifiable h m1 m3 unch.
unfold array_modifiable in |- *; intuition.
transitivity (array_acc m2 v z); auto.
Qed.

Lemma array_modifiable_sub :
 forall (h1 h2:store) (m1 m2:memarray) (unch1 unch2:array_mod_loc),
   store_extends h1 h2 ->
   (forall (v:value) (z:Z), alive h1 v -> unch1 v z -> unch2 v z) ->
   array_modifiable h2 m1 m2 unch2 -> array_modifiable h1 m1 m2 unch1.
unfold store_extends, array_modifiable in |- *; intros.
assert (alive h2 v).
apply alive_extends_stable with h1; auto.
firstorder.
Qed.

(* Combining transitivity and weakening for automatic search *)
Lemma array_modifiable_trans_sub :
 forall (h1 h2:store) (m1 m2 m3:memarray) (unch1 unch2:array_mod_loc),
   (forall (v:value) (z:Z), alive h1 v -> unch1 v z -> unch2 v z) ->
   store_extends h1 h2 ->
   array_modifiable h1 m1 m2 unch1 ->
   array_modifiable h2 m2 m3 unch2 -> array_modifiable h1 m1 m3 unch1.
intros; apply array_modifiable_trans with m2; trivial.
apply array_modifiable_sub with h2 unch2; auto.
Qed.

Lemma array_modifiable_update :
 forall (h:store) (m:memarray) (unch:array_mod_loc) 
   (v:value) (i:Z) (a:A),
   ~ unch v i -> array_modifiable h m (array_update m v i a) unch.
red in |- *; intuition.
rewrite array_acc_update_other; auto.
case (eq_value_dec v v0); auto.
intro; subst; right.
red in |- *; intro; subst; intuition.
Qed.

Lemma array_modifiable_fresh_update :
 forall (h:store) (m1 m2:memarray) (unch:array_mod_loc) 
   (v:value) (i:Z) (a:A),
   fresh h v ->
   array_modifiable h m1 m2 unch ->
   array_modifiable h m1 (array_update m2 v i a) unch.
red in |- *; intuition.
rewrite array_acc_update_other; auto.
left; apply fresh_alive_diff with h; trivial.
Qed.

End Memarray.
Hint Resolve array_modifiable_refl array_modifiable_update.
Hint Resolve array_acc_same.

(** * Allocation of new values *)

Definition new_ref (h:store) := (Ref (new h)).
Lemma new_is_fresh : forall t:store, fresh t (new_ref t).
unfold fresh; simpl.
intros; rewrite find_new; auto.
Save.

Hint Resolve new_is_fresh.

Lemma alive_new_diff :
 forall (t:store) (v:value), alive t v -> v != new_ref t.
red in |- *; red in |- *; intros; subst.
apply (new_is_fresh (t:=t)); auto.
Qed.


Lemma new_alive_diff :
 forall (t:store) (v:value), alive t v -> new_ref t != v.
red in |- *; red in |- *; intros.
apply (new_is_fresh (t:=t)); rewrite H0; auto.
Qed.
Hint Resolve new_alive_diff alive_new_diff.

Lemma new_not_null : forall t:store, new_ref t != Null.
auto.
Qed.
Hint Resolve new_not_null.


Definition allocate (h:store) (v:value) (k:tag) : store :=
  match v with
  | Null => h
  | Ref a => add h a k
  end.

Lemma has_tag_allocated :
 forall (h:store) (v1 v2:value) (k:tag),
   v1 = v2 -> v1 != Null -> has_tag (allocate h v1 k) v2 k.
Proof.
unfold has_tag, allocate in |- *; simple destruct v2; intros; subst;
 simpl in |- *; trivial.
case H0; trivial.
apply contains_add; trivial.
Qed.

Lemma tag_of_allocate_eq :
 forall (h:store) (v1 v2:value) (k:tag),
   v1 = v2 -> v1 != Null -> tag_of (allocate h v1 k) v2 = Some k.
intros; apply has_tag_eq.
apply has_tag_allocated; trivial.
Qed.

Lemma typeof_new_arrint :
 forall (h:store) (n:Z) (v:value),
   v = new_ref h ->
   0 <= n -> typeof (allocate h v (ArrInt n)) v ArrIntType.
Proof.
intros; apply type_arr_int with n; auto.
apply has_tag_allocated; auto.
subst; auto.
Qed.

Lemma typeof_new_arrbool :
 forall (h:store) (n:Z) (v:value),
   v = new_ref h ->
   0 <= n -> typeof (allocate h v (ArrBool n)) v ArrBoolType.
Proof.
intros; apply type_arr_bool with n; auto.
apply has_tag_allocated; auto.
subst; auto.
Qed.

Lemma typeof_new_arrobj :
 forall (h:store) (n:Z) (v:value) (t:javaType),
   v = new_ref h ->
   0 <= n -> typeof (allocate h v (ArrObj n t)) v (ArrayType t).
Proof.
intros; apply type_arr with n; auto.
apply has_tag_allocated; auto.
subst; auto.
Qed.


Lemma typeof_new_obj :
 forall (h:store) (v:value) (c:classId),
   v = new_ref h -> typeof (allocate h v (Obj c)) v (ClassType c).
Proof.
intros; apply type_obj; auto.
apply has_tag_allocated; auto.
subst; auto.
Qed.

Hint Resolve typeof_new_arrint typeof_new_arrbool
  typeof_new_arrobj typeof_new_obj.
 
Lemma tag_of_allocate_other :
 forall (h:store) (v1 v2:value) (t:tag),
   v1 != v2 -> tag_of (allocate h v1 t) v2 = tag_of h v2.
Proof.
unfold tag_of, allocate in |- *; simple destruct v1; auto.
simple destruct v2; intros; auto.
apply find_add_other.
red in |- *; intro; subst; apply H; trivial.
Qed.

Lemma allocate_extends :
 forall (h:store) (v:value) (t:tag),
   fresh h v -> store_extends h (allocate h v t).
red in |- *; intros.
symmetry  in |- *; apply tag_of_allocate_other.
apply fresh_alive_diff with h; auto.
Qed.

Lemma alive_new :
 forall (h:store) (v1 v2:value) (t:tag),
   v1 = v2 -> alive (allocate h v1 t) v2.
Proof.
simple destruct v2; intros; subst; auto.
unfold alive,allocate.
rewrite find_add; auto.
Qed.

Hint Resolve allocate_extends alive_new.

(*
Lemma alive_elim : (h:store)(v1,v2:value)(t:tag)
     (alive (allocate h v1 t) v2)->(alive h v2)\/v1=v2.
Proof.
Destruct v2; Simpl; Auto.
Case v1; Simpl; Auto; Intros a b k.
Case (eq_adrObject_dec a b); Intro.
Subst; Auto.
Rewrite find_add_other; Auto.
Save.

Lemma alive_stable : (h:store)(v1,v2:value)(t:tag)
   (alive h v2)->(alive (allocate h v1 t) v2).
Proof.
Destruct v2; Simpl; Auto.
Case v1; Simpl; Auto; Intros a b k nk.
Case (eq_adrObject_dec a b); Intro.
Rewrite find_add_eq; Auto.
Rewrite find_add_other; Auto.
Save.

Hints Resolve alive_stable.
Lemma fresh_stable :
     (h:store)(v1,v2:value)(t:tag)
     (fresh (allocate h v1 t) v2)->(fresh h v2).
Proof.
Unfold fresh; Intros h v1 v2 t.
Red; Intros na av2; Apply na; Apply alive_stable; Trivial.
Save.


Lemma fresh_diff :
     (h:store)(v1,v2:value)(t:tag)
     (fresh (allocate h v1 t) v2)->v1!=v2.
Proof.
Unfold fresh; Intros h v1 v2 t f eqab.
Apply f; Apply alive_new; Auto.
Save.


Lemma instanceof_stable : 
   (h:store)(ty:javaType)(v1,v2:value)(t:tag)
   (fresh h v1)->
   (instanceof h v2 ty)->(instanceof (allocate h v1 t) v2 ty).
Proof.
Unfold instanceof; Intros h ty v1 v2 t fv1.
Case (eq_value_dec v1 v2); Intro.
Rewrite tag_of_allocate_eq; Trivial.
Subst v1; Rewrite (tag_of_fresh fv1).
Intro; Case (fresh_not_null fv1); Trivial.
Apply (fresh_not_null fv1); Trivial.
Rewrite tag_of_allocate_other; Trivial.
Save.

Hints Resolve instanceof_alive instanceof_stable.
*)

Lemma typeof_extends_stable :
 forall (h h':store) (t:javaType) (v:value),
   typeof h v t -> store_extends h h' -> typeof h' v t.
Proof.
intros h h' t v iof ex.
assert (alive h v).
apply typeof_alive with t; trivial.
red in iof, ex; red in |- *.
replace (tag_of h' v) with (tag_of h v); firstorder.
Save.

Lemma instanceof_extends_stable :
 forall (h h':store) (t:javaType) (v:value),
   instanceof h v t -> store_extends h h' -> instanceof h' v t.
Proof.
intros h h' t v iof ex.
case iof; intuition.
exists x; intuition.
apply typeof_extends_stable with h; trivial.
Qed.

Definition memory_obj := memory value.
Definition memarray_obj := memarray value.

(*
Definition wf_memory [h:store;t:javaType;m:memory_obj] := 
     (v:value)
     ((alive h v)->(instanceof h (acc m v) t))
  /\ ((fresh h v)->(acc m v)=Null).

Lemma wf_memory_instanceof : 
  (h:store;t:javaType;m:memory_obj)(wf_memory h t m)->
  (v:value)(instanceof h (acc m v) t).
Proof.
Unfold wf_memory; Intros.
Case (H v); Intros; Case (alive_dec h v); Intro; Auto.
Rewrite H1; Auto.
Save.
Hints Resolve wf_memory_instanceof.

Lemma wf_memory_alive : 
  (h:store;t:javaType;m:memory_obj)(wf_memory h t m)->
  (v:value)(alive h (acc m v)).
Proof.
Intros; Apply instanceof_alive with t; Auto.
Save.

Hints Resolve wf_memory_alive.

Lemma wf_memory_extends_stable : 
  (h,h':store)(t:javaType)(m:memory_obj)
  (store_extends h h')->(wf_memory h t m)->(wf_memory h' t m).
Proof.
Unfold store_extends wf_memory; Intros.
Case (H0 v); Clear H0; Intuition.
Case (alive_dec h v); Intros; Auto.
Apply instanceof_extends_stable with h; Auto.
Rewrite H1; Auto.
Apply H1; Auto.
Apply fresh_extends_stable with h'; Trivial.
Save.

Lemma wf_memory_update :
 (h:store;t:javaType;m:memory_obj)(this,v:value)
 (wf_memory h t m)->(alive h this)->
 (instanceof h v t)->(wf_memory h t (update m this v)).
Proof.
Unfold wf_memory; Intros.
Case (H v0); Clear H; Intros.
Case (eq_value_dec this v0); Intro.
Rewrite acc_update_eq; Intuition.
Subst this; Case H3; Trivial.
Rewrite acc_update_other; Intuition.
Save.
*)

Lemma arraylength_extends_stable :
 forall (h h':store) (v:value),
   store_extends h h' -> alive h v -> arraylength h v = arraylength h' v.
unfold arraylength in |- *; intros.
rewrite (H v); auto.
Qed.
Hint Resolve arraylength_extends_stable.

(*
Definition wf_memarray 
   [h:store;m:memarray_obj] := 
     (v:value)
     ((alive h v)
     -> (k:Z)`0<=k<(arraylength h v)`->
        (instanceof h (array_acc m v k) (arrayelemtype h v)))
   /\ ((fresh h v)->(k:Z)(array_acc m v k)=Null).

Lemma wf_memarray_instanceof : 
  (h:store)(m:memarray_obj)(wf_memarray h m)->(v:value)
  (v != Null) -> (t:javaType)(instanceof h v (ArrayType t))
  -> (k:Z)`0<=k<(arraylength h v)`->
        (instanceof h (array_acc m v k) t).
Intros.
Case (H v); Intros; Auto.
Apply instanceof_subType_trans with (arrayelemtype h v).
Apply H3; Auto.
Apply instanceof_alive with (ArrayType t); Auto.
Generalize H1; Unfold instanceof arrayelemtype; 
Case (tag_of h v); Simpl; Auto.
Destruct t0; Simpl; Auto; Try Contradiction.
Destruct a; Simpl; Auto; Try Contradiction.
Intro; Case H0; Trivial.
Save.
*)
Lemma arrayelemtype_extends_stable :
 forall (h h':store) (m:memarray_obj),
   store_extends h h' ->
   forall v:value, alive h v -> arrayelemtype h v = arrayelemtype h' v.
intros.
assert (E := H v); intros; clear H.
unfold arrayelemtype in |- *; rewrite E; auto.
Qed.
(*
Lemma wf_memarray_extends_stable : 
  (h,h':store)(m:memarray_obj)
  (store_extends h h')->(wf_memarray h m)
  ->(wf_memarray h' m).
Proof.
Red; Intros.
Case (H0 v); Clear H0; Intuition.
Case (alive_dec h v); Intros; Auto.
Rewrite <- (arrayelemtype_extends_stable m H); Auto.
Apply instanceof_extends_stable with h; Auto.
Apply H0; Auto.
Rewrite arraylength_extends_stable with h h' v; Auto.
Rewrite H1; Auto.
Apply H1; Auto.
Apply fresh_extends_stable with h'; Trivial.
Save.

Lemma wf_memarray_update :
 (h:store;m:memarray_obj)
 (wf_memarray h m)
  ->(this,v:value)(z:Z)
    (is_valid_array_update h this z v)
  ->(wf_memarray h (array_update m this z v)).
Proof.
Unfold wf_memarray; Intros.
Case H0; Intros.
Assert (alive h this).
Apply is_valid_index_alive with z; Auto.
Case (H v0); Clear H; Intros.
Case (eq_value_dec this v0); Intro.
Split; Intros.
Case (Z_eq_dec z k); Intro.
Rewrite array_acc_update_eq; Intuition.
Subst this; Auto.
Rewrite array_acc_update_other; Intuition.
Rewrite array_acc_update_other; Intuition.
Left; Apply alive_fresh_diff with h; Trivial.
Split; Intros; Rewrite array_acc_update_other; Intuition.
Save.
*)
Lemma arraylength_always_pos :
 forall (h:store) (v:value), 0 <= arraylength h v.
unfold arraylength in |- *; intros.
case (tag_of h v); auto with *.
simple destruct t; auto with *.
simple destruct n; simpl in |- *; auto with *.
Qed.
Hint Resolve arraylength_always_pos.

Lemma arraylength_pos_val_not_null :
 forall (h:store) (v:value), 0 < arraylength h v -> v != Null.
unfold arraylength in |- *; intros h v.
case v; auto with *.
Qed.

Lemma arraylength_pos_not_null_val :
 forall (h:store) (v:value), 0 < arraylength h v -> Null != v.
unfold arraylength in |- *; intros h v.
case v; auto with *.
Qed.

Lemma arraylength_new_arrint :
 forall (h:store) (n:Z) (v:value),
   v = new_ref h -> 0 <= n -> arraylength (allocate h v (ArrInt n)) v = n.
Proof.
unfold arraylength in |- *; intros.
rewrite tag_of_allocate_eq; simpl in |- *; auto.
apply Z_entier_pos; trivial.
subst; auto.
Qed.

Lemma arraylength_new_arrbool :
 forall (h:store) (n:Z) (v:value),
   v = new_ref h -> 0 <= n -> arraylength (allocate h v (ArrBool n)) v = n.
Proof.
unfold arraylength in |- *; intros.
rewrite tag_of_allocate_eq; simpl in |- *; auto.
apply Z_entier_pos; trivial.
subst; auto.
Qed.

Lemma arraylength_new_arrobj :
 forall (h:store) (n:Z) (v:value) (t:javaType),
   v = new_ref h -> 0 <= n -> arraylength (allocate h v (ArrObj n t)) v = n.
Proof.
unfold arraylength in |- *; intros.
rewrite tag_of_allocate_eq; simpl in |- *; auto.
apply Z_entier_pos; trivial.
subst; auto.
Qed.


Lemma is_valid_index_intro :
 forall (h:store) (v:value) (i:Z),
   0 <= i < arraylength h v -> is_valid_index h v i.
Proof.
auto.
Qed.

Hint Resolve is_valid_index_intro.

(** Values of incompatible types are different *)

Lemma subType_comparable :
 forall t1 t2 t3:javaType,
   subType t1 t2 -> subType t1 t3 -> subType t2 t3 \/ subType t3 t2.
simple induction t1; simple destruct t2; simple destruct t3; simpl in |- *;
 intuition.
subst; auto.
subst; auto.
apply extends_comparable with c; trivial.
subst; auto.
Qed.

Lemma instanceof_comparable :
 forall (h:store) (v:value),
   v != Null ->
   forall t1 t2:javaType,
     instanceof h v t1 -> instanceof h v t2 -> subType t1 t2 \/ subType t2 t1.
intros h v; destruct 2; destruct 1; intuition.
apply subType_comparable with x; auto.
rewrite (@typeof_unique h v x x0); auto.
Qed.

Definition not_comparable (t1 t2:javaType) :=
  ~ subType t1 t2 /\ ~ subType t2 t1.

Lemma not_comparable_sym :
 forall t1 t2:javaType, not_comparable t1 t2 -> not_comparable t2 t1.
unfold not_comparable in |- *; intuition.
Qed.

Lemma not_comparable_class_array :
 forall (c:classId) (t:javaType),
   c <> ObjectClass -> not_comparable (ClassType c) (ArrayType t).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_array_class :
 forall (c:classId) (t:javaType),
   c <> ObjectClass -> not_comparable (ArrayType t) (ClassType c).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_class :
 forall c1 c2:classId,
   ~ extends c1 c2 ->
   ~ extends c2 c1 -> not_comparable (ClassType c1) (ClassType c2).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_array :
 forall t1 t2:javaType,
   not_comparable t1 t2 -> not_comparable (ArrayType t1) (ArrayType t2).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_type_diff :
 forall (h:store) (t1 t2:javaType) (v1 v2:value),
   v1 != Null \/ v2 != Null ->
   instanceof h v1 t1 ->
   instanceof h v2 t2 -> not_comparable t1 t2 -> v1 != v2.
unfold not_comparable in |- *; red in |- *; red in |- *; intros; subst; auto.
case instanceof_comparable with (h := h) (v := v2) (t1 := t1) (t2 := t2);
 intuition.
Qed.

(** Modifiable *)

Definition nothing_loc : mod_loc := fun v => True.
Definition value_loc (v1:value) : mod_loc := fun v2 => v1 != v2.
Definition everything_loc : mod_loc := fun v => False.
Definition array_nothing_loc : array_mod_loc := fun v c => True.
Definition array_everything_loc : array_mod_loc := fun v c => False.


Lemma nothing_intro : forall v:value, nothing_loc v.
red in |- *; auto.
Qed.

Lemma array_nothing_intro : forall (v:value) (c:Z), array_nothing_loc v c.
red in |- *; auto.
Qed.

Hint Resolve array_nothing_intro nothing_intro.

Lemma not_everything : forall v:value, ~ everything_loc v.
unfold everything_loc in |- *; intuition.
Qed.

Lemma not_array_everything :
 forall (v:value) (c:Z), ~ array_everything_loc v c.
unfold everything_loc in |- *; intuition.
Qed.

Lemma value_loc_intro : forall v1 v:value, v1 != v -> value_loc v1 v.
red in |- *; trivial.
Qed.

Hint Resolve value_loc_intro.

Lemma not_value_loc_refl : forall v:value, ~ value_loc v v.
unfold value_loc in |- *; intuition.
Qed.

Lemma not_value_loc_eq : forall v1 v:value, v1 = v -> ~ value_loc v1 v.
unfold value_loc in |- *; intuition.
Qed.

Definition inter_loc (l1 l2:mod_loc) v := l1 v /\ l2 v.

Definition array_inter_loc (l1 l2:array_mod_loc) v c := l1 v c /\ l2 v c.

Lemma inter_intro :
 forall (l1 l2:mod_loc) (v:value), l1 v -> l2 v -> inter_loc l1 l2 v.
red in |- *; auto.
Qed.

Lemma array_inter_intro :
 forall (l1 l2:array_mod_loc) (v:value) (c:Z),
   l1 v c -> l2 v c -> array_inter_loc l1 l2 v c.
red in |- *; auto.
Qed.

Hint Resolve inter_intro array_inter_intro.

Lemma not_inter :
 forall (l1 l2:mod_loc) (v:value), ~ l1 v \/ ~ l2 v -> ~ inter_loc l1 l2 v.
unfold inter_loc in |- *; intuition.
Qed.

Lemma array_not_inter :
 forall (l1 l2:array_mod_loc) (v:value) (c:Z),
   ~ l1 v c \/ ~ l2 v c -> ~ array_inter_loc l1 l2 v c.
unfold array_inter_loc in |- *; intuition.
Qed.

Hint Resolve not_inter array_not_inter.

Definition array_loc (v1:value) (c1:Z) : array_mod_loc :=
  fun v c => v1 != v \/ c1 <> c.

Lemma array_loc_intro :
 forall (v1 v:value) (c1 c:Z), v1 != v \/ c1 <> c -> array_loc v1 c1 v c.
red in |- *; trivial.
Qed.

Lemma not_array_refl : forall (v:value) (c:Z), ~ array_loc v c v c.
unfold array_loc in |- *; intuition.
Qed.

Lemma not_array_eq :
 forall (v1 v:value) (c1 c:Z), v1 = v -> c1 = c -> ~ array_loc v c v1 c1.
unfold array_loc in |- *; intuition.
Qed.

Hint Resolve array_loc_intro not_array_refl not_array_eq.

Definition array_all_loc (v1:value) : array_mod_loc := fun v c => v1 != v.

Lemma array_all_intro :
 forall (v1 v:value) (c:Z), v1 != v -> array_all_loc v1 v c.
red in |- *; auto.
Qed.

Lemma not_array_all_refl : forall (v:value) (c:Z), ~ array_all_loc v v c.
unfold array_all_loc in |- *; intuition.
Qed.

Lemma not_array_all_eq :
 forall (v1 v:value) (c:Z), v1 = v -> ~ array_all_loc v v1 c.
unfold array_all_loc in |- *; intuition.
Qed.

Hint Resolve array_all_intro not_array_all_refl not_array_all_eq.

Definition array_sub_loc (v1:value) (z1 z2:Z) : array_mod_loc :=
  fun v i => v1 != v \/ i < z1 \/ z2 < i.

Lemma array_sub_intro_val :
 forall (v1:value) (z1 z2:Z) (v:value) (c:Z),
   v1 != v -> array_sub_loc v1 z1 z2 v c.
unfold array_sub_loc in |- *; intuition.
Qed.

Lemma array_sub_intro_index :
 forall (v v1:value) (z1 z2 i:Z),
   i < z1 \/ z2 < i -> array_sub_loc v1 z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Hint Resolve array_sub_intro_val array_sub_intro_index.

Lemma not_array_sub_refl :
 forall (v:value) (z1 z2 i:Z), z1 <= i <= z2 -> ~ array_sub_loc v z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Lemma not_array_sub_eq :
 forall (v v1:value) (z1 z2 i:Z),
   v = v1 -> z1 <= i <= z2 -> ~ array_sub_loc v1 z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Hint Resolve not_array_sub_refl not_array_sub_eq.

(** Special instantiations for integer, booleans and values, 
    because Why does not allow polymorphism *)

(*
Definition memory_int := memory Z.
Definition memory_bool := memory bool.
*)

Definition access_int := acc (A:=Z).
Definition access_bool := acc (A:=bool).
Definition access_obj := acc (A:=value).

Definition update_int := update (A:=Z).
Definition update_bool := update (A:=bool).
Definition update_obj := update (A:=value).

(*
Definition modifiable_int := modifiable (A:=Z).
Definition modifiable_bool := modifiable (A:=bool).
Definition modifiable_obj := modifiable (A:=value). 

Definition memarray_int := memarray Z.
Definition memarray_bool := memarray bool.
*)

Definition array_access_int := array_acc (A:=Z).
Definition array_access_bool := array_acc (A:=bool). 
Definition array_access_obj := array_acc (A:=value).

Definition array_update_int := array_update (A:=Z).
Definition array_update_bool := array_update (A:=bool).
Definition array_update_obj := array_update (A:=value).

(*
Definition array_modifiable_int := array_modifiable (A:=Z).
Definition array_modifiable_bool := array_modifiable (A:=bool).
Definition array_modifiable_obj := array_modifiable (A:=value).
*)
Opaque adrObject.
Opaque new_ref.

End ObjectAdr.
