

open Format;;
open Java_context_types;;

let output_header form =
  fprintf form "(* Coq Model generated by Krakatoa. Do not edit *)@.";
  fprintf form "@.";
  fprintf form "Add LoadPath \"%s\".@." !Config.krakatoa_lib;
  fprintf form "Require Export preambule.@.";
  fprintf form "Require Export classes.@.";
  fprintf form "Require Export %s_memory.@."
    (if !Config.use_global_memory_model
     then "global" else "local");
  fprintf form "Require Export ZArith.@.";
  fprintf form "Require Export java_base.@.";
  fprintf form "Require Export Rbase.@.";
  fprintf form "Require Export Rfunctions.@.";
  fprintf form "@.";
  fprintf form "@.";
;;

let rec get_all_classes acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_all_classes acc pe.package_entry_contents
       | Interface_entry te ->
	   (te.interface_entry_name, Some "Object")::acc
       | Class_entry te ->
	   begin
	     match te.class_entry_extends with
	       | None -> 
		   if te.class_entry_name = "Object"
		   then (te.class_entry_name, None)::acc
		   else (te.class_entry_name, Some "Object")::acc
	       | Some x -> (te.class_entry_name, Some x.class_entry_name)::acc
	   end
       | _ -> acc)
    acc
    env
;;


let rec topo_sort a b acc =
  match a with
    | [] -> 
	if b=[] then List.rev acc else topo_sort b [] acc
    | x::c -> 
	match x with
	  | (_,None) -> topo_sort c b (x::acc)
	  | (_,Some super) ->
	      if List.mem_assoc super acc
	      then topo_sort c b (x::acc)
	      else topo_sort c (x::b) acc
;;

let get_classes env =
  let l = get_all_classes [] env in
    topo_sort l [] []
;;


let rec get_fields acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_fields acc pe.package_entry_contents
       | Class_entry te ->
	   List.fold_left
	   (fun acc (_,entry) -> match entry with
	      | Instance_variable_entry ive ->
		  if ive.instance_variable_entry_is_static
		  then acc
		  else
		    (ive,te)::acc
	      | _ -> acc)
	   acc
	   te.class_entry_contents
       | _ -> acc)
    acc
    env
;;




let rec output_all_classes form classes =
  List.iter
    (fun (this,super) -> 
       fprintf form "| %s : classIdI@." this)
       classes
;;

let rec output_all_superclasses form classes =
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> fprintf form "  | %s => None@." this
	 | Some x -> fprintf form "  | %s => (Some %s)@." this x)
    classes
;;

let rec output_all_wf_classes form classes =
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None ->  
	     fprintf form 
	     "Lemma wf_%s : (Acc (fun i1 i2 => (def (super i2) i1)) %s).@."
	     this this;
	     fprintf form 
	       "constructor; simpl; intros.@.";
	     fprintf form 
	       "case (undef_not_def (A:=classId) (x:=None)) with y; auto.@.";
	     fprintf form 
	       "Defined.@.";
	     fprintf form 
	       "@."
	 | Some x -> 
	     fprintf form 
	     "Lemma wf_%s : (Acc (fun i1 i2 => (def (super i2) i1)) %s).@."
	     this this;
	     fprintf form 
	       "constructor; simpl; intros.@.";
	     fprintf form 
	       "replace y with %s; auto.@." x;
	     fprintf form 
	       "apply wf_%s.@." x;
	     fprintf form 
	       "Defined.@.";
	     fprintf form 
	       "@.")
    classes
;;

let output_classes form env =
  let classes = get_classes env in
    fprintf form "(* Classes *)@.";
    fprintf form "@.";
    fprintf form "Inductive classIdI : Set :=@.";
    output_all_classes form classes;
    fprintf form ".@.";
    fprintf form "Definition classId := classIdI.@.";
    fprintf form "@.";
    fprintf form "Definition ObjectClass := Object.@.";
    fprintf form "@.";
    fprintf form "Definition eq_classId_dec : forall (i1 i2:classId), {i1=i2}+{~i1=i2}.@.";
    fprintf form "decide equality. Defined.@.";
    fprintf form "@.";
    fprintf form "Definition super (i:classId) : (option classId) :=@.";
    fprintf form "  match i with@.";
    output_all_superclasses form classes;
    fprintf form "  end.@.";
    fprintf form "@.";
    fprintf form "Lemma superObject : (super ObjectClass)=None.@.";
    fprintf form "auto. Save.@.";
    fprintf form "@.";
    output_all_wf_classes form classes;
    fprintf form "Lemma super_wf : (well_founded (fun i1 i2 => (def (super i2) i1))).@.";
    fprintf form "red; intros.@.";
    fprintf form "case a.@.";
    List.iter
      (fun  (this,_) -> fprintf form "apply wf_%s.@." this)
      classes;
    fprintf form "Defined.@.";
    fprintf form "@.";
;;

let rec fprintf_type form t =
  match t with
  | Integer_type -> fprintf form "(IntT classId)"
  | Float_type -> fprintf form "(FloatT classId)"
  | Boolean_type -> fprintf form "(BoolT classId)"
  | Null_type -> fprintf form "(NullT classId)"
  | Class_type(te) -> fprintf form "(ClassT classId %s)" te.class_entry_name
  | Interface_type(te) -> 
      (*i fprintf form "(InterfaceT classId %s)" te.interface_entry_name i*)
      fprintf form "(ClassT classId %s)" te.interface_entry_name
  | Array_type(t) -> fprintf form "(ArrayT classId %a)" fprintf_type t
  | Type_type -> assert false
;;

let name_field (ive,te) =
  (ive.instance_variable_entry_trans_name,
   ive.instance_variable_entry_type,
   te.class_entry_name)
;;

let output_fields form env fields = 
  let l = List.map name_field fields in
  fprintf form "(* fields *)@.";
  fprintf form "@.";
  fprintf form "Inductive fieldIdI : Set :=@.";
  List.iter
      (fun (f,t,c) ->
	 fprintf form "| %sI : fieldIdI@." f)
      l;
    fprintf form ".@.";
    fprintf form "@.";
    fprintf form "Definition fieldId := fieldIdI.@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "Definition %s : fieldId := %sI.@." f f)
      l;
    fprintf form "@.";
    fprintf form "Definition eq_fieldId_dec : @.";
    fprintf form " forall (f1 f2:fieldId), {f1=f2}+{~f1=f2}.@.";
    fprintf form "Proof. decide equality. Defined.@.";
    fprintf form "@.";
    fprintf form "Definition defined_in : fieldId -> classId :=@.";
    fprintf form " fun (f:fieldId) =>@.";
    fprintf form " match f with@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "| %s => %s@." f c)
      l;
    fprintf form " end.@.";
    fprintf form "@.";
    fprintf form "Lemma no_field_in_Object :@.";
    fprintf form "  forall (f:fieldId),~(defined_in f)=ObjectClass.@.";
    fprintf form "Proof. induction f; auto. Save.@.";
    fprintf form "@.";
    fprintf form "(* types of fields *)@.";
    fprintf form "@.";
    fprintf form "Definition fieldT : fieldId -> (javaT classId) :=@.";
    fprintf form " fun (f:fieldId) =>@.";
    fprintf form " match f with@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "| %s => %a@." f fprintf_type t)
      l;
    fprintf form "end.@.";
    fprintf form "@.";
;;

 


let output_model form env =
  output_header form;
  fprintf form "Module KrakClasses.@.";
  fprintf form "@.";
  output_classes form env;
  if !Config.use_global_memory_model
  then
    begin
      let fields = get_fields [] env in
      output_fields form env fields;
    end;
  fprintf form "End KrakClasses.@.";
  fprintf form "@.";
  fprintf form "Import KrakClasses.@.";
  fprintf form "Module KrakObj := (ObjectAdr KrakClasses).@.";
  fprintf form "Module KrakGen := KrakObj.CGen.@.";
  fprintf form "@.";
(*  output_tactic form *)
;;

let output_sx_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x -> 
	     fprintf form 
	       "@[(BG_PUSH@ (FORALL (h v) (IMPLIES@ (instanceof h v (ClassType %s))@ "
	       this;
	     fprintf form 
	       "(instanceof h v (ClassType %s)))))@]@.@."
	       x) 
    classes;
  fprintf form "@.";
(*  output_tactic form *)
;;

let output_rv_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x -> 
	     fprintf form 
	       "@[(forall H V (->@ (instanceof H V (classType %s))@ "
	       this;
	     fprintf form 
	       "(instanceof H V (classType %s))))@]@.@."
	       x) 
    classes;
  fprintf form "@.";
(*  output_tactic form *)
;;


let file_copy src form =
  let cin = open_in src in
  try
    while true do
      let s = input_line cin in fprintf form "%s\n" s
    done
  with End_of_file ->
    close_in cin

let output_pvs_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun (this,_) -> fprintf form "%s : classId@\n" this)
    classes;
  fprintf form "@\n";
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x -> 
	     fprintf form 
	       "@[<hov 2>instanceof_%s_%s: AXIOM@ forall (h:store) (v:value): instanceof(h,v,ClassType(%s))@ implies instanceof(h,v,ClassType(%s))@]@.@." this x this x) 
    classes;
(*  output_tactic form *)
;;



(* makefile *)

let rec print_all_paths form env =
  List.iter
    (fun (_,entry) -> match entry with
       | Package_entry pe ->
	   if pe.package_entry_directory <> !Config.source_package then
	     () (*i fprintf form " \\@.\t-I %s" pe.package_entry_directory i*);
	   print_all_paths form pe.package_entry_contents
       | _ -> ())
    env
;;


let string_tr s ch1 ch2 =
  let t = String.copy s in
  for i=0 to pred(String.length t) do
    if String.get t i = ch1 then String.set t i ch2
  done;
  t
;;

let rec find_methods_to_prove acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   find_methods_to_prove acc pe.package_entry_contents
       | Class_entry te ->
	   let acc =
	     List.fold_left
	       (fun acc (_,e) ->
		  match e with
		    | Method_entry l ->
			List.fold_left
			  (fun acc me ->
			     if me.method_entry_routine.routine_entry_produce_body
			     then 
			       (te.class_entry_name ^ "_" ^ me.method_entry_name)
			       ::acc
			     else acc)
			  acc l
		    | _ -> acc)
	       acc te.class_entry_contents
	   in
	   List.fold_left
	     (fun acc ce ->
		if ce.constructor_entry_routine.routine_entry_produce_body
		then ce.constructor_entry_trans_name::acc
		else acc)
	     acc te.class_entry_constructors
       | _ -> acc)
    acc
    env
;;

let find_methods_to_prove env =
  (* reverse list to get methods in the same order as in the files *)
 List.rev (find_methods_to_prove [] env)
;;

let output_krak_validations form env spec_files =
  fprintf form "@.";
  fprintf form "Require Import Krak_spec_why.@.";
  let l = find_methods_to_prove env in
  List.iter
    (fun f -> 
       fprintf form "Require Import %s_valid.@." f)
    l;
  fprintf form "@.";
  fprintf form "Definition typeof := fun (A:Set) (_:A) => A.@.";
  fprintf form "@.";
  List.iter
    (fun f -> 
       fprintf form 
	 "Definition %s_krakatoa_validation : (typeof _ %s_parameter) := %s_body.@."
	 f f f)
    l;
  fprintf form "@."
;;



let output_makefile form env spec_files =
  let cmdline = ref "" in
  for 
    i=0 to Array.length Sys.argv - 1 do
      cmdline := !cmdline ^ " " ^ Sys.argv.(i);
  done; 
  fprintf form "# makefile generated by Krakatoa. Do not edit@.";
  fprintf form "@.";
  fprintf form "#@.";
  fprintf form "# This makefile was generated by the command line :@.";
  fprintf form "#%s" !cmdline;
  fprintf form "@.";
  fprintf form "#@.";
  fprintf form "@.";
  fprintf form "KRAKATOALIB=%s@." !Config.krakatoa_lib;
  fprintf form "@.";
  fprintf form "COQDIRS=-I $(KRAKATOALIB)%s" !Config.coqdirs;
  print_all_paths form env;
  fprintf form "@.";  
  fprintf form "@.";  
  fprintf form "WHYEXEC=why";
  fprintf form "@.";  
  fprintf form "@.";  
  fprintf form "WHY=$(WHYEXEC) --coq-v8%s --coq-tactic \"intuition\"@." 
    (if !Config.why_validation then " --valid" else ""); 
  fprintf form "@.";

  if !Config.output_language = `SIMPLIFY
  then
    begin
      fprintf form "@.";  
      fprintf form "WHYSIMPLIFY=$(WHYEXEC) --simplify --no-simplify-prelude@." ;
      fprintf form "@.";
    end;

  if !Config.output_language = `HARVEY
  then
    begin
      fprintf form "@.";  
      fprintf form "WHYHARVEY=$(WHYEXEC) --harvey --no-harvey-prelude@." ;
      fprintf form "@.";
    end;

  if !Config.output_language = `PVS
  then
    begin
      fprintf form "@.";  
      fprintf form "WHYPVS=$(WHYEXEC) --pvs --pvs-preamble \"importing Krak_model\"@." ;
      fprintf form "@.";
    end;

  (* the main makefile rule *)
  fprintf form "all:";
  let l = find_methods_to_prove env in
  List.iter (fun s -> fprintf form " %s_why.vo" s) l;
  if !Config.why_validation then
    begin
      List.iter (fun s -> fprintf form " %s_valid.vo" s) l;
      fprintf form " Krak_validations.vo"
    end;
  fprintf form "@.";
  fprintf form "@.";
  if !Config.output_language = `SIMPLIFY then
    begin
      fprintf form "simplify:@.";
      fprintf form "\tfor i in";
      List.iter (fun s -> fprintf form " %s_why" s) l;
      fprintf form "; do make $$i.sxcheck; done@.";
      fprintf form "@.";
    end;
  if !Config.output_language = `HARVEY then
    begin
      fprintf form "harvey:";
      List.iter (fun s -> fprintf form " %s_why.rvcheck" s) l;
      fprintf form "@.";
      fprintf form "@.";
    end;
  if !Config.output_language = `PVS then
    begin
      fprintf form "pvs:";
      List.iter (fun s -> fprintf form " %s_why.pvs" s) l;
      fprintf form "@.";
      fprintf form "\techo '(typecheck \"Krak_model\")' > pvsbatch.el@.";
      List.iter 
	(fun s -> 
	   fprintf form "\techo '(typecheck \"%s_why\")' >> pvsbatch.el@." s)
	l;
      fprintf form "\tpvs -q -v 3 -batch -l pvsbatch.el@.";
      fprintf form "@.";
    end;

  (* rule for building memory_model.why and memory_model_why.v *)
  let source_model =
    if !Config.use_global_memory_model
    then "global" else "local"
  in
  fprintf form "%s_memory.why: $(KRAKATOALIB)/%s_memory_template.why@." 
    source_model source_model;
  fprintf form "\tcp $(KRAKATOALIB)/%s_memory_template.why %s_memory.why@.@." 
    source_model source_model;
  fprintf form "%s_memory_why.v: $(KRAKATOALIB)/%s_memory_template_why.v@." 
    source_model source_model;
  fprintf form "\tcp $(KRAKATOALIB)/%s_memory_template_why.v %s_memory_why.v@.@." 
    source_model source_model;
  if !Config.why_validation then 
    begin
      fprintf form "%s_memory_valid.v: $(KRAKATOALIB)/%s_memory_template_valid.v@." 
	source_model source_model;
      fprintf form "\tcp $(KRAKATOALIB)/%s_memory_template_valid.v %s_memory_valid.v@.@." 
	source_model source_model;
    end;


  (* rule for executing why on files generated by krakatoa *)
  fprintf form "Krak_spec_why.v";
  if !Config.why_validation then fprintf form " Krak_spec_valid.v";
  fprintf form " \\@.";
  List.iter
    (fun s -> fprintf form "%s_why.v \\@." s)
    l;
  fprintf form "\t: %s_memory.why" source_model;
  fprintf form " \\@.\t\tKrak_spec.why";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s.why" s)
    l;
  fprintf form "@.\t@@echo Running why on generated programs...@.";
  fprintf form 
    "@.\t$(WHY) --coq-preamble \"Load spec_imports.\" %s_memory.why" 
    source_model;
  fprintf form " \\@.\t\tKrak_spec.why";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s.why" s)
    l;
  fprintf form "@.@.";

  (* running why with simplify output *)
  if !Config.output_language = `SIMPLIFY
  then
    begin
      fprintf form "Krak_spec_why.sx \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.sx \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYSIMPLIFY) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

  if !Config.output_language = `HARVEY
  then
    begin
      fprintf form "Krak_spec_why.rv \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.rv \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYHARVEY) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

  if !Config.output_language = `PVS
  then
    begin
      fprintf form "Krak_spec_why.pvs \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.pvs \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYPVS) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

  (* dependency of the coq object on the importations files *)
(*
  List.iter
    (fun f -> 
       let f = Filename.basename f in 
       fprintf form "%s_param.vo: %s_model_imports.vo@." f source_model)
    spec_files;
*)
(*
  fprintf form "Krak_model.vo: %s_model_imports.vo@." source_model;
*)
(*
  fprintf form "Krak_model_why.vo: spec_imports.vo@.";
*)
  List.iter
    (fun s -> fprintf form "%s_why.vo: Krak_%smodel.vo@." s 
	 (if !Config.use_global_memory_model then "global_" else ""))
    l;
  fprintf form "@.";



  (* dependency of the krak validations *)
  if !Config.why_validation then 
    begin
      fprintf form "Krak_validations.vo:";
      List.iter
	(fun s -> fprintf form " %s_valid.vo" s)
	l;
      fprintf form "@.@.";
    end;

  fprintf form "# generic rules@.";
  fprintf form "@.";
  fprintf form ".SUFFIXES: .v .vo@.";
  fprintf form "@.";
  fprintf form ".v.vo: @.";
  fprintf form "\tcoqc $(COQDIRS)%s $<@." !Config.coqopts;
  fprintf form "@.";

  if !Config.output_language = `SIMPLIFY
  then
    begin
      fprintf form ".SUFFIXES: .sx .sxcheck@.";
      fprintf form "@.";
      fprintf form ".sx.sxcheck: @.";
      fprintf form "\tcat $(KRAKATOALIB)/local_memory.sx Krak_model.sx \\@.";
      fprintf form "\t\t$< > $<.all@." ;
      fprintf form "\tSimplify $<.all@." ;
      fprintf form "@.";
    end;

  if !Config.output_language = `HARVEY
  then
    begin
      fprintf form ".SUFFIXES: .rv .rvcheck@.";
      fprintf form "@.";
      fprintf form ".rv.rvcheck: @.";
      fprintf form "\techo '(' > $*.all.rv@." ;      
      fprintf form "\tcat $(KRAKATOALIB)/local_memory.rv Krak_model.rv \\@.";
      fprintf form "\t\t >> $*.all.rv@." ;
      fprintf form "\techo ')' >> $*.all.rv@." ;      
      fprintf form "\trvc $*.all.rv@." ;
      fprintf form "@.";
    end;

  if !Config.output_language = `PVS
  then
    begin
      fprintf form ".SUFFIXES: .pvs@.";
      fprintf form "@.";
    end;

  fprintf form "# regenerate@.";
  fprintf form "@.";
  fprintf form ".PHONY: regenerate@.";
  fprintf form "@.";
  fprintf form "regenerate:@.";
  fprintf form "\tcd .. && %s@." !cmdline;
  fprintf form "@.";
  
  fprintf form "# clean and depend@.";
  fprintf form "@.";
  fprintf form "clean:@.";
  fprintf form "\trm -f *.vo *.bak Krak_spec_why.v *_valid.v %s_memory.why %s_memory_why.v@." 
    source_model source_model;
  fprintf form "@.";
  if !Config.output_language = `COQ then begin
  fprintf form "depend .depend: %s_memory_why.v" source_model;
  fprintf form " \\@.\tKrak_spec_why.v";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s_why.v" s)
    l;
  fprintf form "@.";
  if Version.has_coq then
    begin
      fprintf form "\trm -f .depend@.";
      fprintf form "\tcoqdep $(COQDIRS) *.v >> .depend@.";
    end
  else 
    fprintf form "\ttouch .depend@.";
  fprintf form "@.";
  fprintf form "include .depend@.";
  fprintf form "@."
  end
;;
