
open Java_context_types;;
open Ast_types;;
open Pre_translate;;

let sprintf = Format.sprintf


let type_default_value produce_a_value t =
  match t with
    | Integer_type -> 
	let v = Why.Cte(Why.Prim_int(0)) in
	if produce_a_value then Why.App(Why.Var("Primitive_int"),v) else v
    | Float_type -> 
	let v = Why.Cte(Why.Prim_float("0","","")) in
	if produce_a_value then Why.App(Why.Var("Primitive_float"),v) else v
    | Boolean_type -> 
	let v = Why.Cte(Why.Prim_bool(false)) in
	if produce_a_value then Why.App(Why.Var("Primitive_bool"),v) else v
    | Class_type(_) -> Why.Var("Null")
    | Array_type _ ->  Why.Var("Null")
    | Interface_type _ -> assert false
    | Null_type -> assert false
;;

let update_coerce t accu = match t with 
      | Integer_type -> "int_update", Why.LApp("Primitive_int",[accu])
      | Boolean_type -> "bool_update",Why.LApp("Primitive_bool",[accu])
      | Float_type -> "float_update",Why.LApp("Primitive_float",[accu])
      | Null_type -> ill_typed "tr_field_assignment(null)"
      | Class_type(_) 
      | Array_type(_)
      | Interface_type(_) ->  "update", accu
	

(* translation of types : to primitive type or to "Value" *)

let tr_type t =
  match t with
    | Integer_type -> Why.int_type
    | Float_type -> Why.float_type
    | Boolean_type -> Why.bool_type
    | Class_type(ce) -> 
	Why.base_type(sprintf "value (* %s *)" ce.class_entry_name)
    | Array_type t -> 
	Format.fprintf 
	  Format.str_formatter
	  "value (* %a array *)" Java_context.fprintf_java_type t;
	Why.base_type (Format.flush_str_formatter ())
    | Interface_type ie -> 	
	Why.base_type(sprintf "value (* %s *)" ie.interface_entry_name)
    | Null_type -> Why.unit_type

let tr_type_detailed t =
  match t with
    | Integer_type -> "int"
    | Float_type -> "float"
    | Boolean_type -> "bool"
    | Class_type(ce) -> ce.class_entry_name
    | Array_type _ -> unsupported_no_loc "tr_type_detailed(array_type)"
    | Interface_type _ -> unsupported_no_loc "tr_type_detailed(interface_type)"
    | Null_type -> "unit"

let rec tr_type_in_coq t =
  match t with
    | Integer_type -> Why.LVar("IntType")
    | Float_type -> Why.LVar("FloatType")
    | Boolean_type -> Why.LVar("BoolType")
    | Class_type(ce) -> 
	(*io->ht*) Why.LApp ("ClassType", [ (*i*)
	Why.LVar ce.class_entry_name
	  (*i*) ]) (*i*)
    | Array_type t1 -> 
	Why.LApp
	  ("ArrayType",
	   [tr_type_in_coq t1])
    | Interface_type _ -> 
	unsupported_no_loc "tr_type_in_coq(interface_type)"
    | Null_type -> 
	unsupported_no_loc "tr_type_in_coq(null_type)"

let tr_literal loc l =
  match l with
    | Int(n) -> Why.Cte(Why.Prim_int(n))
    | Float(i,f,e) -> Why.Cte(Why.Prim_float(i,f,e))
    | Bool(b) -> Why.Cte(Why.Prim_bool(b))
    | String(s) -> unsupported loc "tr_literal(string)"
    | Char(c) -> unsupported loc "tr_literal(char)"
    | Null -> Why.Var("Null")
;;

let tempvar_count = ref 0;;

let build_app e args =
  let rec build n e args =
    match args with
      | [] -> e
      | ((Why.Var _) | (Why.Cte _) | (Why.Deref _) as p)::l ->
	  build n (Why.App(e,p)) l
      | p::l ->
	  incr tempvar_count;
	  let v = "krakv"^(string_of_int !tempvar_count) in
	  Why.Let(v,p,
	    build (succ n) (Why.App(e,Why.Var(v))) l)
  in build 1 e args

(* Build a Let only if the expression is complex 
    [v : string,f : Why.expr -> Why.term -> Why.expr; e: Why.expr] 
*)

let why_let_var v f e = match e with 
    (Why.Var x) -> f e (Why.LVar x)
  | (Why.Cte n) -> f e (Why.LConst n)
(*i  | (Why.Deref x) -> f e (Why.LVar x) *)
  | _ -> Why.Let(v,e,f e (Why.LVar v))

let tr_bin_op loc s e1 e2 =
  let s' = 
    match s with
      | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" -> s
      | "add_float" | "sub_float" | "mul_float" | "div_float" -> s
      | "le_int" | "ge_int" | "gt_int" | "lt_int" -> s
      | "le_float" | "ge_float" | "gt_float" | "lt_float" -> s
      | "||" | "&&" -> s
      | "land" -> "Zland"
      | "lor" -> "Zlor"
      | "lsr" -> "Zlsr"
      | "lsl" -> "Zlsl"
      | "asr" -> "Zasr"
      | "eq_int" -> s
      | "eq_float" -> s
      | "ne_int" -> "neq_int"
      | "ne_float" -> "neq_float"
      | "ne_value" | "eq_value" -> s
      | _ -> unsupported loc ("tr_bin_op(" ^ s ^ ")")
  in
  build_app (Why.Var(s')) [e1;e2]
;;

let tr_un_op loc s e =
  let s' = 
   match s with
    | "neg_int" -> s
    | "neg_float" -> s
    | "float_of_int" -> s
    | _ -> unsupported loc ("tr_un_op(" ^ s ^ ")")
  in
  build_app (Why.Var(s')) [e]
;;

let why_one = Why.Cte(Why.Prim_int(1));;

let coercion_term t e =
  match t with
    | Integer_type -> Why.LApp("int_val",[e])
    | Boolean_type -> Why.LApp("bool_val",[e])
    | Float_type -> Why.LApp("float_val",[e])
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_term(null)"
;;

let coercion_expr t e =
  match t with
    | Integer_type -> Why.App(Why.Var("int_val"),e)
    | Boolean_type -> Why.App(Why.Var("bool_val"),e)
    | Float_type -> Why.App(Why.Var("float_val"),e)
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_expr(null)"
;;


(* translation of cast (pas au point !)*)

let rec tr_cast t e = 
  match t with
    | Base_type(s) -> tr_expr e (* TODO ! *)	
    | Type_name(qi) ->
	begin
	  match qi with
	    | [(loc,id)] -> 
		let result_type = Why.Var(id) (* correct ?? *) in
		Why.App(Why.App(Why.Var("cast"),result_type),tr_expr e)
	    | _ ->unsupported e.expr_loc "tr_cast(longname)"
	end
    | Array_type_expr(t) -> unsupported e.expr_loc "tr_cast(array)"

and tr_field_access loc f =
  match f with
    | Super_access(id) -> unsupported loc "tr_field_access(super)"
    | Primary_access(e,id) -> ill_typed "tr_field_access(primary)"
    | Typed_primary_access(e,ive) ->
	begin
	  match e.expr_node with
	    | Static_interface(te) ->
		Why.Var(te.interface_entry_name ^ "_" ^ 
			ive.instance_variable_entry_name)
	    | Static_class(te) ->
		Why.Var(te.class_entry_name ^ "_" ^ 
			ive.instance_variable_entry_name)
	    | _ -> 
		if !Config.use_value_model then
		  Why.App(Why.Var("access_"^ive.instance_variable_entry_name),
			  tr_expr e)
		else
                  why_let_var "krak_access"
		    (fun whye loge -> 
		       let pre =
			 Why.LPred("is_value_cell",[loge;
							    Why.LApp("Field",
					[Why.LVar(ive.instance_variable_entry_name)])])
		       in
		       let acc = Why.App
				   (Why.App
				      (Why.App
					 (Why.Var("access"),
					  Why.Deref("heap")),
					 whye),
				      (Why.App(Why.Var("Field"),
					       Why.Var(ive.instance_variable_entry_name))))
		       in
			 Why.Triple(pre,coercion_expr ive.instance_variable_entry_type acc
				      ,Why.LTrue,None))
		    (tr_expr e)
	end
    | Array_length(e) ->
	Why.App(Why.Var("arraylength"),tr_expr e)
    | _ -> failwith "TODO"

and tr_lexpr e =
  match e.expr_node with
    | Var(x) -> ill_typed "tr_lexpr(var)"
    | Field_access _ -> assert false
    | Typed_var(lve) -> 
	if lve.local_variable_entry_is_ref
	then lve.local_variable_entry_name 
	else 
	  begin
	    (* should never happen ! *)
	    if !Config.use_value_model then
	      Format.printf "Warning: variable %s is not supposed to be modifiable.@." lve.local_variable_entry_name;
	    lve.local_variable_entry_name 
	  end
    | This -> ill_typed "tr_lexpr(this)"
    | _ -> unsupported e.expr_loc "tr_lexpr()"

and tr_field_assignment loc f ive accu = 
  if  !Config.use_value_model then
  match f.expr_node with
    | Var(x) -> ill_typed "tr_field_assignment(var)"
    | This -> ill_typed "tr_field_assignment(this)"
    | Typed_var(lve) -> 
	if not lve.local_variable_entry_is_ref
	then 
	    (* should never happen ! *)
	    Format.printf "Warning: variable %s is not supposed to be modifiable.@." lve.local_variable_entry_name;
	let v = lve.local_variable_entry_name in
	  Why.Assign
	    (v, Why.App
	       (Why.App(Why.Var("update_" ^ ive.instance_variable_entry_name), 
			Why.Deref(v)),
		accu))
    | Field_access fa -> 
	begin
	  match fa with
	    | Super_access(_) -> unsupported loc "tr_field_assignment(super)"
	    | Primary_access(f,_) -> ill_typed "tr_field_assignment(primary)"
	    | Typed_primary_access(f,_) -> 
		unsupported loc "tr_field_assignment(primary)"
	    | Array_length(_) -> unsupported loc "tr_field_assignment(length)"
	    | _ -> failwith "TODO"
	end
    | _ -> unsupported loc "tr_field_assignment()"
  else 
    why_let_var "krak_update"
      (fun whyf logf -> 
	 why_let_var "krak_acc"
	   (fun whyaccu logaccu ->
	      let update,laccu = update_coerce ive.instance_variable_entry_type logaccu in
	      let pre =
		Why.LPred("is_valid_update",[logf;
						   Why.LApp("Field",[Why.LVar(ive.instance_variable_entry_name)]);
						   laccu])
	      in Why.Triple(pre,
			    Why.Assign
			      ("heap", 
			       Why.App
				 (Why.App
				    (Why.App
				       (Why.App(Why.Var(update), 
						Why.Deref("heap")),
						   whyf),
						  (Why.App(Why.Var("Field"),
							   Why.Var(ive.instance_variable_entry_name)))),
				    whyaccu))
			      ,Why.LTrue,None))
	      accu)
      (tr_expr f)

and tr_array_assignment t e1 e2 accu = 
  if !Config.use_value_model then 
let update = match t with
  | Integer_type -> "int_update"
  | Boolean_type -> "bool_update"
  | Float_type -> "float_update"
  | Null_type -> ill_typed "tr_field_assignment(null)"
  | Class_type(_) 
  | Array_type(_)
  | Interface_type(_) ->  "update"
in  match e1.expr_node with
    | Var(x) -> ill_typed "(?)"
    | This -> ill_typed "(?)"
    | Typed_var(lve) -> 
	if not lve.local_variable_entry_is_ref
	then 
	  begin(* should never happen ! *)
	    if !Config.use_value_model then
	      Format.printf "Warning: variable %s is not supposed to be modifiable.@." lve.local_variable_entry_name;
	  end;
	let v = lve.local_variable_entry_name in
	let upd =
	    Why.App(Why.Var(update),Why.Deref(v)) 
	in
	  Why.Assign(v,Why.App(Why.App(upd, tr_expr e2),accu))	
    | Field_access fa -> 
	begin
	  let accu = 
	      build_app	(Why.Var(update))
		[tr_expr e1 ;tr_expr e2; accu]
	  in
	  match fa with
	    | Super_access(_) -> 
		unsupported e1.expr_loc "tr_array_assignment(super)"
	    | Primary_access(f,(loc,id)) -> 
		ill_typed "tr_array_assignment(primary)"
	    | Typed_primary_access(f,ive) -> 
		  tr_field_assignment e1.expr_loc f ive accu
	    | Array_length(_) -> unsupported e1.expr_loc "tr_array_assignment(length)"
	    | _ -> failwith "TODO"
	end
    | _ -> unsupported e1.expr_loc "tr_array_assignment()"

  else 
    why_let_var "krak_update" 
      (fun whye1 loge1 ->
	 why_let_var "krak_index" 
	   (fun whye2 loge2 ->
	      why_let_var "krak_acc"
		(fun whyaccu logaccu -> 
		   let update,laccu = update_coerce t logaccu 
		   in let pre =
		       Why.LPred("is_valid_update",[loge1; Why.LApp("Index",[loge2]);laccu])
		   in         
		     Why.Triple(pre,
			     Why.Assign
			       ("heap", 
				Why.App
				  (Why.App
				     (Why.App
					(Why.App(Why.Var(update), 
						 Why.Deref("heap")),
					 whye1),
					(Why.App(Why.Var("Index"),
						 whye2))),
				     whyaccu))
			       ,Why.LTrue,None))
		accu)
	   (tr_expr e2))
      (tr_expr e1)

and tr_method_call e id args =
  match e.expr_node with
    | Var(x) -> ill_typed "(?)"
    | This -> ill_typed "(?)"
    | Typed_var(lve) -> 
	begin
	  let v = lve.local_variable_entry_name in
	  match lve.local_variable_entry_type with
            | Class_type(ce) ->
		build_app 
		  (Why.App
		     (Why.Var(ce.class_entry_name ^ "_" ^ id ^ "_ext"),
		      Why.Var(v)))
		  args
	    | _ -> unsupported e.expr_loc "tr_method_call(var)"
	end
    | Static_class(ce) ->
	build_app 
	  (Why.Var(ce.class_entry_name ^ "_" ^ id ^ "_ext")) 
	  args
    | _ -> unsupported e.expr_loc "tr_method_call()"

and tr_expr e =
  match e.expr_node with
    | Lit(l) -> tr_literal e.expr_loc l
    | Bin(e1,op,e2) ->
	begin
	  match op with
	    | "&&" -> Why.And(tr_expr e1,tr_expr e2)
	    | "||" -> Why.Or(tr_expr e1,tr_expr e2)
	    | _ ->
		tr_bin_op e.expr_loc op (tr_expr e1) (tr_expr e2)
	end
    | Un(op,e1) ->
	begin
	  match op with
	    | "pre++" -> 
		let v = tr_lexpr e1 in
		Why.Block
		  ([Why.Assign
		      (v,tr_bin_op e.expr_loc "add_int" (Why.Deref(v)) why_one) ;
		    Why.Var(v)])
	    | "post++" -> 
		let v = tr_lexpr e1 in
		Why.Let("temp",Why.Var(v),
			Why.Block
			  ([
			     Why.Assign
			      (v,tr_bin_op e.expr_loc 
				 "add_int" (Why.Deref(v)) why_one) ;
			     Why.Var("temp")]))
	    | "pre--" -> 
		let v = tr_lexpr e1 in
		Why.Block
		  ([Why.Assign
		      (v,tr_bin_op e.expr_loc 
			 "sub_int" (Why.Deref(v)) why_one) ;
		    Why.Var(v)])
	    | "post--" -> 
		let v = tr_lexpr e1 in
		Why.Let("temp",Why.Var(v),
			Why.Block
			  ([
			     Why.Assign
			      (v,tr_bin_op  e.expr_loc 
				 "sub_int" (Why.Deref(v)) why_one) ;
			     Why.Var("temp")]))
	    | _ -> tr_un_op e.expr_loc op (tr_expr e1)
	end
    | Var(_) -> ill_typed "tr_expr(var)"
    | Typed_var(lve) -> 
	if lve.local_variable_entry_is_ref
	then Why.Deref(lve.local_variable_entry_name)
	else Why.Var(lve.local_variable_entry_name)
    | Static_class(_) -> unsupported e.expr_loc "tr_expr(static_class)"
    | Static_interface(_) -> unsupported e.expr_loc "tr_expr(static_interface)"
    | This -> ill_typed "tr_expr(this)" 
    | Assign_name(_,_,_) -> ill_typed "tr_expr(assign_name)"
    | Assign_local_var(lve,op,e1) ->
	begin
	  match op with
	    | "=" ->
		unsupported e.expr_loc "tr_expr(=)"
		  (* Why.Assign(lve.local_variable_entry_name,tr_expr e) *)
		  (* WRONG : should return the value assigned *)
	    | _ -> unsupported e.expr_loc ("tr_expr("^op^")")
	end
    | Assign_field(fa,op,e1) ->
	begin
	  match fa with
	    | Super_access(id) -> 
		unsupported e.expr_loc "tr_expr(assign_field_super)"
	    | Primary_access(f,_) ->
		ill_typed "tr_expr(assign_field_primary)"
	    | Typed_primary_access(f,ive) ->
		unsupported e.expr_loc 
		  (* WRONG : should return the value assigned *)
		  "tr_expr(assign_field_primary)"
		  (* Why.Let("temp_for_update",
		     tr_expr f,
		     Why.App(Why.App(Why.Var("update_" ^ id), 
		     Why.Var("temp_for_update")),
		     tr_expr e)) *)
	    | Array_length(e2) -> 
		ill_typed "tr_expr(assign_field_array_length)"
	    | _ -> failwith "TODO"
	end
    | Assign_array(e1,e2,op,e3) -> ill_typed "tr_expr(assign_array)"
    | Typed_assign_array(t,e1,e2,op,e3) -> 
	unsupported e.expr_loc "tr_expr(array_assign)"
    | If_expr(e1,e2,e3) -> 
	Why.If(tr_expr e1,tr_expr e2,tr_expr e3)
    | Field_access(fa) -> tr_field_access e.expr_loc fa
    | Method_call(e,(_,id),args) ->
	failwith "TODO"
	  (*i tr_method_call e id (List.map tr_expr args) i*)
    | Array_creation(a) -> tr_array_creation a
    | Array_access(_) -> ill_typed "tr_expr(array_access)"
    | Typed_array_access(t,e1,e2) -> 	
	  if !Config.use_value_model then
	    coercion_expr t (build_app (Why.Var("array_access")) [tr_expr e1 ; tr_expr e2])
	  else
	    why_let_var "krak_access"
	      (fun whye1 loge1 ->
		 why_let_var "krak_index"
		   (fun whye2 loge2 ->
		      let pre = Why.LPred("is_value_cell",
						  [loge1;Why.LApp("Index",[loge2])])
		      and acc = Why.App
				  (Why.App
				       (Why.App
					  (Why.Var("access"),Why.Deref("heap")), whye1),
				       (Why.App(Why.Var("Index"),whye2)))
		      in Why.Triple(pre,coercion_expr t acc,Why.LTrue,None))
		   (tr_expr e2))
	      (tr_expr e1)
    | Cast(t,e) -> tr_cast t e
    | Class_instance_creation(_,_) -> 
	ill_typed "tr_expr(new class)"
    | Typed_class_instance_creation(ce,args) -> 
	(* let new_val = (new_obj heap classid) in
           heap := (store_new_ref heap new_val);
           (constr_classId new_val args)
        *)
	let new_val = "krak_new" in
	Why.Let
	  (new_val,
	   Why.App
	     (Why.App(Why.Var("new_obj"), Why.Deref("heap")),
	      Why.Var(ce.constructor_entry_class.class_entry_name)),
	   Why.Block
	     ([Why.Assign
		 ("heap",
		  Why.App
		    (Why.App(Why.Var("store_new_ref"), Why.Deref("heap")),
		     Why.Var(new_val))) ;
	       build_app
		 (Why.Var(ce.constructor_entry_trans_name))
		 (List.map tr_expr args)]))
    | Instanceof(e,t) -> ill_typed "(instanceof)"
    | Typed_instanceof(e1,t) -> 
	unsupported e.expr_loc "tr_expr(instanceof)"
	  (*
	    Why.App(Why.App(Why.Var("instanceof"),tr_expr e),tr_type_in_coq t)
	  *)
    | Result -> assert false (* should never happen *)
    | Old(_) -> assert false (* should never happen *)
    | Quantifier(_,_,_,_,_) -> assert false (* should never happen *)
    | Typed_quantifier(_,_,_,_) -> assert false (* should never happen *)
    | _ -> failwith "TODO"

and tr_array_creation a =
  match a with
    | Explicit_array_creation(e,a) ->
	if !Config.use_value_model then
	  build_app (Why.Var("build_array")) [ tr_expr e ; tr_array_creation a ]
	else
	  (*  il faudrait
	      let new_val = new_array heap length type_des_cellules in	      
	      heap := (store_new_ref heap new_val);
	      for i=0 to length-1 do heap := update heap new_val i a
	      *)

	  Why.Let(
	    "length",tr_expr e,
	    Why.Let
	      ("new_val", (Why.App
			     (Why.App(Why.Var("new_array"),Why.Deref("heap")),
			      Why.Var("length"))),
	       Why.Let
		 ("new_vector", (Why.App
				   (Why.App
				      (Why.Var("new_array_map"),
				       Why.Var("length")),
				      tr_array_creation a)),
		  (Why.Block
		     [Why.Assign
			("heap",
			 Why.App
			   (Why.App
			      (Why.App(Why.Var("store_add_value"),
				       Why.Deref("heap")),
			       Why.Var("new_val")),
			      Why.Var("new_vector")));
		      Why.Var("new_val")]))))
	    
    | Implicit_array_creation(_) -> ill_typed "(?)"
    | Typed_implicit_array_creation(t) -> type_default_value true t


let tr_variable_id loc id =
  match id with
    | Simple_id s -> snd s
    | Array_id _ -> unsupported loc "tr_variable_id(array)"
;;

let tr_variable_initializer loc i =
  match i with
    | Simple_initializer(e) -> tr_expr e
    | Array_initializer _ -> unsupported loc "tr_variable_initializer(array)"
;;

let tr_local_var_declarator loc t v accu =
  let init =
    match v.variable_initializer with
      | None -> type_default_value false t
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(tr_variable_id loc v.variable_id,init,accu)

;;

let tr_local_var_declaration loc lve init acc =
  let init =
    match init with
      | None -> type_default_value false lve.local_variable_entry_type
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(lve.local_variable_entry_name,init,acc)
;;

(*

  [abrupt_termination s] is [true] if [s] always terminate abruptly,
  by a return, a break, a continue or a throw statement. It returns
  [false] if it always terminate nromally, and raises [Not_found] if
  both are possible.

*)


let rec abrupt_termination s =
  match s with
    | Skip | Expr(_) -> false
    | Var_decl(v) -> false
    | Typed_var_decl(v,_,i) -> abrupt_termination i
    | Return(e) -> true
    | Break(l) -> true
    | Continue(l) -> true
    | Label(l,s) -> abrupt_termination s
    | If_statement(e,s1,s2) ->
	let c1=abrupt_termination s1
	and c2=abrupt_termination s2
	in if c1=c2 then c1 else raise Not_found
    | While(loop_inv,e,s) -> abrupt_termination s
    | For(_,_,e,el2,s) 
    | For_decl(_,_,e,el2,s) 
    | For_typed_decl(_,_,_,e,el2,s) -> 
	abrupt_termination s
    | Switch(e,cases) -> 
	unsupported e.expr_loc "abrupt_termination(switch)"
    | Block(s) -> block_abrupt_termination s
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "abrupt_termination(synchronized)"
    | Try (_, _, _) -> 
	unsupported_no_loc "abrupt_termination(try)"
    | Do (inv,s,_) -> abrupt_termination s
    | Throw(_) ->  true
    | _ -> failwith "TODO"
  
and block_abrupt_termination s =
  match s with
    | [] -> false
    | i::l ->
	if abrupt_termination i
	then true
	else block_abrupt_termination l
;;


let rec chop_last_elem l =
  match l with
    | [] -> assert false
    | [x] -> [],x
    | h::l -> let r,x = chop_last_elem l in (h::r),x
;;

let rec make_case_expr acc l =
  match l with
    | [] -> acc
    | (Case c)::l -> 
	make_case_expr 
	  (Why.Or
	     (Why.App
		(Why.App(Why.Var("eq_int"),
			 Why.Var("temp_switch")),
		 tr_expr c),
		acc))
	  l
    | _ -> assert false
;;
		  
let rec well_formed_switch cases =
  match cases with
    | [] -> assert false
    | (case_list,b)::l ->	
	if List.mem Default case_list 
	then
	  begin
	    if l <> [] then Format.printf "WARNING : unreachable statement";
  	    ([],b)
	  end
	else
	  let e = make_case_expr (Why.Cte(Why.Prim_bool(true))) case_list in
	  let (hb,tb) = chop_last_elem b in
	  if l=[] then
	    if tb = Break("") 
	    then ([(e,hb)],[]) 
	    else ([(e,b)],[])
	  else
	    if tb = Break("") 
	    then 
	      let (l',b') = well_formed_switch l in ((e,hb)::l',b')
	    else 
	      failwith "give me a break"
;;	      




let tr_logic_bin_op loc op t1 t2 =
  match op with
    | "&&" -> unsupported loc "tr_logic_bin_op(&&)"
    | "||" -> unsupported loc "tr_logic_bin_op(||)"
    | "==" -> Why.LApp("=",[t1;t2])
    | _ -> Why.LApp(op,[t1;t2])
;;

let tr_logic_un_op loc op t =
  match op with
    | "!" -> unsupported loc "tr_logic_un_op(!)"
    | _ -> Why.LApp(op,[t])
;;

let rec tr_jml_lit loc l =
  match l with
    | Int(n) -> Why.LConst(Why.Prim_int(n))
    | Bool(b) -> Why.LConst(Why.Prim_bool(b))
    | Char(_) -> unsupported loc "tr_jml_lit(char)"
    | String(_) -> unsupported loc "tr_jml_lit(string)"
    | Float(_) -> unsupported loc "tr_jml_lit(float)"
    | Null -> Why.LVar("Null")
;;

let rec oldify t =
  match t with
    | Why.LVar(x) -> 
	if !Config.use_value_model then
	  Why.LVarAtLabel(x,"init")
	else
	  if x="heap" then Why.LVarAtLabel(x,"init") else t
	    
    | Why.LConst(_) -> t
    | Why.LApp(id,args) -> Why.LApp(id,List.map oldify args) 
    | Why.LVarAtLabel(_) -> assert false (* should never happen *)

let rec tr_term this_is_result t =
  match t.expr_node with
    | Lit(l) -> tr_jml_lit t.expr_loc l
    | Var(s) -> ill_typed ("tr_term(var "^(snd s)^" has not been typed)")
    | Typed_var(lve) ->
	let id = lve.local_variable_entry_name in
	if this_is_result && id="this"
	then Why.LVar("result")
	else Why.LVar(id)
    | Bin(t1,op,t2) ->
	tr_logic_bin_op t.expr_loc op 
	  (tr_term this_is_result t1) (tr_term this_is_result t2)
    | Un(op,t2) -> tr_logic_un_op t.expr_loc op (tr_term this_is_result t2)
    | Instanceof(_,_) -> ill_typed "tr_term(instanceof)"
    | Typed_instanceof (e, t) -> 
	(* io->ht *)
	Why.LApp("instanceof",[tr_term this_is_result e;tr_type_in_coq t])
    | Cast (t, e) -> tr_term this_is_result e	(* TODO ! *)
    | Array_access (e1, e2) -> ill_typed "tr_term(array_access)"
    | Typed_array_access (ty, e1, e2) -> 
	let acc = 	  
	  if !Config.use_value_model then
	    Why.LApp("array_access",
		     [tr_term this_is_result e1;
		      tr_term this_is_result e2])
	  else
	    Why.LApp("access",
		     [Why.LVar("heap");
		      tr_term this_is_result e1;
		      Why.LApp("Index",[tr_term this_is_result e2])])
	in coercion_term ty acc
    | Array_creation _ -> unsupported t.expr_loc "tr_term(new_array)"
    | Class_instance_creation (_, _) -> 
	ill_typed "tr_term(new_object)"
    | Typed_class_instance_creation (_, _) -> 
	unsupported t.expr_loc "tr_term(new_object)"
    | Method_call(e,(_,id),args) -> 
	failwith "TODO"
(*i
	begin
	  match e.expr_node with
	    | Static_class(ce) ->
		Why.LApp(ce.type_entry_name ^ "_" ^ id,
			 if !Config.use_value_model then
			   List.map (tr_term this_is_result) args
			 else
			   (Why.LVar("heap")::
			    List.map (tr_term this_is_result) args))
	    | Typed_var(lve) ->
		begin
		  match lve.local_variable_entry_type with
		    | Class_type(ce) ->
			Why.LApp(
			  ce.type_entry_name ^ "_" ^ id,
			  if !Config.use_value_model then
			    List.map (tr_term this_is_result) (e::args)
			  else
			    (Why.LVar("heap")::
			     List.map (tr_term this_is_result) (e::args)))
		    | _ -> ill_typed "(?)"
		end
	    | _ -> 
		unsupported e.expr_loc "tr_term(method_call(novar))"
	end
i*)

    | Field_access fa -> tr_field_access_term t.expr_loc this_is_result fa
    | If_expr (e1, e2, e3) -> 
	Why.LApp("if",[tr_term this_is_result e1;
		       tr_term this_is_result e2;
		       tr_term this_is_result e3])
    | Assign_array (_, _, _, _) -> ill_typed "(?)"
    | Typed_assign_array (_, _, _, _, _) -> 
	unsupported t.expr_loc "tr_term(array_assign)"
    | Assign_field (_, _, _) -> 
	unsupported t.expr_loc "tr_term(field_assign)"	
    | Assign_local_var (_, _, _) -> 
	unsupported t.expr_loc "tr_term(var_assign)"
    | Assign_name (_, _, _) -> ill_typed "tr_term(assign_name)"
    | Static_class _ ->  unsupported t.expr_loc "tr_term(static_class)"
    | Static_interface(ie) -> 
	unsupported t.expr_loc "tr_term(static_interface)"
    | This -> ill_typed "tr_term(this)"
    | Result -> ill_typed "tr_term(\\result)"
    | Old t' -> oldify(tr_term this_is_result t')
    | Quantifier(_) -> ill_typed "tr_term(quantifier)"
    | Typed_quantifier(_) -> 
	unsupported t.expr_loc "tr_term(quantifier)"
    | _ -> failwith "TODO"

and tr_field_access_term loc this_is_result f =
  match f with
    | Super_access(id) -> 
	unsupported loc "tr_field_access_term(super)"
    | Primary_access(_,_) ->
	ill_typed "tr_field_access_term(primary)"
    | Typed_primary_access(e,ive) ->
	begin
	  match e.expr_node with
	    | Static_interface(te) ->
		Why.LVar(te.interface_entry_name ^ "_" ^ 
			 ive.instance_variable_entry_name)
	    | Static_class(te) ->
		Why.LVar(te.class_entry_name ^ "_" ^ 
			 ive.instance_variable_entry_name)
	    | _ ->
		if !Config.use_value_model then
		  Why.LApp("access_"^ive.instance_variable_entry_name,
			   [tr_term this_is_result e])
		else
		  let acc =
		    Why.LApp
		      ("access",
		       [Why.LVar("heap") ; tr_term this_is_result e;
			Why.LApp("Field",
				 [Why.LVar(ive.instance_variable_entry_name)])])
		  in coercion_term ive.instance_variable_entry_type acc
	end
    | Array_length(e) ->
	  Why.LApp("arraylength",
		   [tr_term this_is_result e])
    | _ -> failwith "TODO"

let rec tr_assertion_rec this_is_result a =   
  match a.expr_node with
    | Bin(t1,"==>",t2) ->
	Why.make_impl 
	  (tr_assertion_rec this_is_result t1) 
	  (tr_assertion_rec this_is_result t2)
    | Bin(t1,"<==",t2) ->
	Why.make_impl 
	  (tr_assertion_rec this_is_result t1) 
	  (tr_assertion_rec this_is_result t2)
    | Bin(t1,"<==>",t2) ->
	Why.make_equiv 
	  (tr_assertion_rec this_is_result t1) 
	  (tr_assertion_rec this_is_result t2)
    | Bin(t1,"<=!=>",t2) ->
	Why.LNot(Why.make_equiv 
		   (tr_assertion_rec this_is_result t1) 
		   (tr_assertion_rec this_is_result t2))
    | Bin(t1,"&&",t2) ->
	Why.make_and 
	  (tr_assertion_rec this_is_result t1) 
	  (tr_assertion_rec this_is_result t2)
    | Bin(t1,"||",t2) ->
	Why.make_or
	  (tr_assertion_rec this_is_result t1)
	  (tr_assertion_rec this_is_result t2)
    | Bin(t1,"!",t2) ->
	Why.LNot(tr_assertion_rec this_is_result t2)
    | Lit(Bool(true)) -> Why.LTrue
    | Lit(Bool(false)) -> Why.LFalse
    | Typed_var(lve) ->
	let id = lve.local_variable_entry_name in
	let v = 
	  if this_is_result && id="this"
	  then Why.LVar("result")
	  else Why.LVar(id)
	in
	Why.LPred("eq_bool",[v;Why.LConst(Why.Prim_bool(true))])
    | Typed_quantifier(Forall,idl,e1,e2) -> 
	List.fold_right
	  (fun lve acc ->
	     let ty = tr_type_detailed lve.local_variable_entry_type in
	     Why.LForall(lve.local_variable_entry_name,ty,acc))
	  idl
	  (Why.LImpl(tr_assertion_rec this_is_result e1,
		     tr_assertion_rec this_is_result e2))
    | Typed_quantifier(Exists,idl,e1,e2) -> 
	List.fold_right
	  (fun lve acc ->
	     let ty = tr_type_detailed lve.local_variable_entry_type in
	     Why.LExists(lve.local_variable_entry_name,ty,acc))
	  idl
	  (Why.make_and (tr_assertion_rec this_is_result e1)
		     (tr_assertion_rec this_is_result e2))
    | _ -> 
	failwith "TODO"
	(* Why.Term(tr_term this_is_result a)   *)

let tr_loop_inv inv dec =
  let i = tr_assertion_rec false inv
  and d = tr_term false dec
  in 
  if !Config.use_value_model then (i,d) else 
    (i,d) (*i (Why.make_and i
       (Why.Term(Why.LApp("store_well_typed",[Why.LVar "heap"]))),d) i*)

let tr_assertion_constr_post x = 
(*
  print_endline "tr_assertion_constr_post";
  Format.printf "assertion = %a" Ast.fprintf_expr x;
*)
  tr_assertion_rec true x

let tr_assertion = tr_assertion_rec false


let tr_assign_op loc op =
match op with
  | "+=" -> "add_int"
  | _ -> unsupported loc ("tr_assign_op("^op^")")
;;

let rec tr_statement s = 
  tempvar_count := 0;
  match s with
    | Skip -> Why.Void
    | Expr(e) -> 
	begin
	  match e.expr_node with
	    | Lit(_) -> ill_typed "(?)"
	    | Bin(_,_,_) ->  ill_typed "(?)"
	    | Un(op,e) ->
		begin
		  match op with
		    | "pre++" | "post++" -> 
			let v = tr_lexpr e in
			Why.Assign
			  (v,tr_bin_op e.expr_loc "add_int" (Why.Deref(v)) why_one)
		    | "pre--" | "post--" -> 
			let v = tr_lexpr e in
			Why.Assign
			  (v,tr_bin_op e.expr_loc "sub_int" (Why.Deref(v)) why_one) 
		    | _ -> ill_typed "(?)"
		end
	    | Var(_) -> ill_typed "(?)"
	    | Typed_var(lve) -> ill_typed "(?)"
	    | Static_class(_) -> assert false (* ????????? *)
	    | Static_interface(_) -> assert false (* ????????? *)
	    | This -> ill_typed "(?)"
	    | Assign_name(_,_,_) -> ill_typed "(?)"
	    | Assign_local_var(lve,op,e1) ->
		begin
		  match op with
		    | "=" ->
			Why.Assign(lve.local_variable_entry_name,tr_expr e1)
		    | _ ->
			Why.Assign(lve.local_variable_entry_name,
				   Why.App(Why.App(Why.Var(tr_assign_op e1.expr_loc op),
						   Why.Var(lve.local_variable_entry_name)),tr_expr e1))
		end
	    | Assign_field(fa,op,e1) ->
		begin
		  match fa with
		    | Super_access(id) -> 
			unsupported e.expr_loc "assign_field(super)"
		    | Primary_access(f,(loc,id)) ->	
			ill_typed "assign_field(primary)"
		    | Typed_primary_access(f,ive) ->	
			begin
			  match op with
			    | "=" ->
				tr_field_assignment 
				  e.expr_loc f ive (tr_expr e1)
			    | _ -> 
				unsupported e.expr_loc 
				  ("tr_statement(assign_field("^op^"))")
			end
		    | Array_length(_) -> ill_typed "tr_statement(array_length)"
		    | _ -> failwith "TODO"
		end
	    | Assign_array(e1,e2,op,e3) -> 
		ill_typed "tr_statement(assign_array)"
	    | Typed_assign_array(t,e1,e2,op,e3) -> 
		begin
		  match op with
		  | "=" ->
		      tr_array_assignment t e1 e2 (tr_expr e3)
		  | _ -> 
		      unsupported e.expr_loc 
			("tr_statement(assign_array("^op^"))")
		end			
	    | If_expr(e1,e2,e3) -> ill_typed "tr_statement(if_expr)"
	    | Field_access(fa) -> ill_typed "tr_statement(field_access)"
(*
	    | Method_call(Static_class(ce),(_,id),args) ->
		build_app 
		  (Why.Var("static_method_" ^ (snd ce.class_entry_name) ^ "_" ^ id)) args
	    | Method_call(e,(_,id),args) ->
		Why.Let
		  ("temp_for_method_call",tr_expr e,
		   build_app 
		     (Why.App
			(Why.Var("method_" ^ id ^ "_ext"),
			 Why.Var("temp_for_method_call")))
		     args)
*)
	    | Method_call(e,(_,id),args) -> 
		failwith "TODO"
		  (*i tr_method_call e id (List.map tr_expr args) i*)
	    | Array_creation(a) -> ill_typed "tr_statement(array_creation)"
	    | Array_access(e1,e2) -> ill_typed "tr_statement(array_access)"
	    | Typed_array_access(t, e1,e2) -> ill_typed "tr_statement(typed_array_access)"
	    | Cast(t,e) -> ill_typed "tr_statement(cast)"
	    | Class_instance_creation(_,_) -> ill_typed "tr_statement(new_object)"
	    | Typed_class_instance_creation(ce,args) -> 
		unsupported e.expr_loc "tr_statement(typed_new_object)"
	    | Instanceof(e,t) -> ill_typed "tr_statement(instanceof)"
	    | Typed_instanceof(e,t) -> ill_typed "tr_statement(instanceof result)"
	    | Result -> assert false (* should never happen *)
	    | Old(_) -> assert false (* should never happen *)
	    | Quantifier(_) -> assert false (* should never happen *)
	    | Typed_quantifier(_) -> assert false (* should never happen *)
	    | _ -> failwith "TODO"
	end
    | Var_decl(v) -> ill_typed "tr_statement(var_decl)"
    | Typed_var_decl(v,init,i) -> 
	tr_local_var_declaration Location.dummy v init (tr_statement i)
    | Return(e) -> 
	begin
	  match e with
	    | None -> Why.Void (* return; *) 
	    | Some e -> tr_expr e 
	end
    | Break(l) -> unsupported Location.dummy "tr_statement(break)"
    | Continue(l) -> unsupported Location.dummy "tr_statement(continue)"
    | Label(l,s) -> unsupported Location.dummy "tr_statement(label)"
    | If_statement(e,s1,s2) ->
	Why.If(tr_expr e, tr_statement s1, tr_statement s2)
    | While(l,e,s) ->
	let (inv,var) = tr_loop_inv l.loop_invariant l.loop_decreases in
	Why.make_while (tr_expr e) inv var (tr_statement s)
    | For(l,el1,e,el2,s) -> 
	let (inv,var) = tr_loop_inv l.loop_invariant l.loop_decreases in
	Why.append
	  (tr_block el1)
	  (Why.make_while (tr_expr e) inv var 
	     (tr_block (s::el2)))
    | For_decl(_) -> ill_typed "(?)"
    | For_typed_decl(lve,lvi,l,e,el2,s) -> 
	let (inv,var) = tr_loop_inv l.loop_invariant l.loop_decreases in
	tr_local_var_declaration Location.dummy lve lvi
	  (Why.make_while (tr_expr e) inv var 
	     (tr_block (s::el2)))
    | Switch(e,cases) -> 
	begin
	  try
	    let (l,b) = well_formed_switch cases in
	    Why.Let("temp_switch",
		    tr_expr e,
		    List.fold_right
		      (fun (c,b) acc -> Why.If(c,tr_block b,acc))
		      l (tr_block b))
	  with
	      Failure msg -> 
		unsupported e.expr_loc 
		  ("tr_statement(switch not well formed: " ^ msg ^ ")") 
	end
    | Block(s) -> tr_block s
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "tr_statement(synchronized)"
    | Try (_, _, _) -> unsupported Location.dummy "tr_statement(try)"
    | Do (inv,_, e) -> unsupported e.expr_loc "tr_statement(do)"
    | Throw(e) ->  unsupported e.expr_loc "tr_statement(throw)"
    | _ -> failwith "TODO"

and tr_begin_end i l =
  Why.Block(List.map (tr_statement) (i::l))

and tr_block (l : statement list) : Why.expr = 
  match l with
    | [] -> Why.Void
    | [i] -> tr_statement i
    | i::l -> 
	match i with
	  | Skip -> tr_block l
	  | If_statement(e,s1,s2) ->
	      begin
		try
		  match (abrupt_termination s1,abrupt_termination s2) with
		    | (false,false) -> 
			tr_begin_end i l
		    | (true,true) ->
			Format.printf "WARNING : unreachable statement";
			Why.If
			  (tr_expr e,
			   tr_statement s1,tr_statement s2)
		    | (false,true) ->
			Why.If
			  (tr_expr e,
			   tr_block (s1::l), tr_statement s2)
		    | (true,false) ->
			Why.If
			  (tr_expr e,
			   tr_statement s1,tr_block (s2::l))
		with
		    Not_found ->
		      Format.printf "This statement needs a complex control-flow@.";
		      unsupported Location.dummy "tr_block(if-complex)"
	      end
	  | Typed_var_decl(lve,init,i) -> 
	      tr_statement (Typed_var_decl(lve,init,Block(i::l)))
	      (*
	      Format.printf "***Assert false in tr_block: %a@."
		(Format_misc.fprintf_list Ast.fprintf_statement) (i::l);
	      assert false
	      *)
	  | _ -> tr_begin_end i l
   
;;
  
let tr_parameter p = 
  match p with
    | Local_variable_entry le ->
	let t = tr_type le.local_variable_entry_type in
	(le.local_variable_entry_name,
	 if !Config.use_value_model then
	   if le.local_variable_entry_is_ref then Why.Ref_type(t) else t
	 else t)
    | _ -> ill_typed "(?)"
;;

let rec params_type params return_type = 
  match params with
    | [] -> return_type
    | (id,t)::l ->
	Why.Prod_type(id,t,params_type l return_type)
;;

let arg_pre x t = 
  Why.LPred(t ^ "_invariant", 
	      if !Config.use_value_model then
		[Why.LVar(x)]
	      else
		[Why.LVar("heap");Why.LVar(x)]
	   )
;;

let tr_signals_clause s =
  match s with
    | None -> None
    | Some(t,var,e) ->
	assert (match var with 
		  | None -> Format.eprintf "None!@.";false 
		  | Some v -> 
		      (v.local_variable_entry_name = "result") ||
		      (Format.eprintf "Some %s@." v.local_variable_entry_name;
		       false));
	Some(tr_type_detailed t, tr_assertion e)
;;

let rec args_pre p l =
  match l with
    | [] -> p
    | (x,xe)::r ->
	let t = 
	  match xe with
	    | Local_variable_entry le -> le.local_variable_entry_type
	    | _ -> ill_typed "(?)"
	in
	match t with
	  | Class_type ce ->
	      args_pre 
		(Why.make_and p (arg_pre x ce.class_entry_name))
		r
	  | Array_type _ ->
	      args_pre
		(Why.make_and 
		   p 
		   (if !Config.use_value_model then
		      Why.LPred("isarray", [Why.LVar(x)])
		    else
		      Why.LPred
			   (*i former has_type i*)
			   ("instanceof", 
			    [Why.LVar(x);
			     tr_type_in_coq t]))
		   )
		r
	  | _ -> args_pre p r

let tr_method_declaration spec_only class_name class_invariant jml me body =
  let routine = me.method_entry_routine in
  let l = routine.routine_entry_parameters in
  let params =
    if me.method_entry_is_static
    then
      if l=[]
      then ["tt",Why.unit_type]
      else List.map (fun (_,pe) -> tr_parameter pe) l
    else
      ("this",
       Why.base_type
	 (if !Config.use_value_model then
	    if me.method_entry_is_pure then "value" else "value ref"
	  else "value"))::
      (List.map (fun (_,pe) -> tr_parameter pe) l)
  in
  let (jml_pre,jml_post,jml_signals) = 
    match jml with
      | Typed_compact tcc ->
	  let requires = tr_assertion tcc.typed_compact_requires
	  and ensures = tr_assertion tcc.typed_compact_ensures
	  and signals = tr_signals_clause tcc.typed_compact_signals
	  in      
	    (requires, ensures, signals)
	  | _ -> ill_typed "JML Desugaring failed"
  in
  let (untyped_pre,untyped_post,untyped_signals) =    
    if !Config.use_value_model then
      if me.method_entry_is_static
      then (args_pre jml_pre l,jml_post,jml_signals)
      else ((args_pre 
	       (Why.make_and jml_pre (arg_pre "this" class_name)) l),
	    (if me.method_entry_is_pure then jml_post else
	       Why.make_and jml_post (arg_pre "this" class_name)),
	    (if me.method_entry_is_pure then jml_signals else
	       Option.map 
		 (fun (e,p) -> 
		    (e,Why.make_and p (arg_pre "this" class_name)))
		 jml_signals))
    else
      if me.method_entry_is_static
      then (args_pre jml_pre l,jml_post,jml_signals)
      else ((args_pre 
	       (Why.make_and jml_pre 
		  (Why.make_and 
		     (Why.LPred("ne_value",[Why.LVar "this";Why.LVar "Null"]))
		     (arg_pre "this" class_name))) l),
	    (if me.method_entry_is_pure then jml_post else
	       Why.make_and jml_post (arg_pre "this" class_name)),
	    (if me.method_entry_is_pure then jml_signals else
	       Option.map 
		 (fun (e,p) -> 
		    (e,Why.make_and p (arg_pre "this" class_name)))
		 jml_signals))
  in
  let (pre,post,signals) =
    if !Config.use_value_model then (untyped_pre,untyped_post,untyped_signals)
    else (untyped_pre,untyped_post,untyped_signals)
(*i    else
    (Why.make_and untyped_pre (Why.Term(Why.LApp("store_well_typed",[Why.LVar "heap"]))),
     Why.make_and untyped_post (Why.Term(Why.LApp("store_well_typed",[Why.LVar "heap"]))),
     Option.map
      (fun (e,p) -> 
	 (e,Why.make_and p (Why.Term(Why.LApp("store_well_typed",[Why.LVar "heap"])))))
    untyped_signals)
i*)
  in
  let return_type = tr_type me.method_entry_return_type in
  let why_type = params_type params return_type 
  and why_annot_type = 
    params_type params 
      (Why.Annot_type(pre,return_type,["heap"],
		      (if me.method_entry_is_pure then [] else ["heap"]),
		      post,signals))
  in
  let why_decls =
    let extern_decl =
      Why.External
	(class_name ^ "_" ^ me.method_entry_name ^ "_ext", why_annot_type)
    in
    if me.method_entry_is_pure 
    then 
      (*
	if return_type = Why.Base_type("bool")
      then
	let logic_bool_decl =
	  Why.Logic
	    (class_name ^ "_" ^ me.method_entry_name ^ "_bool" , 
	     why_type)	  
	and logic_prop_decl =
	  Why.Logic
	    (class_name ^ "_" ^ me.method_entry_name, 
	     params_type params (Why.Base_type "prop"))	  
	in [logic_prop_decl;logic_bool_decl;extern_decl]
      else 
	*)
	let logic_decl =
	  Why.Logic
	    (class_name ^ "_" ^ me.method_entry_name, 
	     if !Config.use_value_model then
	       why_type
	     else Why.Prod_type("",Why.base_type("store"),why_type))	  
	in [logic_decl;extern_decl]
    else
      [extern_decl]
  and coq_decls =
    if me.method_entry_is_pure 
    then 
      (*
      if return_type = Why.Base_type("bool")
      then
	let parameter_decl =
	  Why.Parameter
	    (class_name ^ "_" ^ me.method_entry_name, 
	     params_type params (Why.Base_type "Prop"))
	and axiom_decl =
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_axiom", 
	     params_type params 
	       (Why.Annot_type
		  (pre,return_type,
		   Why.LLet
		     ("result",
		      Why.LApp 
			(class_name ^ "_" ^ me.method_entry_name,
			 if !Config.use_value_model then 
			   List.map (fun (x,_) -> Why.LVar x) params
			 else
			   (Why.LVar "heap"::
			    (List.map (fun (x,_) -> Why.LVar x)) params)),
			post))))
	and parameter_bool_decl =
	  Why.Parameter
	    (class_name ^ "_" ^ me.method_entry_name ^ "_bool", 
	     why_type)
	and axiom_bool_decl =
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_bool_axiom", 
	     params_type params 
	       (Why.Annot_type
		  (pre,return_type,
		   Why.LLet
		     ("result",
		      Why.LApp 
			(class_name ^ "_" ^ me.method_entry_name,
			 List.map (fun (x,_) -> Why.LVar x)params),
			post))))
	in      
	[parameter_decl;axiom_decl;parameter_bool_decl;axiom_bool_decl]
	
      else
      *) 
	let parameter_decl =
	  Why.Parameter
	    (class_name ^ "_" ^ me.method_entry_name, 
	     if !Config.use_value_model then
	       why_type
	     else Why.Prod_type("",Why.base_type("store"),why_type))
	and axiom_decl =
	  let coq_type = 
	    params_type params 
	      (Why.Annot_type
		 (pre,return_type,[],[],
		  Why.LLet
		    ("result",
		     Why.LApp 
		       (class_name ^ "_" ^ me.method_entry_name,
			if !Config.use_value_model then 
			   List.map (fun (x,_) -> Why.LVar x) params
			 else
			   (Why.LVar "heap"::
			    (List.map (fun (x,_) -> Why.LVar x)) params)),
		       untyped_post),None))
	  in
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_axiom", 
	     if !Config.use_value_model then coq_type
	     else Why.Prod_type("heap",Why.base_type("store"),coq_type))
	in      
	[parameter_decl;axiom_decl]
    else []
  in
  if spec_only then  why_decls,coq_decls
  else
    match body with
      | Some l ->
	  begin
	    try
	      ((Why.Def
		  (class_name ^ "_" ^ me.method_entry_name, 
		   Why.Fun
		     (params,pre,tr_block l,post,None))
		::why_decls),coq_decls)
	    with
	      | Unsupported msg ->
		  Format.printf 
		    "**Warning: translation of body of %s not performed due to unsupported feature in translation: %s@."  
		    me.method_entry_name msg;
		  (why_decls,coq_decls)
	  end
      | None -> why_decls,coq_decls
;;

let tr_constr_declaration spec_only class_name class_invariant jml ce body =
  let routine = ce.constructor_entry_routine in
  let l = routine.routine_entry_parameters in
  let params = 
    if l=[]
    then ["tt",Why.unit_type]
    else List.map (fun (_,pe) -> tr_parameter pe) l
  in
  let (jml_pre,jml_post) = 
    match jml with
  | Typed_compact tcc ->
	let requires = tr_assertion tcc.typed_compact_requires
	and ensures = tr_assertion tcc.typed_compact_ensures
(* i
      Compact(compact) ->
	let requires = 
	  compact.compact_requires
	and ensures = 
	  compact.compact_ensures 
i*)
	in      
	((*tr_assertion*) ((*List.hd*) requires),
	 (*tr_assertion*) ((*List.hd*) ensures))
(*i      |Non_extending _ -> ill_typed "###### Non extending"
      |Extending_also _ ->ill_typed"###### Also"
      |Extending_and _  ->ill_typed"###### And"
      |Typed_compact _  ->ill_typed"###### Typed" i*)
    | _ -> ill_typed "JML Desugaring failed"
  in
  let (pre,post) =    
    (args_pre jml_pre l,Why.make_and jml_post (arg_pre "result" class_name))
  in
  let return_type = Why.base_type("value (*" ^ class_name ^ "*)") in
  let why_type = params_type params return_type in
  let extern_decl =
    Why.External
      (ce.constructor_entry_trans_name, why_type) 
  in  
  if spec_only then [extern_decl], []
  else
    let tr_body = 
      Why.Let_ref("this",
		  Why.Var("create_" ^ class_name),
		  Why.append (tr_block body) (Why.Deref("this"))) 
    in
    ([Why.Def("new_" ^ ce.constructor_entry_class.class_entry_name, 
	      Why.Fun(params,pre,tr_body,post,None));
      extern_decl],[])
;;

let tr_instance_variable_declaration class_name ve = 
  if !Config.verbose >=2 then
    Format.printf "producing declaration of field %s.@."
      ve.instance_variable_entry_name;
  ([
     if !Config.use_value_model then	 
       Why.External
	 ("access_" ^ ve.instance_variable_entry_name , 	     
	  Why.Prod_type
	    ("",Why.base_type("value (*" ^ class_name ^ "*)"),
	     tr_type ve.instance_variable_entry_type))
     else 
       Why.External
	 (ve.instance_variable_entry_name,
	  Why.base_type("fieldId"))
	 ;
     if !Config.use_value_model then	 
       Why.Logic
	 ("access_" ^ ve.instance_variable_entry_name , 	     
	  Why.Prod_type
	    ("",Why.base_type("value (*" ^ class_name ^ "*)"),
	     tr_type ve.instance_variable_entry_type))
     else
       Why.Logic
	 (ve.instance_variable_entry_name, 	     
	  Why.base_type("fieldId"))
   ]
   @
   (if !Config.use_value_model then
     [Why.External
       ("update_" ^ ve.instance_variable_entry_name , 	     
	Why.Prod_type
	  ("",Why.base_type("value (*" ^ class_name ^ "*)"),
	   Why.Prod_type
	     ("",tr_type ve.instance_variable_entry_type,
	      (Why.base_type("value (*" ^ class_name ^ "*)")))))]
   else [])
     ,
     [])
;;


let tr_static_variable_declaration loc v init =
  let i = 
    match init with
    | None -> type_default_value false v.instance_variable_entry_type
    | Some e -> tr_variable_initializer loc e
  in
  ([Why.Logic
      (v.instance_variable_entry_name,
       tr_type v.instance_variable_entry_type);
    Why.External
      (v.instance_variable_entry_name,
       tr_type v.instance_variable_entry_type)],
   [Why.Definition(v.instance_variable_entry_name,i)])

let rec find_method_name m =
  match m with
    | Simple_method_declarator((loc,id),_) -> id
    | Array_method_declarator(m) -> find_method_name m

let name_of_field_declaration f =
  match f with 
    | Typed_method(me,jml,b) -> "method " ^ me.method_entry_name
    | Typed_constructor(ce,jml,eci,b) -> 
	"constructor " ^ ce.constructor_entry_class.class_entry_name
    | Typed_variable(ve,init) -> "field " ^ ve.instance_variable_entry_name
    | Static_initializer(_) -> "static initializer"
    | Method(md,_) -> "method " ^ find_method_name md.method_declarator
    | Constructor(cd) -> "constructor " ^ (snd cd.constr_name)
    | Variable(v) -> "field"
    | Specification(_) -> "jml specification"

let tr_field_declaration class_name class_invariant f = 
  try
    match f with 
      | Typed_method(me,jml,b) -> 
	  tr_method_declaration 
	  (not me.method_entry_routine.routine_entry_produce_body)
	  class_name class_invariant jml me b
      | Typed_constructor(ce,jml,eci,b) ->
	  begin
	    match eci with
	      | Invoke_none -> 
		  tr_constr_declaration 
		  (not (Config.mem_methods 
			  class_name ce.constructor_entry_class.class_entry_name))
		  class_name class_invariant jml ce b
	      | Invoke_this(el) -> 
		  unsupported Location.dummy
		    "tr_field_declaration(constructor_invoke(this))"
	      | Invoke_super(el) -> 
		  unsupported Location.dummy
		    "tr_field_declaration(constructor_invoke(super))"
	  end      
      | Typed_variable(ve,init) ->
	  if ve.instance_variable_entry_is_static
	  then tr_static_variable_declaration Location.dummy ve init
	  else tr_instance_variable_declaration class_name ve
      | Static_initializer _ -> 
	  unsupported Location.dummy "tr_field_declaration(static_init)"
      | Method(_,_) -> 
	  ill_typed "tr_field_declaration(method)"
      | Constructor(_) -> 
	  ill_typed "tr_field_declaration(constructor)"
      | Variable(_) -> 
	  ill_typed "tr_field_declaration(variable)"
      |	Specification(_) -> ([],[]) (* already computed before *)
  with
    | Unsupported msg ->
	Format.printf 
	  "**Warning: translation of %s not performed due to unsupported feature in translation: %s@." 
	  (name_of_field_declaration f) msg;
	([],[])
    | Ill_typed msg -> 
	Format.printf 
	  "**Warning: translation of %s not performed due to typing errors (%s)@."
	  (name_of_field_declaration f) msg;
	([],[])

let collect_class_invariant class_name class_fields =
  try
    List.fold_left
      (fun acc (doc,f) ->
	 match f with
	   | Specification s ->
	       List.fold_left
		 (fun acc s ->
		   match s with
		   | Invariant e -> Why.make_and acc (tr_assertion e)
		   | _ -> acc)
		 acc s
	   | _ -> acc)
      Why.LTrue
      class_fields
  with
    | Unsupported msg ->
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	Why.LTrue
      | Ill_typed msg -> 
	  Format.printf 
	    "**Warning: translation of class invariant of %s not performed due to typing errors (%s)@."
	    class_name msg;
	  Why.LTrue	
;;

let tr_class_declaration ce c =
  let class_name = ce.class_entry_name in
  try
    let inv = collect_class_invariant class_name c.class_fields in
    let create_decl =
      if List.mem `FINAL c.class_modifiers then ([],[])
      else
	([(*i
	    Why.External
	    ("create_" ^ class_name, 
	     Why.Base_type("value (*" ^ class_name ^ "*)"));
	    i*)
	  Why.External(class_name,Why.base_type("classId"));
	  Why.Logic(class_name,Why.base_type("classId"));
	  Why.Logic
	    (class_name ^ "_invariant",
	     if !Config.use_value_model then
	       Why.Prod_type("this",
			     Why.base_type("value (*" ^ class_name ^ "*)"),
			     Why.base_type("prop"))
	     else
	       Why.Prod_type
		 ("",Why.base_type("store"),
		  Why.Prod_type
		    ("this",
		     Why.base_type("value (*" ^ class_name ^ "*)"),
		     Why.base_type("prop")))
	    )],
	 [Why.PropDefinition
	    (class_name ^ "_invariant",	   
	     (if !Config.use_value_model then
		["this",Why.base_type("value")]
	      else
		["heap", Why.base_type("store"); 
		 "this",Why.base_type("value")]  ),
	     Why.make_and
	       (
(*io->ht	       if !Config.use_value_model then
		  Why.Term
		    (Why.LApp
		       ("instanceof", [Why.LVar("this"); Why.LVar(class_name)]))
		else

i*)
		  Why.LPred
		(*i former has_type i*)    
		       ("instanceof", 
			[Why.LVar("this"); 
			 Why.LApp("ClassType",
				  [Why.LVar(class_name)])]))
	       inv)])
    in
      List.fold_left
	(fun (wdacc,cdacc) (_,f) -> 
	   let (wd,cd) = tr_field_declaration class_name inv f
	   in (List.rev_append wd wdacc,List.rev_append cd cdacc))
	create_decl
	c.class_fields

    with
      | Unsupported msg ->
	  Format.printf 
	    "**Warning: translation of class %s not performed due to unsupported feature in translation: %s@." 
	    class_name msg;
	  ([],[])
      | Ill_typed msg -> 
	  Format.printf 
	    "**Warning: translation of class %s not performed due to typing errors (%s)@."
	    class_name msg;
	  ([],[])	
;;

let tr_interface_member_declaration interface_name f =
  try
    match f with
    | Constant_declaration(_) -> 
	ill_typed "tr_interface_member_declaration(constant)"
    | Typed_constant_declaration(ive,i) ->
	failwith "TODO"
	  (*i
	([Why.Def(ive.instance_variable_entry_name, 
		  tr_variable_initializer Location.dummy i)],[])
	    i*)
    | Method_header(_) -> 
	ill_typed "tr_interface_member_declaration(method)"
    | Typed_method_header(_) -> 
	unsupported Location.dummy "tr_interface_member_declaration(method)"
  with
  | Unsupported msg ->
      Format.printf 
	"**Warning: translation of interface member %s not performed due to unsupported feature in translation: %s@." 
	"(?)" msg;
      ([],[])
  | Ill_typed msg -> 
      Format.printf 
	"**Warning: translation of interface member %s not performed due to typing errors (%s)@."
	"(?)" msg;
      ([],[])	
;;

let tr_interface_declaration i =
  List.fold_left
    (fun (wdacc,cdacc) (_,f) -> 
       let (wd,cd) = tr_interface_member_declaration 
		       (snd i.interface_name) f
       in (List.rev_append wd wdacc,List.rev_append cd cdacc))
    ([],[])
    i.interface_members
;;

let tr_type_declaration d = 
  match d with
  | Interface _ -> assert false
  | Class _ -> assert false
  | Typed_interface(_, i) -> tr_interface_declaration i
  | Typed_class(ce,c) -> tr_class_declaration ce c
;;

let tr_compilation_unit a = 
  List.fold_left
    (fun (wdacc,cdacc) (_,d) -> 
       let (wd,cd) = tr_type_declaration d
       in (List.rev_append wd wdacc,List.rev_append cd cdacc)) 
    ([],[])
    a.cu_type_decls;;

(*
Local Variables: 
mode: tuareg
compile-command: "make"
End: 
*)

