
open Format;;

type float_repr = string * string * string (*r int, frac, exp *)

let make_float_repr i f e = (i,f,Option.fold "" (fun x -> x) e)

type constant =
  | Prim_int of int
  | Prim_real of float_repr
  | Prim_bool of bool
;;

let int_of_string s = if s="" then 0 else int_of_string s

let fprint_constant coq form e =
  match e with
    | Prim_int(n) -> 
	if coq then
	  fprintf form "%d" n
	else
	  fprintf form "%d" n
    | Prim_real(i,f,e) -> 
	let l = String.length f in
	if coq then
	  let e = int_of_string e - l in
	  if e=0
	  then
	    fprintf form "(IZR %s%s)" i f
	  else
	    fprintf form "(decimal_exp (IZR %s%s) (%d))" i f e
	else
	  if e=""
	  then fprintf form "%s.%s" i f
	  else fprintf form "%s.%se%s" i f e
    | Prim_bool(b) -> 
	(*if coq then
	  fprintf form "%s" (if b then "True" else "False")
	else
	  *)fprintf form "%b" b
;;


type term = 
  | LConst of constant
  | LApp of string * term list
  | LVar of string
  | LVarAtLabel of string * string     (*r x@L *)
;;

let coq_id s = 
  match s with
    | "eq_bool" -> "eq"
    | "gt_int" -> "Zgt"
    | "lt_int" -> "Zlt"
    | "ge_int" -> "Zge"
    | "le_int" -> "Zle"
    | "eq_int" -> "eq"
    | "neq_int" -> "Zne"
    | "add_int" -> "Zplus"
    | "sub_int" -> "Zminus"
    | "mul_int" -> "Zmult"
    | "div_int" -> "Zdiv"
    | "mod_int" -> "Zmod"
    | "neg_int" -> "Zopp"
    | "gt_real" -> "Rgt"
    | "lt_real" -> "Rlt"
    | "ge_real" -> "Rge"
    | "le_real" -> "Rle"
    | "eq_real" -> "eq" 
    | "real_of_int" -> "IZR"
    | "add_real" -> "Rplus"
    | "sub_real" -> "Rminus"
    | "mul_real" -> "Rmult"
    | "div_real" -> "Rdiv"
    | "neg_real" -> "Ropp"
    | "le_type" -> "subType"
    | "eq_type" -> "eq"
    | "eq_value" | "ne_value" -> assert false
    | "access_obj" | "access_int" | "access_bool" ->
	if !Config.use_global_memory_model
	then s else "access"
    | "array_access_obj" | "array_access_int" | "array_access_bool" ->
	if !Config.use_global_memory_model
	then s else "array_access"
    | _ -> s

let rec iter_term f t =
  match t with
  | LConst(c) -> ()
  | LApp(id,l) -> 
      if id = "array_inter_loc" || id = "array_loc" 
      then fprintf Config.log "iterating on %s@." id;
      f id; List.iter (iter_term f) l
  | LVar(id) -> f id
  | LVarAtLabel(id,l) -> 
      if id = "ISOException_systemInstance"
      then fprintf Config.log "iterating on %s@@%s@." id l;
      f id
;;

let rec fprint_term coq form t =
  match t with
  | LConst(c) -> fprint_constant coq form c
  | LApp("eq_value",[t1;t2]) ->
      fprintf form "@[(%a=%a)@]" 
	(fprint_term coq) t1
	(fprint_term coq) t2
  | LApp("ne_value",[t1;t2]) ->
      if coq then
	begin
	  fprintf form "@[(%a!=%a)@]" 
	    (fprint_term coq) t1
	    (fprint_term coq) t2
	end
      else
	begin
	  fprintf form "@[neqv(%a,%a)@]" 
	    (fprint_term coq) t1
	    (fprint_term coq) t2
	end
  | LApp(id,t::tl) ->
      if coq then
	begin
	  fprintf form "@[(%s@ %a" (coq_id id) (fprint_term coq) t;
	  List.iter (fun t -> fprintf form "@ %a" (fprint_term coq) t) tl;
	  fprintf form ")@]"
	end
      else
	begin
	  fprintf form "@[%s(%a" id (fprint_term coq) t;
	  List.iter (fun t -> fprintf form ",@ %a" (fprint_term coq) t) tl;
	  fprintf form ")@]"
	end
  | LApp(_,_) -> assert false      
  | LVar(id) -> fprintf form "%s" id
  | LVarAtLabel(id,l) -> fprintf form "%s@@%s" id l
;;

type assertion = 
  | LTrue | LFalse
  | LAnd of assertion * assertion
  | LOr of assertion * assertion
  | LNot of assertion
  | LImpl of assertion * assertion
  | LIf of term * assertion * assertion
  | LLet of string * term * assertion
      (*r warning: only for Coq assertions *)
  | LForall of string * string * assertion
      (*r forall x:t.a *)
  | LExists of string * string * assertion
      (*r exists x:t.a *)
  | LPred of string * term list
;;

let make_or a1 a2 =
  match (a1,a2) with
    | (LTrue,_) -> LTrue
    | (_,LTrue) -> LTrue
    | (LFalse,_) -> a2
    | (_,LFalse) -> a1
    | (_,_) -> LOr(a1,a2)

let make_and a1 a2 =
  match (a1,a2) with
    | (LTrue,_) -> a2
    | (_,LTrue) -> a1
    | (LFalse,_) -> LFalse
    | (_,LFalse) -> LFalse
    | (_,_) -> LAnd(a1,a2)

let rec make_and_list l =
  match l with
    | [] -> LTrue
    | f::r -> make_and f (make_and_list r)

let make_impl a1 a2 =
  match (a1,a2) with
    | (LTrue,_) -> a2
    | (_,LTrue) -> LTrue
    | (LFalse,_) -> LTrue
    | (_,LFalse) -> LNot(a1)
    | (_,_) -> LImpl(a1,a2)

let make_equiv a1 a2 =
  match (a1,a2) with
    | (LTrue,_) -> a2
    | (_,LTrue) -> a1
    | (_,_) -> LAnd(LImpl(a1,a2),LImpl(a2,a1))

let coq_base_type t =
  match t with
    | "int" -> "Z"
    | "real" -> "R"
    | _ -> t
;;

let rec iter_assertion f a =
  match a with
  | LTrue -> ()
  | LFalse -> () 
  | LAnd(a1,a2) -> 
      fprintf Config.log "iterating on an 'and'@.";
      iter_assertion f a1; 
      iter_assertion f a2 
  | LOr(a1,a2) -> iter_assertion f a1; iter_assertion f a2 
  | LNot(a1) -> iter_assertion f a1
  | LImpl(a1,a2) -> iter_assertion f a1; iter_assertion f a2 
  | LIf(t,a1,a2) -> 
      iter_term f t; iter_assertion f a1; iter_assertion f a2 
  | LLet(id,t,a) -> iter_term f t; iter_assertion f a
  | LForall(id,t,a) -> f t; iter_assertion f a
  | LExists(id,t,a) -> f t; iter_assertion f a
  | LPred(id,l) -> 
      if id = "array_modifiable"
      then fprintf Config.log "iterating on %s@." id;
      f id; List.iter (iter_term f) l
;;

let rec fprint_assertion coq form a =
  match a with
  | LTrue -> fprintf form "%s" (if coq then "True" else "true")
  | LFalse -> fprintf form "%s" (if coq then "False" else "false")
  | LAnd(a1,a2) -> 
      fprintf form "@[(%a@ %s %a)@]" 
	(fprint_assertion coq) a1 
	(if coq then "/\\" else "and")
	(fprint_assertion coq) a2
  | LOr(a1,a2) -> 
      fprintf form "@[(%a@ %s %a)@]" 
	(fprint_assertion coq) a1 
	(if coq then "\\/" else "or")
	(fprint_assertion coq) a2
  | LNot(a1) -> 
      fprintf form "@[(%s %a)@]" 
	(if coq then "~" else "not")
	(fprint_assertion coq) a1
  | LImpl(a1,a2) -> 
      fprintf form "@[<hv 1>(%a ->@ %a)@]" 
	(fprint_assertion coq) a1 (fprint_assertion coq) a2
  | LIf(t,a1,a2) -> 
      fprintf form "@[<hv 1>(if %a@ then %a@ else %a)@]" 
	(fprint_term coq) t (fprint_assertion coq) a1 (fprint_assertion coq) a2
  | LLet(id,t,a) -> 
      fprintf form "@[<hv 1>(let @[<hv 1>%s %s=@ %a in@]@ %a)@]" id
	(if coq then ":" else "")
	(fprint_term coq) t (fprint_assertion coq) a
  | LForall(id,t,a) -> 
      if coq then
	fprintf form "@[<hv 1>(forall (%s:%s),@ %a)@]" 
	  id (coq_base_type t) (fprint_assertion coq) a
      else
	fprintf form "@[<hv 1>(forall %s:%s.@ %a)@]" 
	  id t (fprint_assertion coq) a
  | LExists(id,t,a) -> 
      if coq then
	fprintf form "@[<hv 1>(exists %s:%s,@ %a)@]" 
	  id (coq_base_type t) (fprint_assertion coq) a
      else
	fprintf form "@[<hv 1>(exists %s:%s.@ %a)@]" 
	  id t (fprint_assertion coq) a
  | LPred(("eq"|"eq_value"),[t1;t2]) ->
      fprintf form "@[(%a=%a)@]" 
	(fprint_term coq) t1
	(fprint_term coq) t2
  | LPred("ne_value",[t1;t2]) ->
      if coq then
	begin
	  fprintf form "@[(%a != %a)@]" 
	    (fprint_term coq) t1
	    (fprint_term coq) t2
	end
      else
	begin
	  fprintf form "@[neqv(%a,%a)@]" 
	    (fprint_term coq) t1
	    (fprint_term coq) t2
	end
  | LPred(id,t::tl) ->
      if coq then
	begin
	  fprintf form "@[(%s@ %a" (coq_id id) (fprint_term coq) t;
	  List.iter (fun t -> fprintf form "@ %a" (fprint_term coq) t) tl;
	  fprintf form ")@]"
	end
      else
	begin
	  fprintf form "@[%s(%a" id (fprint_term coq) t;
	  List.iter (fun t -> fprintf form ",@ %a" (fprint_term coq) t) tl;
	  fprintf form ")@]"
	end
  | LPred(_,_) -> assert false      
;;

(*s types *)


type why_type = 
  | Prod_type of string * why_type * why_type      (*r (x:t1)->t2 *)
  | Base_type of why_type list * string       (*r int, real, int list, ... *)
  | Ref_type of why_type
  | Annot_type of 
      assertion * why_type * 
      string list * string list * assertion * ((string * assertion) option)
	(*r { P } t reads r writes w { Q | E => R } *)
;;

let base_type s = Base_type([],s);;
let int_type = base_type "int";;
let bool_type = base_type "bool";;
let real_type = base_type "real";;
let unit_type = base_type "unit";;

let rec iter_why_type f t =
  match t with
    | Prod_type(_,t1,t2) ->
	iter_why_type f t1; iter_why_type f t2
    | Base_type(tl,id) -> List.iter (iter_why_type f) tl; f id
    | Ref_type(t) -> iter_why_type f t 
    | Annot_type (pre,t,reads,writes,post,signals) ->
	fprintf Config.log "iterating on a precondition@.";
	iter_assertion f pre;
	iter_why_type f t;
	List.iter f reads;
	List.iter f writes;
	fprintf Config.log "iterating on a postcondition@.";
	iter_assertion f post;
	Option.iter (fun (_,a) -> iter_assertion f a) signals
;;


let rec fprint_comma_string_list form l =
  match l with
    | [] -> ()
    | x::l -> 
	fprintf form ",%s" x;
	fprint_comma_string_list form l
;;

let rec fprint_type coq anon form t = 
  match t with
    | Prod_type(id,t1,t2) ->
	if id="" or anon then
	  fprintf form "@[<hv 1>%a ->@ %a@]" 
	    (fprint_type coq anon) t1 (fprint_type coq anon) t2
	else
	  if coq then
	    fprintf form "@[<hv 1>forall (%s:%a),@,%a@]" id
	      (fprint_type coq anon) t1 (fprint_type coq anon) t2
	  else
	    fprintf form "@[<hv 1>%s:%a ->@ %a@]" id
	      (fprint_type coq anon) t1 (fprint_type coq anon) t2
    | Base_type([],s) -> 
	if coq
	then
	  fprintf form "%s" (coq_base_type s)
	else
	  fprintf form "%s" s
    | Base_type((t::r) as tl,s) -> 
	if coq
	then
	  begin
	    fprintf form "(%s" s;
	    List.iter (fun t -> fprintf form " %a" (fprint_type coq anon) t) tl;
	    fprintf form ")"
	  end
	else
	  if r=[]
	  then
	    fprintf form "%a %s" (fprint_type coq anon) t s
	  else
	    begin
	      fprintf form "(%a" (fprint_type coq anon) t; 
	      List.iter (fun t -> fprintf form ",%a" (fprint_type coq anon) t) r;
	      fprintf form ") %s" s
	  end
    | Ref_type(t) -> 
	fprintf form "%a ref" (fprint_type coq anon) t
    | Annot_type(p,t,reads,writes,q,signals) ->
	if coq then
	  fprint_assertion coq form (make_impl p q)
	else
	  begin
	    fprintf form "@[@[<hv 2>{ "; 
	    if p <> LTrue 
	    then fprint_assertion coq form p;
	    fprintf form "}@]@ %a@ " (fprint_type coq anon) t;
	    begin
	      match List.sort compare reads with
		| [] -> ()
		| r::l -> 
		    fprintf form "reads %s%a@ " r fprint_comma_string_list l
	    end;
	    begin
	      match List.sort compare writes with
		| [] -> ()
		| r::l -> 
		    fprintf form "writes %s%a@ " r fprint_comma_string_list l
	    end;
	    begin
	      match signals with
		| None -> 
		    fprintf form "@[<hv 2>{ %a }@]@]" (fprint_assertion coq) q
		| Some(e,r) ->
		    fprintf form 
		    "raises %s@ @[<hv 2>{ %a@ | @[<hv 2>%s =>@ %a@] }@]@]" 
		      e
		      (fprint_assertion coq) q
		      e 
		      (fprint_assertion coq) r
	    end		    
		    
	  end
;;


let fprint_why_type = fprint_type false 

let fprint_coq_type = fprint_type true

let rec fprint_logic_type sep form t = 
  match t with
    | Prod_type(_,Base_type([],s),t2) ->
	fprintf form "@[<hv 1>%s%s%a@]" sep s (fprint_logic_type ", ") t2
    | Prod_type(_,Base_type(t::tl,s),t2) ->
	fprintf form "@[<hv 1>%s(%a" sep (fprint_why_type true) t;
	List.iter (fun t -> fprintf form ",%a" (fprint_why_type true) t) tl;
	fprintf form ") %s %a@]" s (fprint_logic_type ", ") t2
    | Base_type([],s) -> fprintf form "-> %s" s
    | Base_type(t::tl,s) -> 
	fprintf form "-> (%a" (fprint_why_type true) t;
	List.iter (fun t -> fprintf form ",%a" (fprint_why_type true) t) tl;
	fprintf form ") %s" s 
    | Ref_type(_) -> assert false (* should never happen *)
    | Prod_type _ -> assert false (* should never happen *)
    | Annot_type _ -> assert false (* should never happen *)
;;

let fprint_logic_type = fprint_logic_type ""

(*s expressions *)

type expr =
  | Cte of constant
  | Var of string
  | And of expr * expr
  | Or of expr * expr
  | Not of expr
  | Void
  | Deref of string
  | If of expr * expr * expr
  | While of 
      expr (* loop condition *)
      * assertion (* invariant *) 
      * term (* variant *) 
      * expr list (* loop body *)
  | Block of expr list
  | Assign of string * expr
  | Let of string * expr * expr
  | Let_ref of string * expr * expr
  | App of expr * expr
  | Raise of string * expr 
  | Try of expr * string * string * expr
  | Fun of (string * why_type) list * 
      assertion * expr * assertion * ((string * assertion) option)
  | Triple of assertion * expr * assertion * ((string * assertion) option)
  | Any of assertion option * why_type * assertion 
  | Assert of assertion  (*r only in blocks *)
  | Label of string
;;

let rec make_app_rec accu l =
  match l with
    | [] -> accu
    | e::r -> make_app_rec (App(accu,e)) r
;;

let make_app id l = make_app_rec (Var(id)) l

let make_while cond inv var e =
  let body = 
    match e with
      | Block(l) -> l
      | _ -> [e]
  in While(cond,inv,var,body)
;;

let make_label label e =
  let body = 
    match e with
      | Block(l) -> l
      | _ -> [e]
  in Block(Label label::body)
;;

let make_pre pre e =  Triple(pre,e,LTrue,None)

let append block e =
  match block with
    | Block(l) -> Block(l@[e])
    | _ -> Block([block;e])
;;

let rec iter_expr f e =
  match e with
    | Cte(c) -> ()
    | Var(id) -> f id
    | And(e1,e2) -> iter_expr f e1; iter_expr f e2
    | Or(e1,e2) -> iter_expr f e1; iter_expr f e2
    | Not(e1) -> iter_expr f e1
    | Void -> ()
    | Deref(id) -> f id
    | If(e1,e2,e3) ->
	iter_expr f e1; iter_expr f e2; iter_expr f e3
    | While(e1,inv,var,e2) ->
	iter_expr f e1; 
	iter_assertion f inv; 
	iter_term f var; 
	List.iter (iter_expr f) e2
    | Block(el) -> List.iter (iter_expr f) el
    | Assign(id,e) -> f id; iter_expr f e
    | Let(id,e1,e2) -> iter_expr f e1; iter_expr f e2
    | Let_ref(id,e1,e2) -> iter_expr f e1; iter_expr f e2
    | App(e1,e2) -> iter_expr f e1; iter_expr f e2
    | Raise(id,e) -> iter_expr f e
    | Try(e1,exc,id,e2) -> iter_expr f e1; iter_expr f e2
    | Fun(params,pre,body,post,signals) ->
	iter_assertion f pre;
	iter_expr f body;
	iter_assertion f post;
	Option.iter (fun (_,a) -> iter_assertion f a) signals
    | Triple(pre,e,post,exceps) ->
	iter_assertion f pre;
	iter_expr f e;
	iter_assertion f post;
	Option.iter (fun (_,a) -> iter_assertion f a) exceps
    | Any (pre,t,post) ->
	Option.iter (iter_assertion f) pre;
	iter_why_type f t;
	iter_assertion f post
    | Assert(e) -> iter_assertion f e
    | Label s -> ()
	  
let rec fprint_expr form e =
  match e with
    | Cte(c) -> fprint_constant false form c
    | Var(id) -> fprintf form "%s" id
    | And(e1,e2) ->
	fprintf form "@[(%a && %a)@]" 
	  fprint_expr e1 fprint_expr e2
    | Or(e1,e2) ->
	fprintf form "@[(%a || %a)@]" 
	  fprint_expr e1 fprint_expr e2
    | Not(e1) ->
	fprintf form "@[(not %a)@]" 
	  fprint_expr e1 
    | Void -> fprintf form "void"
    | Deref(id) -> fprintf form "!%s" id
    | If(e1,e2,e3) ->
	fprintf form "@[<hv 0>if %a@ @[<hv 1>then@ %a@]@ @[<hv 1>else@ %a@]@]" 
	  fprint_expr e1 fprint_expr e2 fprint_expr e3
    | While(e1,inv,var,e2) ->
	fprintf form 
	  "@[<hv 0>while %a do@ @[<hv 1>@[<hv 2>{ @[<hv 2>invariant@ %a@]@ @[<hv 2>variant@ %a@] }@]@ %a@]@ done@]" 
	  fprint_expr e1 
	  (fprint_assertion false) inv
	  (fprint_term false) var
	  fprint_expr_list e2
    | Block(el) ->
	fprintf form "@[<hv 0>begin@ @[<hv 1>  %a@]@ end@]" fprint_expr_list el
    | Assign(id,e) ->
	fprintf form "@[<hv 1>%s := %a@]" 
	  id fprint_expr e
    | Let(id,e1,e2) ->
	fprintf form "@[<hv 0>let %s = %a in@ %a@]" id
	  fprint_expr e1 fprint_expr e2
    | Let_ref(id,e1,e2) ->
	fprintf form "@[<hv 0>let %s = ref %a in@ %a@]" id
	  fprint_expr e1 fprint_expr e2
    | App(e1,e2) ->
	fprintf form "@[<hv 1>(%a %a)@]" fprint_expr e1 fprint_expr e2
    | Raise(id,e) ->
(*
	if e = Void
	then fprintf form "@[<hv 1>raise@ %s@]" id  
	else 
*)
	fprintf form "@[<hv 1>raise@ (%s@ %a)@]" id fprint_expr e 
    | Try(e1,exc,id,e2) ->
	fprintf form "@[<hv 1>try@ %a@ with@ %s %s ->@ %a end@]" 
	  fprint_expr e1 exc id fprint_expr e2
    | Fun(params,pre,body,post,signals) ->
	fprintf form "@[<hv 1>fun @[";
	List.iter 
	  (fun (x,t) -> fprintf form "(%s : %a) " x (fprint_why_type false) t) 
	  params;
	fprintf form "@]->@ @[<hv 0>{ "; 
	if pre <> LTrue 
	then fprint_assertion false form pre;
	fprintf form " }@ begin@ label init;@ %a@ end@]" fprint_expr body;
	begin
	  match signals with
	    | None -> 
		fprintf form "@[<hv 2>{ %a }@]@]" (fprint_assertion false) post
	    | Some(e,r) ->
		fprintf form 
		  "@[<hv 2>{ %a@ | @[<hv 2>%s =>@ %a@] }@]@]" 
		  (fprint_assertion false) post
		  e 
		  (fprint_assertion false) r
	end		    

    | Triple(pre,e,LTrue,None) ->
	fprintf form "@[<hv 0>{ %a }@ (%a)@ { }@]" 
	  (fprint_assertion false) pre
	  fprint_expr e
    | Triple(pre,e,post,exceps) ->
	fprintf form "@[<hv 0>{ %a }@ (%a)@ " 
	  (fprint_assertion false) pre
	  fprint_expr e;
	begin
	  match exceps with
	    | None -> 
		fprintf form "{ %a }@]" 
		  (fprint_assertion false) post
	    | Some(e,r) ->
		fprintf form 
		  "@[<hv 2>{ %a@ | @[<hv 2>%s =>@ %a@] }@]@]" 
		  (fprint_assertion false) post
		  e 
		  (fprint_assertion false) r
	end
    | Any (None,t,post) ->
	fprintf form "@[<hv 0>[{ }@ (%a)@ { %a }]@]" 
	  (fprint_why_type false) t (fprint_assertion false) post
    | Any (Some pre,t,post) ->
	fprintf form "@[<hv 0>[{ %a }@ (%a)@ { %a }]@]" 
	  (fprint_assertion false) pre 
	  (fprint_why_type false) t (fprint_assertion false) post
    | Assert(e) ->
	fprintf form "@[<hv 0>assert@ { %a }@]" 
	  (fprint_assertion false) e
    | Label s ->
	fprintf form "@[<hv 0>label@ %s@]" s
	  

and fprint_expr_list form l =
  match l with
    | [] -> ()
    | e::l ->
	fprintf form "%a" fprint_expr e;
	fprint_expr_end_list form l

and fprint_expr_end_list form l =
  match l with
    | [] -> ()
    | e::l ->
	fprintf form ";@ %a" fprint_expr e;
	fprint_expr_end_list form l
;;

type why_decl =
  | Param of string * why_type         (*r parameter in why *)
(*
  | Paramfun of string * why_type         (*r parameter in why *)
*)
  | Def of string * expr               (*r global let in why *)
(*
  | Deffun of string * expr            (*r global let in why *)
*)
  | External of string * why_type      (*r external decl in why *)
  | Logic of string * why_type         (*r logic decl in why *)

type coq_decl =
  | Parameter  of string * why_type    (*r Parameter in Coq *)
  | Definition of string * expr        (*r Definition in Coq *) 
  | PropDefinition of string * (string * why_type) list * assertion  (*r Definition in Coq *) 
  | Axiom of string * why_type         (*r Axiom in Coq *)
(*
  | HintUnfold of string                 (*r Hint Unfold in Coq *)
*)
  | Verbatim of string                 (*r Text in Coq *)


let get_why_id d =
  match d with
    | Param(id,_) -> id
(*
    | Paramfun(id,_) -> id
*)
    | External(id,_) -> id
    | Logic(id,_) -> id
    | Def(id,_) -> id
(*
    | Deffun(id,_) -> id
*)
      
let iter_why_decl f d =
  match d with
    | Param(_,t) -> iter_why_type f t
(*
    | Paramfun(id,t) -> iter_why_type f t
*)
    | External(id,t) -> iter_why_type f t
    | Logic(id,t) -> iter_why_type f t
    | Def(id,t) -> iter_expr f t
(*
    | Deffun(id,t) -> iter_expr f t
*)
      
let get_coq_id d =
  match d with
    | Parameter(id,_) -> id
    | Axiom(id,_) -> id
    | Definition(id,_) -> id
    | PropDefinition(id,_,_) -> id
(*
    | HintUnfold(id) -> assert false
*)
    | Verbatim(s) -> assert false


type state = [`TODO | `RUNNING | `DONE ];;

type 'a decl = { mutable state : state; decl : 'a };;

module StringMap = Map.Make(String);;

exception Recursion;;

let rec do_topo get_id decl_map iter_fun output_fun d =
  match d.state with
    | `DONE -> ()
    | `RUNNING -> raise Recursion
    | `TODO ->
	fprintf Config.log "try output decl of %s@." (get_id d.decl);
	d.state <- `RUNNING;
	iter_fun
	  (fun id ->
	     try 
	       let s = StringMap.find id decl_map in
	       do_topo get_id decl_map iter_fun output_fun s
	     with
		 Not_found -> ())
	  d.decl;	
	output_fun d.decl;
	d.state <- `DONE
;;


let build_map get_id decl_list =
  List.fold_left
    (fun acc decl ->
       let id = get_id decl in
       StringMap.add id { state = `TODO ; decl = decl } acc)
    StringMap.empty
    decl_list
;;

let fprint_why_decl form d =
  match d with
    | Param(id,t) ->
	fprintf form "@[<hv 1>parameter %s :@ %a@]@.@." id 
	  (fprint_why_type false) t
(*
    | Paramfun(id,t) ->
	fprintf form "@[<hv 1>parameter %s :@ %a@]@.@." id 
	  (fprint_why_type false) t
*)
    | External(id,t) ->
	fprintf form "@[<hv 1>external parameter %s :@ %a@]@.@." id 
	  (fprint_why_type false) t
    | Logic(id,t) ->
	fprintf form "@[<hv 1>external logic %s :@ %a@]@.@." id 
	  fprint_logic_type t
    | Def(id,e) ->
	fprintf form "@[<hv 1>let %s =@ %a@]@.@." id fprint_expr e
(*
    | Deffun(id,e) ->
	fprintf form "@[<hv 1>let %s =@ %a@]@.@." id fprint_expr e
*)

let iter_coq_decl f d =
  match d with
    | Parameter(id,t) -> iter_why_type f t
    | Axiom(id,t) -> iter_why_type f t
    | Definition(id,e) -> iter_expr f e
    | PropDefinition(id,args,a) -> iter_assertion f a
(*
    | HintUnfold(id) -> f id
*)
    | Verbatim (s) -> assert false

let fprint_coq_decl form d =
  match d with
    | Parameter(id,t) -> 
	fprintf form "@[<hv 1>Parameter %s :@ %a.@]@.@." id 
	  (fprint_coq_type true) t
    | Axiom(id,t) -> 
	fprintf form "@[<hv 1>Axiom %s :@ %a.@]@.@." id 
	  (fprint_coq_type false) t
    | Definition(id,e) ->
	fprintf form "@[<hv 1>Definition %s :=@ %a.@]@.@." id fprint_expr e
    | PropDefinition(id,args,a) ->
	fprintf form "@[<hv 1>Definition %s := fun@ " id;
	List.iter 
	  (fun (x,t) -> fprintf form "(%s : %a)@ " x (fprint_coq_type true) t) 
	  args;
	fprintf form "=> %a.@]@.@." (fprint_assertion true) a
(*
    | HintUnfold(id) ->
	fprintf form "Hints Unfold %s.@." id
*)
    | Verbatim (s) -> fprintf form "%s@." s


let output_decls get_id iter_decl output_decl decls =
  let map = build_map get_id decls in
  StringMap.iter
    (fun id decl ->
       do_topo get_id map iter_decl output_decl decl)
    map
;;

let output_why_decls form decls =
(* partition is necessary because Logic may have same name as parameter *)
  let (logic,other) = 
    List.partition
      (function Logic _ -> true | _ -> false) 
      decls
  in
  List.iter (fprint_why_decl form) logic;
  output_decls get_why_id iter_why_decl (fprint_why_decl form) other
;;

let output_coq_decls form decls =
  let (coqdef,other) =
    List.partition
      (function (* HintUnfold _ *) | Verbatim _ -> false 
	 | _ -> true) 
      decls
  in
  output_decls get_coq_id iter_coq_decl (fprint_coq_decl form) coqdef;
  List.iter (fprint_coq_decl form) other;
;;



(*

Simplify output

*)

let fprint_sx_constant form e =
  match e with
    | Prim_int(n) -> 
	fprintf form "%d" n
    | Prim_real(i,f,"") -> 
	fprintf form "|%s.%s|" i f
    | Prim_real(i,f,e) -> 
	fprintf form "|%s.%se%s|" i f e
    | Prim_bool(b) -> 
	fprintf form "|@@%b|" b
;;

let sx_fun_symbol id =
  match id with
    | "eq_value" -> "EqBool"
    | "ne_value" -> "NeqBool"
    | "add_int" -> "+"
    | "mul_int" -> "*"
    | "sub_int" -> "-"
    | _ -> id
;;

(*
let fprintf_sx_id form id = 
  if String.get id 0 = '_'
  then fprintf form "U%s" id
  else fprintf form "%s" id
*)

let rec fprint_sx_term form t =
  match t with
  | LConst(c) -> fprint_sx_constant form c
  | LApp(id,l) ->
      fprintf form "@[(%s" (sx_fun_symbol id);
      List.iter (fun t -> fprintf form "@ %a" fprint_sx_term t) l;
      fprintf form ")@]"
  | LVar(id) -> fprintf form "%s" id
  | LVarAtLabel(id,l) -> fprintf form "|%s@@%s|" id l
;;

let sx_pred_symbol id =
  match id with
    | "eq_value" -> "EQ"
    | "ne_value" -> "NEQ"
    | "eq_bool" -> "EQ"
    | "neq_bool" -> "NEQ"
    | "eq_int" -> "EQ"
    | "neq_int" -> "NEQ"
    | "le_int" -> "<="
    | "lt_int" -> "<"
    | "ge_int" -> ">="
    | "gt_int" -> ">"
    | "eq _" -> "EQ"
    | "eq" -> "EQ"
    | "eq ?" -> "EQ"
    | _ -> id
;;


let rec fprint_sx_assertion form a =
  match a with
  | LTrue -> fprintf form "%s" "TRUE"
  | LFalse -> fprintf form "%s" "FALSE"
  | LAnd(a1,a2) -> 
      fprintf form "@[(AND@ %a@ %a)@]" 
	fprint_sx_assertion a1 
	fprint_sx_assertion a2
  | LOr(a1,a2) -> 
      fprintf form "@[(OR@ %a@ %a)@]" 
	fprint_sx_assertion a1 
	fprint_sx_assertion a2
  | LNot(a1) -> 
      fprintf form "@[(NOT@ %a)@]" 
      fprint_sx_assertion a1
  | LImpl(a1,a2) -> 
      fprintf form "@[(IMPLIES@ %a@ %a)@]" 
      fprint_sx_assertion a1 
      fprint_sx_assertion a2
  | LIf(t,a1,a2) -> 
      assert false
  | LLet(id,t,a) -> 
      fprintf form "@[(FORALL (%s)@ (IMPLIES@ (EQ@ %s@ %a)@ %a))@]" 
	id id
	fprint_sx_term t fprint_sx_assertion a
  | LForall(id,t,a) -> 
      fprintf form "@[(FORALL (%s)@ %a)@]" 
      id fprint_sx_assertion a
  | LExists(id,t,a) -> 
      fprintf form "@[<hv 1>(EXISTS (%s) %a)@]" 
      id fprint_sx_assertion a
  | LPred(id,l) ->
      fprintf form "@[(%s" (sx_pred_symbol id);
      List.iter (fun t -> fprintf form "@ %a" fprint_sx_term t) l;
      fprintf form ")@]"
;;

let rec fprint_sx_type form t = 
  match t with
    | Prod_type(id,t1,t2) ->
	if id=""
	then
	  fprintf form "@[<hv 1>(IMPLIES@ %a@ %a)@]" 
	    fprint_sx_type t1 fprint_sx_type t2
	else
	  fprintf form "@[<hv 1>(FORALL (%s)@ %a)@]" 
	    id fprint_sx_type t2
    | Base_type([],s) -> fprintf form "%s" s
    | Base_type(tl,s) -> assert false
    | Ref_type(t) -> assert false
    | Annot_type(p,t,reads,writes,q,signals) ->
	fprint_sx_assertion form (make_impl p q)
;;


let rec fprint_sx_expr form e =
  match e with
    | Cte(c) -> fprint_sx_constant form c
    | Var(id) -> fprintf form "%s" id
    | And(e1,e2) ->
	fprintf form "@[(AndBool %a %a)@]" 
	  fprint_sx_expr e1 fprint_sx_expr e2
    | Or(e1,e2) ->
	fprintf form "@[(OrBool %a %a)@]" 
	  fprint_sx_expr e1 fprint_sx_expr e2
    | Not(e1) ->
	fprintf form "@[(NotBool %a)@]" 
	  fprint_sx_expr e1 
    | Void -> assert false
    | Deref(id) -> assert false
    | If(e1,e2,e3) -> assert false
    | While(e1,inv,var,e2) ->
	assert false
    | Block(el) ->
	assert false
    | Assign(id,e) ->
	assert false
    | Let(id,e1,e2) ->
	assert false
    | Let_ref(id,e1,e2) ->
	assert false
    | App(e1,e2) ->
	assert false
    | Raise(id,e) ->
	assert false
    | Try(e1,exc,id,e2) ->
	assert false
    | Fun(params,pre,body,post,signals) ->
	assert false
    | Triple(pre,e,LTrue,None) ->
	assert false
    | Triple(pre,e,post,exceps) ->
	assert false
    | Any _ -> 
	assert false
    | Assert(e) ->
	assert false
    | Label s ->
	assert false

let fprint_sx_decl form d =
  match d with
    | Parameter(id,t) -> 
	fprintf form ";; Parameter %s : ...@.@." id (* fprint_sx_type t *)
    | Axiom(id,t) ->  
	fprintf form ";; Axiom %s@.@[<hv 1>(BG_PUSH@ %a)@]@.@." 
	id fprint_sx_type t
    | Definition(id,e) ->
	fprintf form ";; Definition %s@.@[<hv 1>(BG_PUSH@ (EQ@ %s@ %a))@]@.@." 
	id id fprint_sx_expr e
    | PropDefinition(id,args,a) ->
	fprintf form ";; Predicate %s@.@[<hv 1>(DEFPRED (%s" id id;
	List.iter 
	  (fun (x,t) -> fprintf form " %s" x) 
	  args;
	fprintf form ")@ %a)@]@.@." fprint_sx_assertion a
(*
    | HintUnfold(id) -> ()
*)
    | Verbatim (s) -> 
	fprintf form ";; Verbatim %s@.@." s

let output_sx_decls form decls =
  let coqdef =
    List.filter
      (function (* HintUnfold _ *) | Verbatim _ -> false 
	 | _ -> true)  
      decls
  in
  output_decls get_coq_id iter_coq_decl (fprint_sx_decl form) coqdef;
;;

(*

haRVey output

*)

let fprint_rv_constant form e =
  match e with
    | Prim_int(n) -> 
	fprintf form "arith_%d" n
    | Prim_real(i,f,e) -> 
	fprintf form "%s.%se%s" i f e
    | Prim_bool(b) -> 
	fprintf form "%b" b
;;

let rv_fun_symbol id =
  match id with
    | "eq_value" -> "eq_bool"
    | "ne_value" -> "neq_bool"
    | "add_int"
    | _ -> id
;;

let rec fprint_rv_term form t =
  match t with
  | LConst(c) -> fprint_rv_constant form c
  | LApp(id,l) ->
      fprintf form "@[(%s" (rv_fun_symbol id);
      List.iter (fun t -> fprintf form "@ %a" fprint_rv_term t) l;
      fprintf form ")@]"
  | LVar(id) -> fprintf form "%s" id
  | LVarAtLabel(id,l) -> fprintf form "|%s@@%s|" id l
;;

let rv_pred_symbol id =
  match id with
    | "eq_value" -> "="
(*
    | "ne_value" -> "NEQ"
*)
    | "eq_int" -> "="
(*
    | "neq_int" -> "NEQ"
*)
(*
    | "le_int" -> "<="
    | "lt_int" -> "<"
    | "ge_int" -> ">="
    | "gt_int" -> ">"
*)
    | "eq _" -> "="
    | _ -> id
;;


let rec fprint_rv_assertion form a =
  match a with
  | LTrue -> fprintf form "%s" "true"
  | LFalse -> fprintf form "%s" "false"
  | LAnd(a1,a2) -> 
      fprintf form "@[(and@ %a@ %a)@]" 
	fprint_rv_assertion a1 
	fprint_rv_assertion a2
  | LOr(a1,a2) -> 
      fprintf form "@[(or@ %a@ %a)@]" 
	fprint_rv_assertion a1 
	fprint_rv_assertion a2
  | LNot(a1) -> 
      fprintf form "@[(not@ %a)@]" 
      fprint_rv_assertion a1
  | LImpl(a1,a2) -> 
      fprintf form "@[(->@ %a@ %a)@]" 
      fprint_rv_assertion a1 
      fprint_rv_assertion a2
  | LIf(t,a1,a2) -> 
      assert false
  | LLet(id,t,a) -> 
      fprintf form "@[(forall %s@ (->@ (=@ %s@ %a)@ %a))@]" id id
	fprint_rv_term t fprint_rv_assertion a
  | LForall(id,t,a) -> 
      fprintf form "@[(forall %s@ %a)@]" 
      id fprint_rv_assertion a
  | LExists(id,t,a) -> 
      fprintf form "@[<hv 1>(exists %s %a)@]" 
      id fprint_rv_assertion a
  | LPred(id,l) ->
      begin
	match id with
	  | "neq_int" | "ne_value" ->
	      fprint_rv_assertion form (LNot (LPred("=",l)))
	  | _ ->
	      fprintf form "@[(%s" (rv_pred_symbol id);
	      List.iter (fun t -> fprintf form "@ %a" fprint_rv_term t) l;
	      fprintf form ")@]"
      end
;;

let rec fprint_rv_type form t = 
  match t with
    | Prod_type(id,t1,t2) ->
	if id=""
	then
	  fprintf form "@[<hv 1>(->@ %a@ %a)@]" 
	    fprint_rv_type t1 fprint_rv_type t2
	else
	  fprintf form "@[<hv 1>(forall %s@ %a)@]" 
	    id fprint_rv_type t2
    | Base_type([],s) -> fprintf form "%s" s
    | Base_type(tl,s) -> assert false
    | Ref_type(t) -> assert false
    | Annot_type(p,t,reads,writes,q,signals) ->
	fprint_rv_assertion form (make_impl p q)
;;


let rec fprint_rv_expr form e =
  match e with
    | Cte(c) -> fprint_rv_constant form c
    | Var(id) -> fprintf form "%s" id
    | And(e1,e2) ->
	fprintf form "@[(and_bool %a %a)@]" 
	  fprint_rv_expr e1 fprint_rv_expr e2
    | Or(e1,e2) ->
	fprintf form "@[(or_bool %a %a)@]" 
	  fprint_rv_expr e1 fprint_rv_expr e2
    | Not(e1) ->
	fprintf form "@[(not_bool %a)@]" 
	  fprint_rv_expr e1 
    | Void -> assert false
    | Deref(id) -> assert false
    | If(e1,e2,e3) -> assert false
    | While(e1,inv,var,e2) ->
	assert false
    | Block(el) ->
	assert false
    | Assign(id,e) ->
	assert false
    | Let(id,e1,e2) ->
	assert false
    | Let_ref(id,e1,e2) ->
	assert false
    | App(e1,e2) ->
	assert false
    | Raise(id,e) ->
	assert false
    | Try(e1,exc,id,e2) ->
	assert false
    | Fun(params,pre,body,post,signals) ->
	assert false
    | Triple(pre,e,LTrue,None) ->
	assert false
    | Triple(pre,e,post,exceps) ->
	assert false
    | Any _ ->
	assert false
    | Assert(e) ->
	assert false
    | Label s ->
	assert false

let fprint_rv_decl form d =
  match d with
    | Parameter(id,t) -> 
	fprintf form ";; Parameter %s : ...@.@." id (* fprint_rv_type t *)
    | Axiom(id,t) ->  
	fprintf form ";; Axiom %s@.@[<hv 1>%a@]@.@." 
	id fprint_rv_type t
    | Definition(id,e) ->
	fprintf form ";; Definition %s@.@[<hv 1>(=@ %s@ %a)@]@.@." 
	id id fprint_rv_expr e
    | PropDefinition(id,args,a) ->
	fprintf form ";; Predicate %s@." id;
	fprintf form "@[<hv 1>(forall";
	List.iter 
	  (fun (x,t) -> fprintf form " %s" x) 
	  args;
	fprintf form "@ @[<hv 1>(<-> (%s" id;
	List.iter 
	  (fun (x,t) -> fprintf form " %s" x) 
	  args;
	fprintf form ")@ %a)@])@]@.@." fprint_rv_assertion a
(*
    | HintUnfold(id) -> ()
*)
    | Verbatim (s) -> 
	fprintf form ";; Verbatim %s@.@." s





let output_rv_decls form decls =
  let coqdef =
    List.filter
      (function (* HintUnfold _ *) | Verbatim _ -> false 
	 | _ -> true) 
      decls
  in
  output_decls get_coq_id iter_coq_decl (fprint_rv_decl form) coqdef
;;





(*

PVS output

*)

(*
let rationalize s =
  let n = String.length s in
  let i = String.index s '.' in
  let d = n - i - 1 in
  String.sub s 0 i ^ String.sub s (succ i) d, "1" ^ String.make d '0'
*)

let fprint_pvs_constant form e =
  match e with
    | Prim_int(n) -> 
	fprintf form "(%d::int)" n
    | Prim_real(i,f,e) ->
	let f = if f = "0" then "" else f in
	let l = String.length f in
	let e = int_of_string e - l in
	if e=0
	then
	  fprintf form "%s%s" i f 
	else
	  if e < 0
	  then
	    fprintf form "(%s%s / 1%s)" i f (String.make (-e) '0')
	  else
	    fprintf form "(%s%s%s)" i f (String.make e '0')
(*
    | Prim_real("0" | "0.0") -> 
	fprintf form "0"
    | Prim_real(f) -> 
	let n,d = rationalize f in
	if d = "1" then 
	  fprintf form "%s" n 
	else 
	  fprintf form "(%s / %s)" n d
*)
    | Prim_bool(b) -> 
	fprintf form "%b" b
;;

let pvs_fun_symbol id =
  match id with
    | "eq_value" -> "="
    | "ne_value" -> "/="
    | "add_int" | "add_real" -> "+"
    | "mul_int" | "mul_real" -> "*"
    | "sub_int" | "sub_real" -> "-"
    | "div_real" -> "/"
    | "neg_int" | "neg_real" -> "-"
    | _ -> id
;;

let rec print_list sep print fmt = function
  | [] -> ()
  | [x] -> print fmt x
  | x :: r -> print fmt x; sep fmt (); print_list sep print fmt r

let comma fmt () = fprintf fmt ",@ "

let rec fprint_pvs_term form t =
  match t with
  | LConst(c) -> fprint_pvs_constant form c
  | LApp(id,l) ->
      let s = pvs_fun_symbol id in
      fprintf form "@[%s(%a)@]" s (print_list comma fprint_pvs_term) l
  | LVar(id) -> fprintf form "%s" id
  | LVarAtLabel(id,l) -> assert false
;;

let pvs_infix id =
  match id with
    | "eq_value" -> "="
    | "ne_value" -> "/="
    | "eq_bool" -> "="
    | "neq_bool" -> "/="
    | "eq_int" | "eq_real" -> "="
    | "neq_int" | "neq_real" -> "/="
    | "le_int" | "le_real" -> "<="
    | "lt_int" | "lt_real" -> "<"
    | "ge_int" | "ge_real" -> ">="
    | "gt_int" | "gt_real" -> ">"
    | "eq _" -> "="
    | "eq" -> "="
    | "eq ?" -> "="
    | _ -> raise Exit
;;


let rec uncurry = function
  | Prod_type (_,t1,t2) -> let l,t2 = uncurry t2 in t1::l, t2
  | _ as t2 -> [], t2

let rec fprint_pvs_type form t = 
  match t with
    | Prod_type _ ->
	let l1,t2 = uncurry t in 
	fprintf form "[%a -> %a]" 
	  (print_list comma fprint_pvs_type) l1 fprint_pvs_type t2
    | Base_type([],s) -> fprintf form "%s" s
    | Base_type(tl,s) -> 
	fprintf form "%s[%a]" s (print_list comma fprint_pvs_type) tl
    | Ref_type(t) -> assert false
    | Annot_type(p,t,reads,writes,q,signals) -> assert false

let rec fprint_pvs_type_assertion form t =
  match t with
    | Prod_type(id,t1,t2) ->
	fprintf form "@[(FORALL (%s:%a):@ %a)@]" 
	  id fprint_pvs_type t1 fprint_pvs_type_assertion t2
    | Base_type([],s) -> fprintf form "%s" s
    | Base_type(tl,s) -> 
	fprintf form "%s[%a]" s (print_list comma fprint_pvs_type) tl
    | Ref_type(t) -> assert false
    | Annot_type(p,t,reads,writes,q,signals) ->
	fprint_pvs_assertion form (make_impl p q)
      
and pvs_type_of_string form = function
  | "unit" | "bool" | "int" | "value" | "real" as s -> fprintf form "%s" s
  | _ -> assert false

and fprint_pvs_assertion form a =
  match a with
  | LTrue -> fprintf form "%s" "TRUE"
  | LFalse -> fprintf form "%s" "FALSE"
  | LAnd(a1,a2) -> 
      fprintf form "@[(%a AND@ %a)@]" 
	fprint_pvs_assertion a1 
	fprint_pvs_assertion a2
  | LOr(a1,a2) -> 
      fprintf form "@[(%a OR@ %a)@]" 
	fprint_pvs_assertion a1 
	fprint_pvs_assertion a2
  | LNot(a1) -> 
      fprintf form "@[(NOT@ %a)@]" 
      fprint_pvs_assertion a1
  | LImpl(a1,a2) -> 
      fprintf form "@[(%a IMPLIES@ %a)@]" 
      fprint_pvs_assertion a1 
      fprint_pvs_assertion a2
  | LIf(t,a1,a2) -> 
      assert false
  | LLet(id,t,a) -> 
      fprintf form "@[(LET %s = %a@ IN %a)@]"
	id fprint_pvs_term t fprint_pvs_assertion a
  | LForall(id,t,a) -> 
      fprintf form "@[(FORALL (%s:%a):@ %a)@]" 
      id pvs_type_of_string t fprint_pvs_assertion a
  | LExists(id,t,a) -> 
      fprintf form "@[<hv 1>(EXISTS (%s:%a): %a)@]" 
      id pvs_type_of_string t fprint_pvs_assertion a
  | LPred(id,l) ->
      (try
	 let s = pvs_infix id in
	 (match l with
	    | [t1; t2] ->
		fprintf form "@[(%a %s@ %a)@]" 
	          fprint_pvs_term t1 s fprint_pvs_term t2
	    | _ -> raise Exit)
       with Exit ->
	 fprintf form "@[%s(%a)@]" id (print_list comma fprint_pvs_term) l)
;;



let rec fprint_pvs_expr form e =
  match e with
    | Cte(c) -> fprint_pvs_constant form c
    | Var(id) -> fprintf form "%s" id
    | And(e1,e2) ->
	fprintf form "@[(AndBool %a %a)@]" 
	  fprint_pvs_expr e1 fprint_pvs_expr e2
    | Or(e1,e2) ->
	fprintf form "@[(OrBool %a %a)@]" 
	  fprint_pvs_expr e1 fprint_pvs_expr e2
    | Not(e1) ->
	fprintf form "@[(NotBool %a)@]" 
	  fprint_pvs_expr e1 
    | Void -> assert false
    | Deref(id) -> assert false
    | If(e1,e2,e3) -> assert false
    | While(e1,inv,var,e2) ->
	assert false
    | Block(el) ->
	assert false
    | Assign(id,e) ->
	assert false
    | Let(id,e1,e2) ->
	assert false
    | Let_ref(id,e1,e2) ->
	assert false
    | App(e1,e2) ->
	assert false
    | Raise(id,e) ->
	assert false
    | Try(e1,exc,id,e2) ->
	assert false
    | Fun(params,pre,body,post,signals) ->
	assert false
    | Triple(pre,e,LTrue,None) ->
	assert false
    | Triple(pre,e,post,exceps) ->
	assert false
    | Any _ ->
	assert false
    | Assert(e) ->
	assert false
    | Label s ->
	assert false

let fprint_pvs_decl form d =
  match d with
    | Parameter(id,t) -> 
	fprintf form "%% Parameter %s@.@[%s: %a@]@.@." id id fprint_pvs_type t
    | Axiom(id,t) ->  
	fprintf form "%% Axiom %s@.@[<hv 1>%s: AXIOM %a@]@.@."
	id id fprint_pvs_type_assertion t
    | Definition(id,e) ->
	fprintf form "%% Definition %s@.@[<hv 1>(BG_PUSH@ (EQ@ %s@ %a))@]@.@." 
	id id fprint_pvs_expr e
    | PropDefinition(id,args,a) ->
	fprintf form "%% Predicate %s@.@[<hv 1>%s(%a) : bool =@\n%a@\n@\n" 
	  id id
	  (print_list comma 
	     (fun form (x,t) -> fprintf form "%s:%a" x fprint_pvs_type t)) args
	  fprint_pvs_assertion a
    | Verbatim (s) -> 
	fprintf form "%% Verbatim %s@.@." s

let output_pvs_decls form decls =
  let coqdef =
    List.filter
      (function (* HintUnfold _ *) | Verbatim _ -> false 
	 | _ -> true)  
      decls
  in
  output_decls get_coq_id iter_coq_decl (fprint_pvs_decl form) coqdef;
;;

